@@ -1,7 +1,13 @@
 package com.artifex.mupdf.fitz;
 
+import java.util.Date;
+
 public class PDFAnnotation extends Annotation
 {
+	static {
+		Context.init();
+	}
+
 	private PDFAnnotation(long p) { super(p); }
 
 	/* IMPORTANT: Keep in sync with mupdf/pdf/annot.h */
@@ -32,9 +38,20 @@ public class PDFAnnotation extends Annotation
 	public static final int TYPE_3D = 24;
 	public static final int TYPE_UNKNOWN = -1;
 
+	public static final int LINE_ENDING_NONE = 0;
+	public static final int LINE_ENDING_SQUARE = 1;
+	public static final int LINE_ENDING_CIRCLE = 2;
+	public static final int LINE_ENDING_DIAMOND = 3;
+	public static final int LINE_ENDING_OPEN_ARROW = 4;
+	public static final int LINE_ENDING_CLOSED_ARROW = 5;
+	public static final int LINE_ENDING_BUTT = 6;
+	public static final int LINE_ENDING_R_OPEN_ARROW = 7;
+	public static final int LINE_ENDING_R_CLOSED_ARROW = 8;
+	public static final int LINE_ENDING_SLASH = 9;
+
 	public native int getType();
 	public native int getFlags();
-	public native int setFlags(int flags);
+	public native void setFlags(int flags);
 	public native String getContents();
 	public native void setContents(String contents);
 	public native Rect getRect();
@@ -45,14 +62,27 @@ public class PDFAnnotation extends Annotation
 	public native void setColor(float[] color);
 	public native float[] getInteriorColor();
 	public native void setInteriorColor(float[] color);
+	public native String getAuthor();
+	public native void setAuthor(String author);
+	protected native long getModificationDateNative();
+	protected native void setModificationDate(long time);
+	public Date getModificationDate() {
+		return new Date(getModificationDateNative());
+	}
+	public void setModificationDate(Date date) {
+		setModificationDate(date.getTime());
+	}
 
 	public native int[] getLineEndingStyles();
 	public native void setLineEndingStyles(int startStyle, int endStyle);
+	public void setLineEndingStyles(int[] styles) {
+		setLineEndingStyles(styles[0], styles[1]);
+	}
 
 	public native float[] getVertices();
 	public native void setVertices(float[] vertices);
-	public native float[] getQuadPoints();
-	public native void setQuadPoints(float[] quadPoints);
+	public native float[][] getQuadPoints();
+	public native void setQuadPoints(float[][] quadPoints);
 	public native float[][] getInkList();
 	public native void setInkList(float[][] inkList);
 
