@@ -1,5 +1,8 @@
+#include "mupdf/fitz.h"
 #include "fitz-imp.h"
 
+#include <string.h>
+
 fz_device *
 fz_new_device_of_size(fz_context *ctx, int size)
 {
@@ -13,33 +16,40 @@ fz_close_device(fz_context *ctx, fz_device *dev)
 {
 	if (dev == NULL)
 		return;
-	if (dev->close_device)
-		dev->close_device(ctx, dev);
 
-	/* Don't call more than once! */
-	dev->close_device = NULL;
+	fz_try(ctx)
+	{
+		if (dev->close_device)
+			dev->close_device(ctx, dev);
+	}
+	fz_always(ctx)
+	{
+		dev->close_device = NULL; /* Don't call more than once! */
 
-	/* And disable all further device calls. */
-	dev->fill_path = NULL;
-	dev->stroke_path = NULL;
-	dev->clip_path = NULL;
-	dev->clip_stroke_path = NULL;
-	dev->fill_text = NULL;
-	dev->stroke_text = NULL;
-	dev->clip_text = NULL;
-	dev->clip_stroke_text = NULL;
-	dev->ignore_text = NULL;
-	dev->fill_shade = NULL;
-	dev->fill_image = NULL;
-	dev->fill_image_mask = NULL;
-	dev->clip_image_mask = NULL;
-	dev->pop_clip = NULL;
-	dev->begin_mask = NULL;
-	dev->end_mask = NULL;
-	dev->begin_group = NULL;
-	dev->end_group = NULL;
-	dev->begin_tile = NULL;
-	dev->end_tile = NULL;
+		/* And disable all further device calls. */
+		dev->fill_path = NULL;
+		dev->stroke_path = NULL;
+		dev->clip_path = NULL;
+		dev->clip_stroke_path = NULL;
+		dev->fill_text = NULL;
+		dev->stroke_text = NULL;
+		dev->clip_text = NULL;
+		dev->clip_stroke_text = NULL;
+		dev->ignore_text = NULL;
+		dev->fill_shade = NULL;
+		dev->fill_image = NULL;
+		dev->fill_image_mask = NULL;
+		dev->clip_image_mask = NULL;
+		dev->pop_clip = NULL;
+		dev->begin_mask = NULL;
+		dev->end_mask = NULL;
+		dev->begin_group = NULL;
+		dev->end_group = NULL;
+		dev->begin_tile = NULL;
+		dev->end_tile = NULL;
+	}
+	fz_catch(ctx)
+		fz_rethrow(ctx);
 }
 
 fz_device *
@@ -106,22 +116,22 @@ pop_clip_stack(fz_context *ctx, fz_device *dev)
 
 void
 fz_fill_path(fz_context *ctx, fz_device *dev, const fz_path *path, int even_odd, const fz_matrix *ctm,
-	fz_colorspace *colorspace, const float *color, float alpha)
+	fz_colorspace *colorspace, const float *color, float alpha, const fz_color_params *color_params)
 {
 	if (dev->error_depth)
 		return;
 	if (dev->fill_path)
-		dev->fill_path(ctx, dev, path, even_odd, ctm, colorspace, color, alpha);
+		dev->fill_path(ctx, dev, path, even_odd, ctm, colorspace, color, alpha, color_params);
 }
 
 void
 fz_stroke_path(fz_context *ctx, fz_device *dev, const fz_path *path, const fz_stroke_state *stroke, const fz_matrix *ctm,
-	fz_colorspace *colorspace, const float *color, float alpha)
+	fz_colorspace *colorspace, const float *color, float alpha, const fz_color_params *color_params)
 {
 	if (dev->error_depth)
 		return;
 	if (dev->stroke_path)
-		dev->stroke_path(ctx, dev, path, stroke, ctm, colorspace, color, alpha);
+		dev->stroke_path(ctx, dev, path, stroke, ctm, colorspace, color, alpha, color_params);
 }
 
 void
@@ -192,22 +202,22 @@ fz_clip_stroke_path(fz_context *ctx, fz_device *dev, const fz_path *path, const
 
 void
 fz_fill_text(fz_context *ctx, fz_device *dev, const fz_text *text, const fz_matrix *ctm,
-	fz_colorspace *colorspace, const float *color, float alpha)
+	fz_colorspace *colorspace, const float *color, float alpha, const fz_color_params *color_params)
 {
 	if (dev->error_depth)
 		return;
 	if (dev->fill_text)
-		dev->fill_text(ctx, dev, text, ctm, colorspace, color, alpha);
+		dev->fill_text(ctx, dev, text, ctm, colorspace, color, alpha, color_params);
 }
 
 void
 fz_stroke_text(fz_context *ctx, fz_device *dev, const fz_text *text, const fz_stroke_state *stroke, const fz_matrix *ctm,
-	fz_colorspace *colorspace, const float *color, float alpha)
+	fz_colorspace *colorspace, const float *color, float alpha, const fz_color_params *color_params)
 {
 	if (dev->error_depth)
 		return;
 	if (dev->stroke_text)
-		dev->stroke_text(ctx, dev, text, stroke, ctm, colorspace, color, alpha);
+		dev->stroke_text(ctx, dev, text, stroke, ctm, colorspace, color, alpha, color_params);
 }
 
 void
@@ -302,31 +312,31 @@ fz_pop_clip(fz_context *ctx, fz_device *dev)
 }
 
 void
-fz_fill_shade(fz_context *ctx, fz_device *dev, fz_shade *shade, const fz_matrix *ctm, float alpha)
+fz_fill_shade(fz_context *ctx, fz_device *dev, fz_shade *shade, const fz_matrix *ctm, float alpha, const fz_color_params *color_params)
 {
 	if (dev->error_depth)
 		return;
 	if (dev->fill_shade)
-		dev->fill_shade(ctx, dev, shade, ctm, alpha);
+		dev->fill_shade(ctx, dev, shade, ctm, alpha, color_params);
 }
 
 void
-fz_fill_image(fz_context *ctx, fz_device *dev, fz_image *image, const fz_matrix *ctm, float alpha)
+fz_fill_image(fz_context *ctx, fz_device *dev, fz_image *image, const fz_matrix *ctm, float alpha, const fz_color_params *color_params)
 {
 	if (dev->error_depth)
 		return;
 	if (dev->fill_image)
-		dev->fill_image(ctx, dev, image, ctm, alpha);
+		dev->fill_image(ctx, dev, image, ctm, alpha, color_params);
 }
 
 void
 fz_fill_image_mask(fz_context *ctx, fz_device *dev, fz_image *image, const fz_matrix *ctm,
-	fz_colorspace *colorspace, const float *color, float alpha)
+	fz_colorspace *colorspace, const float *color, float alpha, const fz_color_params *color_params)
 {
 	if (dev->error_depth)
 		return;
 	if (dev->fill_image_mask)
-		dev->fill_image_mask(ctx, dev, image, ctm, colorspace, color, alpha);
+		dev->fill_image_mask(ctx, dev, image, ctm, colorspace, color, alpha, color_params);
 }
 
 void
@@ -363,7 +373,7 @@ fz_clip_image_mask(fz_context *ctx, fz_device *dev, fz_image *image, const fz_ma
 }
 
 void
-fz_begin_mask(fz_context *ctx, fz_device *dev, const fz_rect *area, int luminosity, fz_colorspace *colorspace, const float *bc)
+fz_begin_mask(fz_context *ctx, fz_device *dev, const fz_rect *area, int luminosity, fz_colorspace *colorspace, const float *bc, const fz_color_params *color_params)
 {
 	if (dev->error_depth)
 	{
@@ -376,7 +386,7 @@ fz_begin_mask(fz_context *ctx, fz_device *dev, const fz_rect *area, int luminosi
 		if (dev->hints & FZ_MAINTAIN_CONTAINER_STACK)
 			push_clip_stack(ctx, dev, area, fz_device_container_stack_in_mask);
 		if (dev->begin_mask)
-			dev->begin_mask(ctx, dev, area, luminosity, colorspace, bc);
+			dev->begin_mask(ctx, dev, area, luminosity, colorspace, bc, color_params);
 	}
 	fz_catch(ctx)
 	{
@@ -413,7 +423,7 @@ fz_end_mask(fz_context *ctx, fz_device *dev)
 }
 
 void
-fz_begin_group(fz_context *ctx, fz_device *dev, const fz_rect *area, int isolated, int knockout, int blendmode, float alpha)
+fz_begin_group(fz_context *ctx, fz_device *dev, const fz_rect *area, fz_colorspace *cs, int isolated, int knockout, int blendmode, float alpha)
 {
 	if (dev->error_depth)
 	{
@@ -426,7 +436,7 @@ fz_begin_group(fz_context *ctx, fz_device *dev, const fz_rect *area, int isolate
 		if (dev->hints & FZ_MAINTAIN_CONTAINER_STACK)
 			push_clip_stack(ctx, dev, area, fz_device_container_stack_is_group);
 		if (dev->begin_group)
-			dev->begin_group(ctx, dev, area, isolated, knockout, blendmode, alpha);
+			dev->begin_group(ctx, dev, area, cs, isolated, knockout, blendmode, alpha);
 	}
 	fz_catch(ctx)
 	{
@@ -510,3 +520,30 @@ fz_render_flags(fz_context *ctx, fz_device *dev, int set, int clear)
 	if (dev->render_flags)
 		dev->render_flags(ctx, dev, set, clear);
 }
+
+void
+fz_set_default_colorspaces(fz_context *ctx, fz_device *dev, fz_default_colorspaces *default_cs)
+{
+	if (dev->set_default_colorspaces)
+		dev->set_default_colorspaces(ctx, dev, default_cs);
+}
+
+void fz_begin_layer(fz_context *ctx, fz_device *dev, const char *layer_name)
+{
+	if (dev->begin_layer)
+		dev->begin_layer(ctx, dev, layer_name);
+}
+
+void fz_end_layer(fz_context *ctx, fz_device *dev)
+{
+	if (dev->end_layer)
+		dev->end_layer(ctx, dev);
+}
+
+const fz_rect *
+fz_device_current_scissor(fz_context *ctx, fz_device *dev)
+{
+	if (dev->container_len > 0)
+		return &dev->container[dev->container_len-1].scissor;
+	return &fz_infinite_rect;
+}
