@@ -1,5 +1,8 @@
+#include "mupdf/fitz.h"
 #include "fitz-imp.h"
 
+#include <string.h>
+
 typedef struct tar_entry_s tar_entry;
 typedef struct fz_tar_archive_s fz_tar_archive;
 
@@ -63,7 +66,7 @@ static void ensure_tar_entries(fz_context *ctx, fz_tar_archive *tar)
 		offset = fz_tell(ctx, file);
 		n = fz_read(ctx, file, (unsigned char *) name, nelem(name));
 		if (n < nelem(name))
-			fz_throw(ctx, FZ_ERROR_GENERIC, "premature end of data in zip entry name");
+			fz_throw(ctx, FZ_ERROR_GENERIC, "premature end of data in tar entry name");
 		name[nelem(name) - 1] = '\0';
 
 		if (strlen(name) == 0)
@@ -72,7 +75,7 @@ static void ensure_tar_entries(fz_context *ctx, fz_tar_archive *tar)
 		fz_seek(ctx, file, 24, 1);
 		n = fz_read(ctx, file, (unsigned char *) octsize, nelem(octsize));
 		if (n < nelem(octsize))
-			fz_throw(ctx, FZ_ERROR_GENERIC, "premature end of data in zip entry size");
+			fz_throw(ctx, FZ_ERROR_GENERIC, "premature end of data in tar entry size");
 		size = otoi(octsize);
 
 		fz_seek(ctx, file, 20, 1);
