@@ -1,11 +1,13 @@
 #include "mupdf/fitz.h"
 
+#include <float.h>
+
 fz_display_list *
 fz_new_display_list_from_page(fz_context *ctx, fz_page *page)
 {
 	fz_display_list *list;
 	fz_rect bounds;
-	fz_device *dev;
+	fz_device *dev = NULL;
 
 	list = fz_new_display_list(ctx, fz_bound_page(ctx, page, &bounds));
 
@@ -32,7 +34,7 @@ fz_display_list *
 fz_new_display_list_from_page_number(fz_context *ctx, fz_document *doc, int number)
 {
 	fz_page *page;
-	fz_display_list *list;
+	fz_display_list *list = NULL;
 
 	page = fz_load_page(ctx, doc, number);
 	fz_try(ctx)
@@ -49,7 +51,7 @@ fz_new_display_list_from_page_contents(fz_context *ctx, fz_page *page)
 {
 	fz_display_list *list;
 	fz_rect bounds;
-	fz_device *dev;
+	fz_device *dev = NULL;
 
 	list = fz_new_display_list(ctx, fz_bound_page(ctx, page, &bounds));
 
@@ -77,7 +79,7 @@ fz_new_display_list_from_annot(fz_context *ctx, fz_annot *annot)
 {
 	fz_display_list *list;
 	fz_rect bounds;
-	fz_device *dev;
+	fz_device *dev = NULL;
 
 	list = fz_new_display_list(ctx, fz_bound_annot(ctx, annot, &bounds));
 
@@ -106,13 +108,13 @@ fz_new_pixmap_from_display_list(fz_context *ctx, fz_display_list *list, const fz
 	fz_rect rect;
 	fz_irect irect;
 	fz_pixmap *pix;
-	fz_device *dev;
+	fz_device *dev = NULL;
 
 	fz_bound_display_list(ctx, list, &rect);
 	fz_transform_rect(&rect, ctm);
 	fz_round_rect(&irect, &rect);
 
-	pix = fz_new_pixmap_with_bbox(ctx, cs, &irect, alpha);
+	pix = fz_new_pixmap_with_bbox(ctx, cs, &irect, 0, alpha);
 	if (alpha)
 		fz_clear_pixmap(ctx, pix);
 	else
@@ -143,13 +145,13 @@ fz_new_pixmap_from_page_contents(fz_context *ctx, fz_page *page, const fz_matrix
 	fz_rect rect;
 	fz_irect irect;
 	fz_pixmap *pix;
-	fz_device *dev;
+	fz_device *dev = NULL;
 
 	fz_bound_page(ctx, page, &rect);
 	fz_transform_rect(&rect, ctm);
 	fz_round_rect(&irect, &rect);
 
-	pix = fz_new_pixmap_with_bbox(ctx, cs, &irect, alpha);
+	pix = fz_new_pixmap_with_bbox(ctx, cs, &irect, 0, alpha);
 	if (alpha)
 		fz_clear_pixmap(ctx, pix);
 	else
@@ -180,13 +182,13 @@ fz_new_pixmap_from_annot(fz_context *ctx, fz_annot *annot, const fz_matrix *ctm,
 	fz_rect rect;
 	fz_irect irect;
 	fz_pixmap *pix;
-	fz_device *dev;
+	fz_device *dev = NULL;
 
 	fz_bound_annot(ctx, annot, &rect);
 	fz_transform_rect(&rect, ctm);
 	fz_round_rect(&irect, &rect);
 
-	pix = fz_new_pixmap_with_bbox(ctx, cs, &irect, alpha);
+	pix = fz_new_pixmap_with_bbox(ctx, cs, &irect, 0, alpha);
 	if (alpha)
 		fz_clear_pixmap(ctx, pix);
 	else
@@ -217,13 +219,13 @@ fz_new_pixmap_from_page(fz_context *ctx, fz_page *page, const fz_matrix *ctm, fz
 	fz_rect rect;
 	fz_irect irect;
 	fz_pixmap *pix;
-	fz_device *dev;
+	fz_device *dev = NULL;
 
 	fz_bound_page(ctx, page, &rect);
 	fz_transform_rect(&rect, ctm);
 	fz_round_rect(&irect, &rect);
 
-	pix = fz_new_pixmap_with_bbox(ctx, cs, &irect, alpha);
+	pix = fz_new_pixmap_with_bbox(ctx, cs, &irect, 0, alpha);
 	if (alpha)
 		fz_clear_pixmap(ctx, pix);
 	else
@@ -252,7 +254,7 @@ fz_pixmap *
 fz_new_pixmap_from_page_number(fz_context *ctx, fz_document *doc, int number, const fz_matrix *ctm, fz_colorspace *cs, int alpha)
 {
 	fz_page *page;
-	fz_pixmap *pix;
+	fz_pixmap *pix = NULL;
 
 	page = fz_load_page(ctx, doc, number);
 	fz_try(ctx)
@@ -265,10 +267,10 @@ fz_new_pixmap_from_page_number(fz_context *ctx, fz_document *doc, int number, co
 }
 
 fz_stext_page *
-fz_new_stext_page_from_display_list(fz_context *ctx, fz_display_list *list, fz_stext_sheet *sheet, const fz_stext_options *options)
+fz_new_stext_page_from_display_list(fz_context *ctx, fz_display_list *list, const fz_stext_options *options)
 {
 	fz_stext_page *text;
-	fz_device *dev;
+	fz_device *dev = NULL;
 	fz_rect mediabox;
 
 	if (list == NULL)
@@ -277,7 +279,7 @@ fz_new_stext_page_from_display_list(fz_context *ctx, fz_display_list *list, fz_s
 	text = fz_new_stext_page(ctx, fz_bound_display_list(ctx, list, &mediabox));
 	fz_try(ctx)
 	{
-		dev = fz_new_stext_device(ctx, sheet, text, options);
+		dev = fz_new_stext_device(ctx, text, options);
 		fz_run_display_list(ctx, list, dev, &fz_identity, NULL, NULL);
 		fz_close_device(ctx, dev);
 	}
@@ -295,10 +297,10 @@ fz_new_stext_page_from_display_list(fz_context *ctx, fz_display_list *list, fz_s
 }
 
 fz_stext_page *
-fz_new_stext_page_from_page(fz_context *ctx, fz_page *page, fz_stext_sheet *sheet, const fz_stext_options *options)
+fz_new_stext_page_from_page(fz_context *ctx, fz_page *page, const fz_stext_options *options)
 {
 	fz_stext_page *text;
-	fz_device *dev;
+	fz_device *dev = NULL;
 	fz_rect mediabox;
 
 	if (page == NULL)
@@ -307,8 +309,8 @@ fz_new_stext_page_from_page(fz_context *ctx, fz_page *page, fz_stext_sheet *shee
 	text = fz_new_stext_page(ctx, fz_bound_page(ctx, page, &mediabox));
 	fz_try(ctx)
 	{
-		dev = fz_new_stext_device(ctx, sheet, text, options);
-		fz_run_page(ctx, page, dev, &fz_identity, NULL);
+		dev = fz_new_stext_device(ctx, text, options);
+		fz_run_page_contents(ctx, page, dev, &fz_identity, NULL);
 		fz_close_device(ctx, dev);
 	}
 	fz_always(ctx)
@@ -325,14 +327,14 @@ fz_new_stext_page_from_page(fz_context *ctx, fz_page *page, fz_stext_sheet *shee
 }
 
 fz_stext_page *
-fz_new_stext_page_from_page_number(fz_context *ctx, fz_document *doc, int number, fz_stext_sheet *sheet, const fz_stext_options *options)
+fz_new_stext_page_from_page_number(fz_context *ctx, fz_document *doc, int number, const fz_stext_options *options)
 {
 	fz_page *page;
-	fz_stext_page *text;
+	fz_stext_page *text = NULL;
 
 	page = fz_load_page(ctx, doc, number);
 	fz_try(ctx)
-		text = fz_new_stext_page_from_page(ctx, page, sheet, options);
+		text = fz_new_stext_page_from_page(ctx, page, options);
 	fz_always(ctx)
 		fz_drop_page(ctx, page);
 	fz_catch(ctx)
@@ -343,24 +345,14 @@ fz_new_stext_page_from_page_number(fz_context *ctx, fz_document *doc, int number
 int
 fz_search_display_list(fz_context *ctx, fz_display_list *list, const char *needle, fz_rect *hit_bbox, int hit_max)
 {
-	fz_stext_sheet *sheet = NULL;
-	fz_stext_page *text = NULL;
-	int count;
-
-	fz_var(sheet);
-	fz_var(text);
+	fz_stext_page *text;
+	int count = 0;
 
+	text = fz_new_stext_page_from_display_list(ctx, list, NULL);
 	fz_try(ctx)
-	{
-		sheet = fz_new_stext_sheet(ctx);
-		text = fz_new_stext_page_from_display_list(ctx, list, sheet, NULL);
 		count = fz_search_stext_page(ctx, text, needle, hit_bbox, hit_max);
-	}
 	fz_always(ctx)
-	{
 		fz_drop_stext_page(ctx, text);
-		fz_drop_stext_sheet(ctx, sheet);
-	}
 	fz_catch(ctx)
 		fz_rethrow(ctx);
 	return count;
@@ -369,24 +361,14 @@ fz_search_display_list(fz_context *ctx, fz_display_list *list, const char *needl
 int
 fz_search_page(fz_context *ctx, fz_page *page, const char *needle, fz_rect *hit_bbox, int hit_max)
 {
-	fz_stext_sheet *sheet = NULL;
-	fz_stext_page *text = NULL;
-	int count;
-
-	fz_var(sheet);
-	fz_var(text);
+	fz_stext_page *text;
+	int count = 0;
 
+	text = fz_new_stext_page_from_page(ctx, page, NULL);
 	fz_try(ctx)
-	{
-		sheet = fz_new_stext_sheet(ctx);
-		text = fz_new_stext_page_from_page(ctx, page, sheet, NULL);
 		count = fz_search_stext_page(ctx, text, needle, hit_bbox, hit_max);
-	}
 	fz_always(ctx)
-	{
 		fz_drop_stext_page(ctx, text);
-		fz_drop_stext_sheet(ctx, sheet);
-	}
 	fz_catch(ctx)
 		fz_rethrow(ctx);
 	return count;
@@ -396,7 +378,7 @@ int
 fz_search_page_number(fz_context *ctx, fz_document *doc, int number, const char *needle, fz_rect *hit_bbox, int hit_max)
 {
 	fz_page *page;
-	int count;
+	int count = 0;
 
 	page = fz_load_page(ctx, doc, number);
 	fz_try(ctx)
@@ -409,72 +391,27 @@ fz_search_page_number(fz_context *ctx, fz_document *doc, int number, const char
 }
 
 fz_buffer *
-fz_new_buffer_from_stext_page(fz_context *ctx, fz_stext_page *text, const fz_rect *sel, int crlf)
+fz_new_buffer_from_stext_page(fz_context *ctx, fz_stext_page *page)
 {
+	fz_stext_block *block;
+	fz_stext_line *line;
+	fz_stext_char *ch;
 	fz_buffer *buf;
-	fz_rect hitbox;
-	float x0, y0, x1, y1;
-	int block_num;
-	int need_newline;
-	int i;
-
-	need_newline = 0;
-
-	if (fz_is_infinite_rect(sel))
-	{
-		x0 = y0 = -FLT_MAX;
-		x1 = y1 = FLT_MAX;
-	}
-	else
-	{
-		x0 = sel->x0;
-		y0 = sel->y0;
-		x1 = sel->x1;
-		y1 = sel->y1;
-	}
 
 	buf = fz_new_buffer(ctx, 256);
 	fz_try(ctx)
 	{
-		for (block_num = 0; block_num < text->len; block_num++)
+		for (block = page->first_block; block; block = block->next)
 		{
-			fz_stext_line *line;
-			fz_stext_block *block;
-			fz_stext_span *span;
-
-			if (text->blocks[block_num].type != FZ_PAGE_BLOCK_TEXT)
-				continue;
-
-			block = text->blocks[block_num].u.text;
-			for (line = block->lines; line < block->lines + block->len; line++)
+			if (block->type == FZ_STEXT_BLOCK_TEXT)
 			{
-				int saw_text = 0;
-				for (span = line->first_span; span; span = span->next)
+				for (line = block->u.t.first_line; line; line = line->next)
 				{
-					for (i = 0; i < span->len; i++)
-					{
-						int c;
-						fz_stext_char_bbox(ctx, &hitbox, span, i);
-						c = span->text[i].c;
-						if (c < 32)
-							c = 0xFFFD;
-						if (hitbox.x1 >= x0 && hitbox.x0 <= x1 && hitbox.y1 >= y0 && hitbox.y0 <= y1)
-						{
-							saw_text = 1;
-							if (need_newline)
-							{
-								if (crlf)
-									fz_append_byte(ctx, buf, '\r');
-								fz_append_byte(ctx, buf, '\n');
-								need_newline = 0;
-							}
-							fz_append_rune(ctx, buf, c);
-						}
-					}
+					for (ch = line->first_char; ch; ch = ch->next)
+						fz_append_rune(ctx, buf, ch->c);
+					fz_append_byte(ctx, buf, '\n');
 				}
-
-				if (saw_text)
-					need_newline = 1;
+				fz_append_byte(ctx, buf, '\n');
 			}
 		}
 	}
@@ -488,59 +425,86 @@ fz_new_buffer_from_stext_page(fz_context *ctx, fz_stext_page *text, const fz_rec
 }
 
 fz_buffer *
-fz_new_buffer_from_display_list(fz_context *ctx, fz_display_list *list, const fz_rect *sel, int crlf, const fz_stext_options *options)
+fz_new_buffer_from_display_list(fz_context *ctx, fz_display_list *list, const fz_stext_options *options)
 {
-	fz_stext_sheet *sheet;
 	fz_stext_page *text;
-	fz_buffer *buf;
+	fz_buffer *buf = NULL;
 
-	sheet = fz_new_stext_sheet(ctx);
+	text = fz_new_stext_page_from_display_list(ctx, list, options);
 	fz_try(ctx)
-	{
-		text = fz_new_stext_page_from_display_list(ctx, list, sheet, options);
-		buf = fz_new_buffer_from_stext_page(ctx, text, sel, crlf);
-	}
+		buf = fz_new_buffer_from_stext_page(ctx, text);
 	fz_always(ctx)
-		fz_drop_stext_sheet(ctx, sheet);
+		fz_drop_stext_page(ctx, text);
 	fz_catch(ctx)
 		fz_rethrow(ctx);
-	fz_drop_stext_page(ctx, text);
 	return buf;
 }
 
 fz_buffer *
-fz_new_buffer_from_page(fz_context *ctx, fz_page *page, const fz_rect *sel, int crlf, const fz_stext_options *options)
+fz_new_buffer_from_page(fz_context *ctx, fz_page *page, const fz_stext_options *options)
 {
-	fz_stext_sheet *sheet;
 	fz_stext_page *text;
-	fz_buffer *buf;
+	fz_buffer *buf = NULL;
 
-	sheet = fz_new_stext_sheet(ctx);
+	text = fz_new_stext_page_from_page(ctx, page, options);
 	fz_try(ctx)
-	{
-		text = fz_new_stext_page_from_page(ctx, page, sheet, options);
-		buf = fz_new_buffer_from_stext_page(ctx, text, sel, crlf);
-	}
+		buf = fz_new_buffer_from_stext_page(ctx, text);
 	fz_always(ctx)
-		fz_drop_stext_sheet(ctx, sheet);
+		fz_drop_stext_page(ctx, text);
 	fz_catch(ctx)
 		fz_rethrow(ctx);
-	fz_drop_stext_page(ctx, text);
 	return buf;
 }
 
 fz_buffer *
-fz_new_buffer_from_page_number(fz_context *ctx, fz_document *doc, int number, const fz_rect *sel, int crlf, const fz_stext_options *options)
+fz_new_buffer_from_page_number(fz_context *ctx, fz_document *doc, int number, const fz_stext_options *options)
 {
 	fz_page *page;
-	fz_buffer *buf;
+	fz_buffer *buf = NULL;
 
 	page = fz_load_page(ctx, doc, number);
 	fz_try(ctx)
-		buf = fz_new_buffer_from_page(ctx, page, sel, crlf, options);
+		buf = fz_new_buffer_from_page(ctx, page, options);
 	fz_always(ctx)
 		fz_drop_page(ctx, page);
 	fz_catch(ctx)
 		fz_rethrow(ctx);
 	return buf;
 }
+
+void
+fz_write_image_as_data_uri(fz_context *ctx, fz_output *out, fz_image *image)
+{
+	fz_compressed_buffer *cbuf;
+	fz_buffer *buf;
+
+	cbuf = fz_compressed_image_buffer(ctx, image);
+
+	if (cbuf && cbuf->params.type == FZ_IMAGE_JPEG)
+	{
+		int type = fz_colorspace_type(ctx, image->colorspace);
+		if (type == FZ_COLORSPACE_GRAY || type == FZ_COLORSPACE_RGB)
+		{
+			fz_write_string(ctx, out, "image/jpeg;base64,");
+			fz_write_base64_buffer(ctx, out, cbuf->buffer, 1);
+			return;
+		}
+	}
+	if (cbuf && cbuf->params.type == FZ_IMAGE_PNG)
+	{
+		fz_write_string(ctx, out, "image/png;base64,");
+		fz_write_base64_buffer(ctx, out, cbuf->buffer, 1);
+		return;
+	}
+
+	buf = fz_new_buffer_from_image_as_png(ctx, image, NULL);
+	fz_try(ctx)
+	{
+		fz_write_string(ctx, out, "image/png;base64,");
+		fz_write_base64_buffer(ctx, out, buf, 1);
+	}
+	fz_always(ctx)
+		fz_drop_buffer(ctx, buf);
+	fz_catch(ctx)
+		fz_rethrow(ctx);
+}
