@@ -1,10 +1,13 @@
+#include "mupdf/fitz.h"
 #include "mupdf/pdf.h"
 
+#include <string.h>
+
 #define IS_NUMBER \
 	'+':case'-':case'.':case'0':case'1':case'2':case'3':\
 	case'4':case'5':case'6':case'7':case'8':case'9'
 #define IS_WHITE \
-	'\000':case'\011':case'\012':case'\014':case'\015':case'\040'
+	'\x00':case'\x09':case'\x0a':case'\x0c':case'\x0d':case'\x20'
 #define IS_HEX \
 	'0':case'1':case'2':case'3':case'4':case'5':case'6':\
 	case'7':case'8':case'9':case'A':case'B':case'C':\
@@ -35,6 +38,11 @@ static inline int iswhite(int ch)
 		ch == '\040';
 }
 
+static inline int fz_isprint(int ch)
+{
+	return ch >= ' ' && ch <= '~';
+}
+
 static inline int unhex(int ch)
 {
 	if (ch >= '0' && ch <= '9') return ch - '0';
@@ -63,115 +71,167 @@ lex_comment(fz_context *ctx, fz_stream *f)
 	} while ((c != '\012') && (c != '\015') && (c != EOF));
 }
 
+/* Fast(ish) but inaccurate strtof, with Adobe overflow handling. */
+static float acrobat_compatible_atof(char *s)
+{
+	int neg = 0;
+	int i = 0;
+
+	while (*s == '-')
+	{
+		neg = 1;
+		++s;
+	}
+	while (*s == '+')
+	{
+		++s;
+	}
+
+	while (*s >= '0' && *s <= '9')
+	{
+		/* We deliberately ignore overflow here.
+		 * Tests show that Acrobat handles * overflows in exactly the same way we do:
+		 * 123450000000000000000678 is read as 678.
+		 */
+		i = i * 10 + (*s - '0');
+		++s;
+	}
+
+	if (*s == '.')
+	{
+		float v = i;
+		float n = 0;
+		float d = 1;
+		++s;
+		while (*s >= '0' && *s <= '9')
+		{
+			n = 10 * n + (*s - '0');
+			d = 10 * d;
+			++s;
+		}
+		v += n / d;
+		return neg ? -v : v;
+	}
+	else
+	{
+		return neg ? -i : i;
+	}
+}
+
+/* Fast but inaccurate atoi. */
+static int fast_atoi(char *s)
+{
+	int neg = 0;
+	int i = 0;
+
+	while (*s == '-')
+	{
+		neg = 1;
+		++s;
+	}
+	while (*s == '+')
+	{
+		++s;
+	}
+
+	while (*s >= '0' && *s <= '9')
+	{
+		/* We deliberately ignore overflow here. */
+		i = i * 10 + (*s - '0');
+		++s;
+	}
+
+	return neg ? -i : i;
+}
+
 static int
 lex_number(fz_context *ctx, fz_stream *f, pdf_lexbuf *buf, int c)
 {
-	int neg = 0;
-	fz_off_t i = 0;
-	int n;
-	int d;
-	float v;
+	char *s = buf->scratch;
+	char *e = buf->scratch + buf->size - 1; /* leave space for zero terminator */
+	char *isreal = (c == '.' ? s : NULL);
+	int neg = (c == '-');
+	int isbad = 0;
 
-	/* Initially we might have +, -, . or a digit */
-	switch (c)
+	*s++ = c;
+
+	c = fz_read_byte(ctx, f);
+
+	/* skip extra '-' signs at start of number */
+	if (neg)
 	{
-	case '.':
-		goto loop_after_dot;
-	case '-':
-		neg = 1;
-		break;
-	case '+':
-		break;
-	default: /* Must be a digit */
-		i = c - '0';
-		break;
+		while (c == '-')
+			c = fz_read_byte(ctx, f);
 	}
 
-	while (1)
+	while (s < e)
 	{
-		c = fz_read_byte(ctx, f);
 		switch (c)
 		{
+		case IS_WHITE:
+		case IS_DELIM:
+			fz_unread_byte(ctx, f);
+			goto end;
+		case EOF:
+			goto end;
 		case '.':
-			goto loop_after_dot;
+			if (isreal)
+				isbad = 1;
+			isreal = s;
+			*s++ = c;
+			break;
 		case RANGE_0_9:
-			/* We deliberately ignore overflow here. We tried
-			 * code that returned INT_MIN/MAX as appropriate,
-			 * but this causes loss of data (see Bug695950.pdf
-			 * for an example). Tests show that Acrobat handles
-			 * overflows in exactly the same way we do (i.e.
-			 * 123450000000000000000678 is read as 678). */
-			i = 10*i + c - '0';
+			*s++ = c;
 			break;
 		default:
-			fz_unread_byte(ctx, f);
-			/* Fallthrough */
-		case EOF:
-			if (neg)
-				i = -i;
-			buf->i = i;
-			return PDF_TOK_INT;
+			isbad = 1;
+			*s++ = c;
+			break;
 		}
+		c = fz_read_byte(ctx, f);
 	}
 
-	/* In here, we've seen a dot, so can accept just digits */
-loop_after_dot:
-	n = 0;
-	d = 1;
-	while (1)
+end:
+	*s = '\0';
+	if (isbad)
+		return PDF_TOK_ERROR;
+	if (isreal)
 	{
-		c = fz_read_byte(ctx, f);
-		switch (c)
-		{
-		case RANGE_0_9:
-			if (d >= INT_MAX/10)
-				goto underflow;
-			n = n*10 + (c - '0');
-			d *= 10;
-			break;
-		default:
-			fz_unread_byte(ctx, f);
-			/* Fallthrough */
-		case EOF:
-			v = (float)i + ((float)n / (float)d);
-			if (neg)
-				v = -v;
-			buf->f = v;
-			return PDF_TOK_REAL;
-		}
+		/* We'd like to use the fastest possible atof
+		 * routine, but we'd rather match acrobats
+		 * handling of broken numbers. As such, we
+		 * spot common broken cases and call an
+		 * acrobat compatible routine where required. */
+		if (neg > 1 || isreal - buf->scratch >= 10)
+			buf->f = acrobat_compatible_atof(buf->scratch);
+		else
+			buf->f = fz_atof(buf->scratch);
+		return PDF_TOK_REAL;
 	}
-
-underflow:
-	/* Ignore any digits after here, because they are too small */
-	while (1)
+	else
 	{
-		c = fz_read_byte(ctx, f);
-		switch (c)
-		{
-		case RANGE_0_9:
-			break;
-		default:
-			fz_unread_byte(ctx, f);
-			/* Fallthrough */
-		case EOF:
-			v = (float)i + ((float)n / (float)d);
-			if (neg)
-				v = -v;
-			buf->f = v;
-			return PDF_TOK_REAL;
-		}
+		buf->i = fast_atoi(buf->scratch);
+		return PDF_TOK_INT;
 	}
 }
 
 static void
-lex_name(fz_context *ctx, fz_stream *f, pdf_lexbuf *buf)
+lex_name(fz_context *ctx, fz_stream *f, pdf_lexbuf *lb)
 {
-	char *s = buf->scratch;
-	int n = buf->size;
+	char *s = lb->scratch;
+	char *e = s + fz_mini(127, lb->size);
+	int c;
 
-	while (n > 1)
+	while (1)
 	{
-		int c = fz_read_byte(ctx, f);
+		if (s == e)
+		{
+			if (e - lb->scratch >= 127)
+				fz_throw(ctx, FZ_ERROR_SYNTAX, "name too long");
+			s += pdf_lexbuf_grow(ctx, lb);
+			e = lb->scratch + fz_mini(127, lb->size);
+		}
+		c = fz_read_byte(ctx, f);
 		switch (c)
 		{
 		case IS_WHITE:
@@ -182,58 +242,45 @@ lex_name(fz_context *ctx, fz_stream *f, pdf_lexbuf *buf)
 			goto end;
 		case '#':
 		{
-			int d;
-			c = fz_read_byte(ctx, f);
-			switch (c)
+			int hex[2];
+			int i;
+			for (i = 0; i < 2; i++)
 			{
-			case RANGE_0_9:
-				d = (c - '0') << 4;
-				break;
-			case RANGE_a_f:
-				d = (c - 'a' + 10) << 4;
-				break;
-			case RANGE_A_F:
-				d = (c - 'A' + 10) << 4;
-				break;
-			default:
-				fz_unread_byte(ctx, f);
-				/* fallthrough */
-			case EOF:
-				goto end;
+				c = fz_peek_byte(ctx, f);
+				switch (c)
+				{
+				case RANGE_0_9:
+					if (i == 1 && c == '0' && hex[0] == 0)
+						goto illegal;
+					hex[i] = fz_read_byte(ctx, f) - '0';
+					break;
+				case RANGE_a_f:
+					hex[i] = fz_read_byte(ctx, f) - 'a' + 10;
+					break;
+				case RANGE_A_F:
+					hex[i] = fz_read_byte(ctx, f) - 'A' + 10;
+					break;
+				default:
+				case EOF:
+					goto illegal;
+				}
 			}
-			c = fz_read_byte(ctx, f);
-			switch (c)
-			{
-			case RANGE_0_9:
-				c -= '0';
-				break;
-			case RANGE_a_f:
-				c -= 'a' - 10;
-				break;
-			case RANGE_A_F:
-				c -= 'A' - 10;
-				break;
-			default:
-				fz_unread_byte(ctx, f);
-				/* fallthrough */
-			case EOF:
-				*s++ = d;
-				n--;
-				goto end;
-			}
-			*s++ = d + c;
-			n--;
+			*s++ = (hex[0] << 4) + hex[1];
 			break;
+illegal:
+			if (i == 1)
+				fz_unread_byte(ctx, f);
+			*s++ = '#';
+			continue;
 		}
 		default:
 			*s++ = c;
-			n--;
 			break;
 		}
 	}
 end:
 	*s = '\0';
-	buf->len = s - buf->scratch;
+	lb->len = s - lb->scratch;
 }
 
 static int
@@ -256,7 +303,7 @@ lex_string(fz_context *ctx, fz_stream *f, pdf_lexbuf *lb)
 		switch (c)
 		{
 		case EOF:
-			goto end;
+			return PDF_TOK_ERROR;
 		case '(':
 			bal++;
 			*s++ = c;
@@ -272,7 +319,7 @@ lex_string(fz_context *ctx, fz_stream *f, pdf_lexbuf *lb)
 			switch (c)
 			{
 			case EOF:
-				goto end;
+				return PDF_TOK_ERROR;
 			case 'n':
 				*s++ = '\n';
 				break;
@@ -367,8 +414,9 @@ lex_hex_string(fz_context *ctx, fz_stream *f, pdf_lexbuf *lb)
 			}
 			break;
 		case '>':
-		case EOF:
 			goto end;
+		case EOF:
+			return PDF_TOK_ERROR;
 		default:
 			fz_warn(ctx, "ignoring invalid character in hex string");
 		}
@@ -410,8 +458,13 @@ pdf_token_from_keyword(char *key)
 	case 'x':
 		if (!strcmp(key, "xref")) return PDF_TOK_XREF;
 		break;
-	default:
-		break;
+	}
+
+	while (*key)
+	{
+		if (!fz_isprint(*key))
+			return PDF_TOK_ERROR;
+		++key;
 	}
 
 	return PDF_TOK_KEYWORD;
@@ -469,32 +522,21 @@ pdf_lex(fz_context *ctx, fz_stream *f, pdf_lexbuf *buf)
 		case '(':
 			return lex_string(ctx, f, buf);
 		case ')':
-			fz_warn(ctx, "lexical error (unexpected ')')");
-			continue;
+			return PDF_TOK_ERROR;
 		case '<':
 			c = fz_read_byte(ctx, f);
 			if (c == '<')
-			{
 				return PDF_TOK_OPEN_DICT;
-			}
-			else
-			{
+			if (c != EOF)
 				fz_unread_byte(ctx, f);
-				return lex_hex_string(ctx, f, buf);
-			}
+			return lex_hex_string(ctx, f, buf);
 		case '>':
 			c = fz_read_byte(ctx, f);
 			if (c == '>')
-			{
 				return PDF_TOK_CLOSE_DICT;
-			}
-			fz_warn(ctx, "lexical error (unexpected '>')");
-			if (c == EOF)
-			{
-				return PDF_TOK_EOF;
-			}
-			fz_unread_byte(ctx, f);
-			continue;
+			if (c != EOF)
+				fz_unread_byte(ctx, f);
+			return PDF_TOK_ERROR;
 		case '[':
 			return PDF_TOK_OPEN_ARRAY;
 		case ']':
@@ -533,31 +575,23 @@ pdf_lex_no_string(fz_context *ctx, fz_stream *f, pdf_lexbuf *buf)
 			lex_name(ctx, f, buf);
 			return PDF_TOK_NAME;
 		case '(':
-			continue;
+			return PDF_TOK_ERROR; /* no strings allowed */
 		case ')':
-			continue;
+			return PDF_TOK_ERROR; /* no strings allowed */
 		case '<':
 			c = fz_read_byte(ctx, f);
 			if (c == '<')
-			{
 				return PDF_TOK_OPEN_DICT;
-			}
-			else
-			{
-				continue;
-			}
+			if (c != EOF)
+				fz_unread_byte(ctx, f);
+			return PDF_TOK_ERROR; /* no strings allowed */
 		case '>':
 			c = fz_read_byte(ctx, f);
 			if (c == '>')
-			{
 				return PDF_TOK_CLOSE_DICT;
-			}
-			if (c == EOF)
-			{
-				return PDF_TOK_EOF;
-			}
-			fz_unread_byte(ctx, f);
-			continue;
+			if (c != EOF)
+				fz_unread_byte(ctx, f);
+			return PDF_TOK_ERROR;
 		case '[':
 			return PDF_TOK_OPEN_ARRAY;
 		case ']':
@@ -576,47 +610,45 @@ pdf_lex_no_string(fz_context *ctx, fz_stream *f, pdf_lexbuf *buf)
 	}
 }
 
-void pdf_print_token(fz_context *ctx, fz_buffer *fzbuf, int tok, pdf_lexbuf *buf)
+void pdf_append_token(fz_context *ctx, fz_buffer *fzbuf, int tok, pdf_lexbuf *buf)
 {
 	switch (tok)
 	{
 	case PDF_TOK_NAME:
-		fz_buffer_printf(ctx, fzbuf, "/%s", buf->scratch);
+		fz_append_printf(ctx, fzbuf, "/%s", buf->scratch);
 		break;
 	case PDF_TOK_STRING:
 		if (buf->len >= buf->size)
 			pdf_lexbuf_grow(ctx, buf);
 		buf->scratch[buf->len] = 0;
-		fz_buffer_cat_pdf_string(ctx, fzbuf, buf->scratch);
+		fz_append_pdf_string(ctx, fzbuf, buf->scratch);
 		break;
 	case PDF_TOK_OPEN_DICT:
-		fz_buffer_printf(ctx, fzbuf, "<<");
+		fz_append_string(ctx, fzbuf, "<<");
 		break;
 	case PDF_TOK_CLOSE_DICT:
-		fz_buffer_printf(ctx, fzbuf, ">>");
+		fz_append_string(ctx, fzbuf, ">>");
 		break;
 	case PDF_TOK_OPEN_ARRAY:
-		fz_buffer_printf(ctx, fzbuf, "[");
+		fz_append_byte(ctx, fzbuf, '[');
 		break;
 	case PDF_TOK_CLOSE_ARRAY:
-		fz_buffer_printf(ctx, fzbuf, "]");
+		fz_append_byte(ctx, fzbuf, ']');
 		break;
 	case PDF_TOK_OPEN_BRACE:
-		fz_buffer_printf(ctx, fzbuf, "{");
+		fz_append_byte(ctx, fzbuf, '{');
 		break;
 	case PDF_TOK_CLOSE_BRACE:
-		fz_buffer_printf(ctx, fzbuf, "}");
+		fz_append_byte(ctx, fzbuf, '}');
 		break;
 	case PDF_TOK_INT:
-		fz_buffer_printf(ctx, fzbuf, "%d", buf->i);
+		fz_append_printf(ctx, fzbuf, "%ld", buf->i);
 		break;
 	case PDF_TOK_REAL:
-		{
-			fz_buffer_printf(ctx, fzbuf, "%g", buf->f);
-		}
+		fz_append_printf(ctx, fzbuf, "%g", buf->f);
 		break;
 	default:
-		fz_buffer_printf(ctx, fzbuf, "%s", buf->scratch);
+		fz_append_data(ctx, fzbuf, buf->scratch, buf->len);
 		break;
 	}
 }
