@@ -1,16 +1,17 @@
+#include "mupdf/fitz.h"
 #include "mupdf/pdf.h"
+#include "pdf-imp.h"
+
+#include <string.h>
 
 /* Scan file for objects and reconstruct xref table */
 
-/* Define in PDF 1.7 to be 8388607, but mupdf is more lenient. */
-#define MAX_OBJECT_NUMBER (10 << 20)
-
 struct entry
 {
 	int num;
 	int gen;
-	int ofs;
-	int stm_ofs;
+	int64_t ofs;
+	int64_t stm_ofs;
 	int stm_len;
 };
 
@@ -28,7 +29,7 @@ static void add_root(fz_context *ctx, pdf_obj *obj, pdf_obj ***roots, int *num_r
 }
 
 int
-pdf_repair_obj(fz_context *ctx, pdf_document *doc, pdf_lexbuf *buf, fz_off_t *stmofsp, int *stmlenp, pdf_obj **encrypt, pdf_obj **id, pdf_obj **page, fz_off_t *tmpofs, pdf_obj **root)
+pdf_repair_obj(fz_context *ctx, pdf_document *doc, pdf_lexbuf *buf, int64_t *stmofsp, int *stmlenp, pdf_obj **encrypt, pdf_obj **id, pdf_obj **page, int64_t *tmpofs, pdf_obj **root)
 {
 	fz_stream *file = doc->file;
 	pdf_token tok;
@@ -48,7 +49,7 @@ pdf_repair_obj(fz_context *ctx, pdf_document *doc, pdf_lexbuf *buf, fz_off_t *st
 
 	if (tok == PDF_TOK_OPEN_DICT)
 	{
-		pdf_obj *dict, *obj;
+		pdf_obj *obj, *dict = NULL;
 
 		fz_try(ctx)
 		{
@@ -281,6 +282,11 @@ orphan_object(fz_context *ctx, pdf_document *doc, pdf_obj *obj)
 	doc->orphans[doc->orphans_count++] = obj;
 }
 
+static int is_white(int c)
+{
+	return c == '\x00' || c == '\x09' || c == '\x0a' || c == '\x0c' || c == '\x0d' || c == '\x20';
+}
+
 void
 pdf_repair_xref(fz_context *ctx, pdf_document *doc)
 {
@@ -299,7 +305,7 @@ pdf_repair_xref(fz_context *ctx, pdf_document *doc)
 
 	int num = 0;
 	int gen = 0;
-	fz_off_t tmpofs, stm_ofs, numofs = 0, genofs = 0;
+	int64_t tmpofs, stm_ofs, numofs = 0, genofs = 0;
 	int stm_len;
 	pdf_token tok;
 	int next;
@@ -319,13 +325,16 @@ pdf_repair_xref(fz_context *ctx, pdf_document *doc)
 	fz_var(list);
 	fz_var(obj);
 
+	fz_warn(ctx, "repairing PDF document");
+
 	if (doc->repair_attempted)
 		fz_throw(ctx, FZ_ERROR_GENERIC, "Repair failed already - not trying again");
 	doc->repair_attempted = 1;
 
 	doc->dirty = 1;
-	/* Can't support incremental update after repair */
-	doc->freeze_updates = 1;
+	doc->freeze_updates = 1; /* Can't support incremental update after repair */
+
+	pdf_forget_xref(ctx, doc);
 
 	fz_seek(ctx, doc->file, 0, 0);
 
@@ -346,7 +355,7 @@ pdf_repair_xref(fz_context *ctx, pdf_document *doc)
 			{
 				if (memcmp(&buf->scratch[j], "%PDF", 4) == 0)
 				{
-					fz_seek(ctx, doc->file, (fz_off_t)(j + 8), 0); /* skip "%PDF-X.Y" */
+					fz_seek(ctx, doc->file, (int64_t)(j + 8), 0); /* skip "%PDF-X.Y" */
 					break;
 				}
 			}
@@ -366,14 +375,15 @@ pdf_repair_xref(fz_context *ctx, pdf_document *doc)
 				fz_throw(ctx, FZ_ERROR_GENERIC, "cannot tell in file");
 
 			fz_try(ctx)
-			{
 				tok = pdf_lex_no_string(ctx, doc->file, buf);
-			}
 			fz_catch(ctx)
 			{
 				fz_rethrow_if(ctx, FZ_ERROR_TRYLATER);
-				fz_warn(ctx, "ignoring the rest of the file");
-				break;
+				fz_warn(ctx, "skipping ahead to next token");
+				do
+					c = fz_read_byte(ctx, doc->file);
+				while (c != EOF && !is_white(c));
+				continue;
 			}
 
 			/* If we have the next token already, then we'll jump
@@ -423,7 +433,7 @@ pdf_repair_xref(fz_context *ctx, pdf_document *doc)
 					break;
 				}
 
-				if (num <= 0 || num > MAX_OBJECT_NUMBER)
+				if (num <= 0 || num > PDF_MAX_OBJECT_NUMBER)
 				{
 					fz_warn(ctx, "ignoring object with invalid object number (%d %d R)", num, gen);
 					goto have_next_token;
@@ -469,45 +479,51 @@ pdf_repair_xref(fz_context *ctx, pdf_document *doc)
 					continue;
 				}
 
-				obj = pdf_dict_get(ctx, dict, PDF_NAME_Encrypt);
-				if (obj)
+				fz_try(ctx)
 				{
-					pdf_drop_obj(ctx, encrypt);
-					encrypt = pdf_keep_obj(ctx, obj);
+					obj = pdf_dict_get(ctx, dict, PDF_NAME_Encrypt);
+					if (obj)
+					{
+						pdf_drop_obj(ctx, encrypt);
+						encrypt = pdf_keep_obj(ctx, obj);
+					}
+
+					obj = pdf_dict_get(ctx, dict, PDF_NAME_ID);
+					if (obj && (!id || !encrypt || pdf_dict_get(ctx, dict, PDF_NAME_Encrypt)))
+					{
+						pdf_drop_obj(ctx, id);
+						id = pdf_keep_obj(ctx, obj);
+					}
+
+					obj = pdf_dict_get(ctx, dict, PDF_NAME_Root);
+					if (obj)
+						add_root(ctx, obj, &roots, &num_roots, &max_roots);
+
+					obj = pdf_dict_get(ctx, dict, PDF_NAME_Info);
+					if (obj)
+					{
+						pdf_drop_obj(ctx, info);
+						info = pdf_keep_obj(ctx, obj);
+					}
 				}
+				fz_always(ctx)
+					pdf_drop_obj(ctx, dict);
+				fz_catch(ctx)
+					fz_rethrow(ctx);
 
-				obj = pdf_dict_get(ctx, dict, PDF_NAME_ID);
-				if (obj && (!id || !encrypt || pdf_dict_get(ctx, dict, PDF_NAME_Encrypt)))
-				{
-					pdf_drop_obj(ctx, id);
-					id = pdf_keep_obj(ctx, obj);
-				}
-
-				obj = pdf_dict_get(ctx, dict, PDF_NAME_Root);
-				if (obj)
-					add_root(ctx, obj, &roots, &num_roots, &max_roots);
-
-				obj = pdf_dict_get(ctx, dict, PDF_NAME_Info);
-				if (obj)
-				{
-					pdf_drop_obj(ctx, info);
-					info = pdf_keep_obj(ctx, obj);
-				}
-
-				pdf_drop_obj(ctx, dict);
 				obj = NULL;
 			}
 
 			else if (tok == PDF_TOK_EOF)
+			{
 				break;
+			}
+
 			else
 			{
-				if (tok == PDF_TOK_ERROR)
-					fz_read_byte(ctx, doc->file);
 				num = 0;
 				gen = 0;
 			}
-
 		}
 
 		if (listlen == 0)
@@ -552,11 +568,17 @@ pdf_repair_xref(fz_context *ctx, pdf_document *doc)
 				pdf_obj *old_obj = NULL;
 				dict = pdf_load_object(ctx, doc, list[i].num);
 
-				length = pdf_new_int(ctx, doc, list[i].stm_len);
-				pdf_dict_get_put_drop(ctx, dict, PDF_NAME_Length, length, &old_obj);
-				if (old_obj)
-					orphan_object(ctx, doc, old_obj);
-				pdf_drop_obj(ctx, dict);
+				fz_try(ctx)
+				{
+					length = pdf_new_int(ctx, doc, list[i].stm_len);
+					pdf_dict_get_put_drop(ctx, dict, PDF_NAME_Length, length, &old_obj);
+					if (old_obj)
+						orphan_object(ctx, doc, old_obj);
+				}
+				fz_always(ctx)
+					pdf_drop_obj(ctx, dict);
+				fz_catch(ctx)
+					fz_rethrow(ctx);
 			}
 		}
 
@@ -595,7 +617,6 @@ pdf_repair_xref(fz_context *ctx, pdf_document *doc)
 
 		if (roots)
 		{
-			int i;
 			for (i = num_roots-1; i > 0; i--)
 			{
 				if (pdf_is_dict(ctx, roots[i]))
@@ -647,8 +668,6 @@ pdf_repair_xref(fz_context *ctx, pdf_document *doc)
 	}
 	fz_always(ctx)
 	{
-		int i;
-
 		for (i = 0; i < num_roots; i++)
 			pdf_drop_obj(ctx, roots[i]);
 		fz_free(ctx, roots);
