@@ -1,3 +1,4 @@
+#include "mupdf/fitz.h"
 #include "mupdf/pdf.h"
 
 static void
@@ -26,22 +27,20 @@ static void
 pdf_preload_image_resources(fz_context *ctx, pdf_document *doc)
 {
 	int len, k;
-	pdf_obj *obj;
+	pdf_obj *obj = NULL;
 	pdf_obj *type;
-	pdf_obj *res = NULL;
 	fz_image *image = NULL;
 	unsigned char digest[16];
 
 	fz_var(obj);
 	fz_var(image);
-	fz_var(res);
 
 	fz_try(ctx)
 	{
 		len = pdf_count_objects(ctx, doc);
 		for (k = 1; k < len; k++)
 		{
-			obj = pdf_load_object(ctx, doc, k);
+			obj = pdf_new_indirect(ctx, doc, k, 0);
 			type = pdf_dict_get(ctx, obj, PDF_NAME_Subtype);
 			if (pdf_name_eq(ctx, type, PDF_NAME_Image))
 			{
@@ -69,6 +68,11 @@ pdf_preload_image_resources(fz_context *ctx, pdf_document *doc)
 	}
 }
 
+static void pdf_drop_obj_as_void(fz_context *ctx, void *obj)
+{
+	pdf_drop_obj(ctx, obj);
+}
+
 pdf_obj *
 pdf_find_image_resource(fz_context *ctx, pdf_document *doc, fz_image *item, unsigned char digest[16])
 {
@@ -76,7 +80,7 @@ pdf_find_image_resource(fz_context *ctx, pdf_document *doc, fz_image *item, unsi
 
 	if (!doc->resources.images)
 	{
-		doc->resources.images = fz_new_hash_table(ctx, 4096, 16, -1, (fz_hash_table_drop_fn)pdf_drop_obj);
+		doc->resources.images = fz_new_hash_table(ctx, 4096, 16, -1, pdf_drop_obj_as_void);
 		pdf_preload_image_resources(ctx, doc);
 	}
 
@@ -96,7 +100,7 @@ pdf_insert_image_resource(fz_context *ctx, pdf_document *doc, unsigned char dige
 		fz_warn(ctx, "warning: image resource already present");
 	else
 		res = pdf_keep_obj(ctx, obj);
-	return res;
+	return pdf_keep_obj(ctx, res);
 }
 
 /* We do need to come up with an effective way to see what is already in the
@@ -105,15 +109,25 @@ pdf_insert_image_resource(fz_context *ctx, pdf_document *doc, unsigned char dige
  * it may be more problematic. */
 
 pdf_obj *
-pdf_find_font_resource(fz_context *ctx, pdf_document *doc, fz_buffer *item, unsigned char digest[16])
+pdf_find_font_resource(fz_context *ctx, pdf_document *doc, int type, int encoding, fz_buffer *item, unsigned char digest[16])
 {
+	fz_md5 state;
+	unsigned char *data;
+	size_t size;
 	pdf_obj *res;
 
 	if (!doc->resources.fonts)
-		doc->resources.fonts = fz_new_hash_table(ctx, 4096, 16, -1, (fz_hash_table_drop_fn)pdf_drop_obj);
+		doc->resources.fonts = fz_new_hash_table(ctx, 4096, 16, -1, pdf_drop_obj_as_void);
+
+	size = fz_buffer_storage(ctx, item, &data);
 
 	/* Create md5 and see if we have the item in our table */
-	fz_md5_buffer(ctx, item, digest);
+	fz_md5_init(&state);
+	fz_md5_update(&state, (unsigned char*)&type, sizeof type);
+	fz_md5_update(&state, (unsigned char*)&encoding, sizeof encoding);
+	fz_md5_update(&state, data, size);
+	fz_md5_final(&state, digest);
+
 	res = fz_hash_find(ctx, doc->resources.fonts, digest);
 	if (res)
 		pdf_keep_obj(ctx, res);
@@ -128,7 +142,7 @@ pdf_insert_font_resource(fz_context *ctx, pdf_document *doc, unsigned char diges
 		fz_warn(ctx, "warning: font resource already present");
 	else
 		res = pdf_keep_obj(ctx, obj);
-	return res;
+	return pdf_keep_obj(ctx, res);
 }
 
 void
