@@ -1,7 +1,14 @@
+#include "mupdf/fitz.h"
+#include "mupdf/pdf.h"
 #include "pdf-imp.h"
 
 #include <zlib.h>
 
+#include <assert.h>
+#include <limits.h>
+#include <string.h>
+
+#include <stdio.h> /* for debug printing */
 /* #define DEBUG_LINEARIZATION */
 /* #define DEBUG_HEAP_SORT */
 /* #define DEBUG_WRITING */
@@ -60,8 +67,9 @@ struct pdf_write_state_s
 	int do_linear;
 	int do_clean;
 
+	int list_len;
 	int *use_list;
-	fz_off_t *ofs_list;
+	int64_t *ofs_list;
 	int *gen_list;
 	int *renumber_map;
 	int continue_on_error;
@@ -69,10 +77,10 @@ struct pdf_write_state_s
 	/* The following extras are required for linearization */
 	int *rev_renumber_map;
 	int start;
-	fz_off_t first_xref_offset;
-	fz_off_t main_xref_offset;
-	fz_off_t first_xref_entry_offset;
-	fz_off_t file_len;
+	int64_t first_xref_offset;
+	int64_t main_xref_offset;
+	int64_t first_xref_entry_offset;
+	int64_t file_len;
 	int hints_shared_offset;
 	int hintstream_len;
 	pdf_obj *linear_l;
@@ -86,6 +94,7 @@ struct pdf_write_state_s
 	pdf_obj *hints_length;
 	int page_count;
 	page_objects_list *page_object_lists;
+	int crypt_object_number;
 };
 
 /*
@@ -316,12 +325,6 @@ order_ge(int ui, int uj)
 		return 1;
 	else if (uj & USE_OTHER_OBJECTS)
 		return 0;
-	/* Put objects not associated with any page (anything
-	 * not touched by the catalogue) before that... */
-	else if (ui == 0)
-		return 1;
-	else if (uj == 0)
-		return 0;
 	/* Put shared objects before that... */
 	else if (ui & USE_SHARED)
 		return 1;
@@ -618,7 +621,7 @@ static int markobj(fz_context *ctx, pdf_document *doc, pdf_write_state *opts, pd
 		{
 			DEBUGGING_MARKING(indent(); printf("ARRAY[%d/%d]\n", i, n));
 			if (markobj(ctx, doc, opts, pdf_array_get(ctx, obj, i)))
-				pdf_array_put_drop(ctx, obj, i, pdf_new_null(ctx, doc));
+				pdf_array_put(ctx, obj, i, PDF_OBJ_NULL);
 		}
 	}
 
@@ -627,13 +630,37 @@ static int markobj(fz_context *ctx, pdf_document *doc, pdf_write_state *opts, pd
 	return 0;
 }
 
+static void
+expand_lists(fz_context *ctx, pdf_write_state *opts, int num)
+{
+	int i;
+
+	/* objects are numbered 0..num and maybe two additional objects for linearization */
+	num += 3;
+	opts->use_list = fz_resize_array(ctx, opts->use_list, num, sizeof(*opts->use_list));
+	opts->ofs_list = fz_resize_array(ctx, opts->ofs_list, num, sizeof(*opts->ofs_list));
+	opts->gen_list = fz_resize_array(ctx, opts->gen_list, num, sizeof(*opts->gen_list));
+	opts->renumber_map = fz_resize_array(ctx, opts->renumber_map, num, sizeof(*opts->renumber_map));
+	opts->rev_renumber_map = fz_resize_array(ctx, opts->rev_renumber_map, num, sizeof(*opts->rev_renumber_map));
+
+	for (i = opts->list_len; i < num; i++)
+	{
+		opts->use_list[i] = 0;
+		opts->ofs_list[i] = 0;
+		opts->gen_list[i] = 0;
+		opts->renumber_map[i] = i;
+		opts->rev_renumber_map[i] = i;
+	}
+	opts->list_len = num;
+}
+
 /*
  * Scan for and remove duplicate objects (slow)
  */
 
 static void removeduplicateobjs(fz_context *ctx, pdf_document *doc, pdf_write_state *opts)
 {
-	int num, other;
+	int num, other, max_num;
 	int xref_len = pdf_xref_len(ctx, doc);
 
 	for (num = 1; num < xref_len; num++)
@@ -642,7 +669,7 @@ static void removeduplicateobjs(fz_context *ctx, pdf_document *doc, pdf_write_st
 		for (other = 1; other < num; other++)
 		{
 			pdf_obj *a, *b;
-			int differ, newnum, streama, streamb;
+			int newnum, streama = 0, streamb = 0, differ = 0;
 
 			if (num == other || !opts->use_list[num] || !opts->use_list[other])
 				continue;
@@ -713,6 +740,9 @@ static void removeduplicateobjs(fz_context *ctx, pdf_document *doc, pdf_write_st
 
 			/* Keep the lowest numbered object */
 			newnum = fz_mini(num, other);
+			max_num = fz_maxi(num, other);
+			if (max_num >= opts->list_len)
+				expand_lists(ctx, opts, max_num);
 			opts->renumber_map[num] = newnum;
 			opts->renumber_map[other] = newnum;
 			opts->rev_renumber_map[newnum] = num; /* Either will do */
@@ -742,6 +772,9 @@ static void compactxref(fz_context *ctx, pdf_document *doc, pdf_write_state *opt
 	 * object ids be updated to reflect the compaction.
 	 */
 
+	if (xref_len > opts->list_len)
+		expand_lists(ctx, opts, xref_len-1);
+
 	newnum = 1;
 	for (num = 1; num < xref_len; num++)
 	{
@@ -790,8 +823,7 @@ static void renumberobj(fz_context *ctx, pdf_document *doc, pdf_write_state *opt
 					val = pdf_new_null(ctx, doc);
 				else
 					val = pdf_new_indirect(ctx, doc, opts->renumber_map[o], 0);
-				pdf_dict_put(ctx, obj, key, val);
-				pdf_drop_obj(ctx, val);
+				pdf_dict_put_drop(ctx, obj, key, val);
 			}
 			else
 			{
@@ -813,8 +845,7 @@ static void renumberobj(fz_context *ctx, pdf_document *doc, pdf_write_state *opt
 					val = pdf_new_null(ctx, doc);
 				else
 					val = pdf_new_indirect(ctx, doc, opts->renumber_map[o], 0);
-				pdf_array_put(ctx, obj, i, val);
-				pdf_drop_obj(ctx, val);
+				pdf_array_put_drop(ctx, obj, i, val);
 			}
 			else
 			{
@@ -853,8 +884,12 @@ static void renumberobjs(fz_context *ctx, pdf_document *doc, pdf_write_state *op
 			if (pdf_is_indirect(ctx, obj))
 			{
 				obj = pdf_new_indirect(ctx, doc, to, 0);
-				pdf_update_object(ctx, doc, num, obj);
-				pdf_drop_obj(ctx, obj);
+				fz_try(ctx)
+					pdf_update_object(ctx, doc, num, obj);
+				fz_always(ctx)
+					pdf_drop_obj(ctx, obj);
+				fz_catch(ctx)
+					fz_rethrow(ctx);
 			}
 			else
 			{
@@ -1152,7 +1187,7 @@ add_linearization_objs(fz_context *ctx, pdf_document *doc, pdf_write_state *opts
 		opts->renumber_map[params_num] = params_num;
 		opts->rev_renumber_map[params_num] = params_num;
 		opts->gen_list[params_num] = 0;
-		pdf_dict_put_drop(ctx, params_obj, PDF_NAME_Linearized, pdf_new_real(ctx, doc, 1.0));
+		pdf_dict_put_real(ctx, params_obj, PDF_NAME_Linearized, 1.0f);
 		opts->linear_l = pdf_new_int(ctx, doc, INT_MIN);
 		pdf_dict_put(ctx, params_obj, PDF_NAME_L, opts->linear_l);
 		opts->linear_h0 = pdf_new_int(ctx, doc, INT_MIN);
@@ -1179,7 +1214,7 @@ add_linearization_objs(fz_context *ctx, pdf_document *doc, pdf_write_state *opts
 		opts->renumber_map[hint_num] = hint_num;
 		opts->rev_renumber_map[hint_num] = hint_num;
 		opts->gen_list[hint_num] = 0;
-		pdf_dict_put_drop(ctx, hint_obj, PDF_NAME_P, pdf_new_int(ctx, doc, 0));
+		pdf_dict_put_int(ctx, hint_obj, PDF_NAME_P, 0);
 		opts->hints_s = pdf_new_int(ctx, doc, INT_MIN);
 		pdf_dict_put(ctx, hint_obj, PDF_NAME_S, opts->hints_s);
 		/* FIXME: Do we have thumbnails? Do a T entry */
@@ -1190,7 +1225,7 @@ add_linearization_objs(fz_context *ctx, pdf_document *doc, pdf_write_state *opts
 		/* FIXME: Do we have document information? Do an I entry */
 		/* FIXME: Do we have logical structure hierarchy? Do a C entry */
 		/* FIXME: Do L, Page Label hint table */
-		pdf_dict_put_drop(ctx, hint_obj, PDF_NAME_Filter, PDF_NAME_FlateDecode);
+		pdf_dict_put(ctx, hint_obj, PDF_NAME_Filter, PDF_NAME_FlateDecode);
 		opts->hints_length = pdf_new_int(ctx, doc, INT_MIN);
 		pdf_dict_put(ctx, hint_obj, PDF_NAME_Length, opts->hints_length);
 		pdf_get_xref_entry(ctx, doc, hint_num)->stm_ofs = -1;
@@ -1458,13 +1493,13 @@ linearize(fz_context *ctx, pdf_document *doc, pdf_write_state *opts)
 static void
 update_linearization_params(fz_context *ctx, pdf_document *doc, pdf_write_state *opts)
 {
-	fz_off_t offset;
+	int64_t offset;
 	pdf_set_int(ctx, opts->linear_l, opts->file_len);
 	/* Primary hint stream offset (of object, not stream!) */
-	pdf_set_int_offset(ctx, opts->linear_h0, opts->ofs_list[pdf_xref_len(ctx, doc)-1]);
+	pdf_set_int(ctx, opts->linear_h0, opts->ofs_list[pdf_xref_len(ctx, doc)-1]);
 	/* Primary hint stream length (of object, not stream!) */
 	offset = (opts->start == 1 ? opts->main_xref_offset : opts->ofs_list[1] + opts->hintstream_len);
-	pdf_set_int_offset(ctx, opts->linear_h1, offset - opts->ofs_list[pdf_xref_len(ctx, doc)-1]);
+	pdf_set_int(ctx, opts->linear_h1, offset - opts->ofs_list[pdf_xref_len(ctx, doc)-1]);
 	/* Object number of first pages page object (the first object of page 0) */
 	pdf_set_int(ctx, opts->linear_o, opts->page_object_lists->page[0]->object[0]);
 	/* Offset of end of first page (first page is followed by primary
@@ -1472,13 +1507,13 @@ update_linearization_params(fz_context *ctx, pdf_document *doc, pdf_write_state
 	 * primary hint stream counts as part of the first pages data, I think.
 	 */
 	offset = (opts->start == 1 ? opts->main_xref_offset : opts->ofs_list[1] + opts->hintstream_len);
-	pdf_set_int_offset(ctx, opts->linear_e, offset);
+	pdf_set_int(ctx, opts->linear_e, offset);
 	/* Number of pages in document */
 	pdf_set_int(ctx, opts->linear_n, opts->page_count);
 	/* Offset of first entry in main xref table */
-	pdf_set_int_offset(ctx, opts->linear_t, opts->first_xref_entry_offset + opts->hintstream_len);
+	pdf_set_int(ctx, opts->linear_t, opts->first_xref_entry_offset + opts->hintstream_len);
 	/* Offset of shared objects hint table in the primary hint stream */
-	pdf_set_int_offset(ctx, opts->hints_s, opts->hints_shared_offset);
+	pdf_set_int(ctx, opts->hints_s, opts->hints_shared_offset);
 	/* Primary hint stream length */
 	pdf_set_int(ctx, opts->hints_length, opts->hintstream_len);
 }
@@ -1491,9 +1526,9 @@ static void preloadobjstms(fz_context *ctx, pdf_document *doc)
 {
 	pdf_obj *obj;
 	int num;
-	int xref_len = pdf_xref_len(ctx, doc);
 
-	for (num = 0; num < xref_len; num++)
+	/* xref_len may change due to repair, so check it every iteration */
+	for (num = 0; num < pdf_xref_len(ctx, doc); num++)
 	{
 		if (pdf_get_xref_entry(ctx, doc, num)->type == 'o')
 		{
@@ -1525,11 +1560,11 @@ static int isbinarystream(fz_context *ctx, fz_buffer *buf)
 	return 0;
 }
 
-static fz_buffer *hexbuf(fz_context *ctx, unsigned char *p, size_t n)
+static fz_buffer *hexbuf(fz_context *ctx, const unsigned char *p, size_t n)
 {
 	static const char hex[17] = "0123456789abcdef";
 	int x = 0;
-	size_t len = n * 2 + (n / 32) + 2;
+	size_t len = n * 2 + (n / 32) + 1;
 	unsigned char *data = fz_malloc(ctx, len);
 	fz_buffer *buf = fz_new_buffer_from_data(ctx, data, len);
 
@@ -1546,7 +1581,6 @@ static fz_buffer *hexbuf(fz_context *ctx, unsigned char *p, size_t n)
 	}
 
 	*data++ = '>';
-	*data++ = '\n';
 
 	return buf;
 }
@@ -1573,7 +1607,7 @@ static void addhexfilter(fz_context *ctx, pdf_document *doc, pdf_obj *dict)
 			if (pdf_is_dict(ctx, dp))
 			{
 				newdp = pdf_new_array(ctx, doc, 2);
-				pdf_array_push_drop(ctx, newdp, pdf_new_null(ctx, doc));
+				pdf_array_push(ctx, newdp, PDF_OBJ_NULL);
 				pdf_array_push(ctx, newdp, dp);
 				dp = newdp;
 			}
@@ -1582,7 +1616,7 @@ static void addhexfilter(fz_context *ctx, pdf_document *doc, pdf_obj *dict)
 		{
 			pdf_array_insert(ctx, f, PDF_NAME_ASCIIHexDecode, 0);
 			if (pdf_is_array(ctx, dp))
-				pdf_array_insert_drop(ctx, dp, pdf_new_null(ctx, doc), 0);
+				pdf_array_insert(ctx, dp, PDF_OBJ_NULL, 0);
 		}
 		else
 			f = PDF_NAME_ASCIIHexDecode;
@@ -1600,7 +1634,7 @@ static void addhexfilter(fz_context *ctx, pdf_document *doc, pdf_obj *dict)
 		fz_rethrow(ctx);
 }
 
-static fz_buffer *deflatebuf(fz_context *ctx, unsigned char *p, size_t n)
+static fz_buffer *deflatebuf(fz_context *ctx, const unsigned char *p, size_t n)
 {
 	fz_buffer *buf;
 	uLongf csize;
@@ -1626,6 +1660,11 @@ static fz_buffer *deflatebuf(fz_context *ctx, unsigned char *p, size_t n)
 	return buf;
 }
 
+static void write_data(fz_context *ctx, void *arg, const unsigned char *data, int len)
+{
+	fz_write_data(ctx, (fz_output *)arg, data, len);
+}
+
 static void copystream(fz_context *ctx, pdf_document *doc, pdf_write_state *opts, pdf_obj *obj_orig, int num, int gen, int do_deflate)
 {
 	fz_buffer *buf, *tmp;
@@ -1670,17 +1709,14 @@ static void copystream(fz_context *ctx, pdf_document *doc, pdf_write_state *opts
 		addhexfilter(ctx, doc, obj);
 	}
 
-	newlen = pdf_new_int(ctx, doc, (int)len);
-	pdf_dict_put(ctx, obj, PDF_NAME_Length, newlen);
-	pdf_drop_obj(ctx, newlen);
+	newlen = pdf_new_int(ctx, doc, pdf_encrypted_len(ctx, doc->crypt, num, gen, (int)len));
+	pdf_dict_put_drop(ctx, obj, PDF_NAME_Length, newlen);
 
 	fz_write_printf(ctx, opts->out, "%d %d obj\n", num, gen);
-	pdf_print_obj(ctx, opts->out, obj, opts->do_tight);
+	pdf_print_encrypted_obj(ctx, opts->out, obj, opts->do_tight, doc->crypt, num, gen);
 	fz_write_string(ctx, opts->out, "\nstream\n");
-	fz_write_data(ctx, opts->out, data, len);
-	if (len > 0 && data[len-1] != '\n')
-		fz_write_byte(ctx, opts->out, '\n');
-	fz_write_string(ctx, opts->out, "endstream\nendobj\n\n");
+	pdf_encrypt_data(ctx, doc->crypt, num, gen, write_data, opts->out, data, len);
+	fz_write_string(ctx, opts->out, "\nendstream\nendobj\n\n");
 
 	fz_drop_buffer(ctx, buf);
 	pdf_drop_obj(ctx, obj);
@@ -1689,7 +1725,6 @@ static void copystream(fz_context *ctx, pdf_document *doc, pdf_write_state *opts
 static void expandstream(fz_context *ctx, pdf_document *doc, pdf_write_state *opts, pdf_obj *obj_orig, int num, int gen, int do_deflate)
 {
 	fz_buffer *buf, *tmp;
-	pdf_obj *newlen;
 	pdf_obj *obj;
 	int truncated = 0;
 	size_t len;
@@ -1735,23 +1770,19 @@ static void expandstream(fz_context *ctx, pdf_document *doc, pdf_write_state *op
 		addhexfilter(ctx, doc, obj);
 	}
 
-	newlen = pdf_new_int(ctx, doc, (int)len);
-	pdf_dict_put(ctx, obj, PDF_NAME_Length, newlen);
-	pdf_drop_obj(ctx, newlen);
+	pdf_dict_put_int(ctx, obj, PDF_NAME_Length, len);
 
 	fz_write_printf(ctx, opts->out, "%d %d obj\n", num, gen);
-	pdf_print_obj(ctx, opts->out, obj, opts->do_tight);
+	pdf_print_encrypted_obj(ctx, opts->out, obj, opts->do_tight, doc->crypt, num, gen);
 	fz_write_string(ctx, opts->out, "\nstream\n");
 	fz_write_data(ctx, opts->out, data, len);
-	if (len > 0 && data[len-1] != '\n')
-		fz_write_byte(ctx, opts->out, '\n');
-	fz_write_string(ctx, opts->out, "endstream\nendobj\n\n");
+	fz_write_string(ctx, opts->out, "\nendstream\nendobj\n\n");
 
 	fz_drop_buffer(ctx, buf);
 	pdf_drop_obj(ctx, obj);
 }
 
-static int is_image_filter(char *s)
+static int is_image_filter(const char *s)
 {
 	if (!strcmp(s, "CCITTFaxDecode") || !strcmp(s, "CCF") ||
 		!strcmp(s, "DCTDecode") || !strcmp(s, "DCT") ||
@@ -1812,7 +1843,15 @@ static int is_font_stream(fz_context *ctx, pdf_obj *obj)
 	return 0;
 }
 
-static void writeobject(fz_context *ctx, pdf_document *doc, pdf_write_state *opts, int num, int gen, int skip_xrefs)
+static int is_xml_metadata(fz_context *ctx, pdf_obj *obj)
+{
+	if (pdf_name_eq(ctx, pdf_dict_get(ctx, obj, PDF_NAME_Type), PDF_NAME_Metadata))
+		if (pdf_name_eq(ctx, pdf_dict_get(ctx, obj, PDF_NAME_Subtype), PDF_NAME_XML))
+			return 1;
+	return 0;
+}
+
+static void writeobject(fz_context *ctx, pdf_document *doc, pdf_write_state *opts, int num, int gen, int skip_xrefs, int unenc)
 {
 	pdf_xref_entry *entry;
 	pdf_obj *obj;
@@ -1859,13 +1898,13 @@ static void writeobject(fz_context *ctx, pdf_document *doc, pdf_write_state *opt
 	if (!pdf_obj_num_is_stream(ctx, doc, num))
 	{
 		fz_write_printf(ctx, opts->out, "%d %d obj\n", num, gen);
-		pdf_print_obj(ctx, opts->out, obj, opts->do_tight);
+		pdf_print_encrypted_obj(ctx, opts->out, obj, opts->do_tight, unenc ? NULL : doc->crypt, num, gen);
 		fz_write_string(ctx, opts->out, "\nendobj\n\n");
 	}
 	else if (entry->stm_ofs < 0 && entry->stm_buf == NULL)
 	{
 		fz_write_printf(ctx, opts->out, "%d %d obj\n", num, gen);
-		pdf_print_obj(ctx, opts->out, obj, opts->do_tight);
+		pdf_print_encrypted_obj(ctx, opts->out, obj, opts->do_tight, doc->crypt, num, gen);
 		fz_write_string(ctx, opts->out, "\nstream\nendstream\nendobj\n\n");
 	}
 	else
@@ -1878,6 +1917,8 @@ static void writeobject(fz_context *ctx, pdf_document *doc, pdf_write_state *opt
 				do_deflate = 1, do_expand = 0;
 			if (opts->do_compress_fonts && is_font_stream(ctx, obj))
 				do_deflate = 1, do_expand = 0;
+			if (is_xml_metadata(ctx, obj))
+				do_deflate = 0, do_expand = 0;
 			if (do_expand)
 				expandstream(ctx, doc, opts, obj, num, gen, do_deflate);
 			else
@@ -1912,13 +1953,13 @@ static void writexrefsubsect(fz_context *ctx, pdf_write_state *opts, int from, i
 	for (num = from; num < to; num++)
 	{
 		if (opts->use_list[num])
-			fz_write_printf(ctx, opts->out, "%010Zd %05d n \n", opts->ofs_list[num], opts->gen_list[num]);
+			fz_write_printf(ctx, opts->out, "%010lu %05d n \n", opts->ofs_list[num], opts->gen_list[num]);
 		else
-			fz_write_printf(ctx, opts->out, "%010Zd %05d f \n", opts->ofs_list[num], opts->gen_list[num]);
+			fz_write_printf(ctx, opts->out, "%010lu %05d f \n", opts->ofs_list[num], opts->gen_list[num]);
 	}
 }
 
-static void writexref(fz_context *ctx, pdf_document *doc, pdf_write_state *opts, int from, int to, int first, int main_xref_offset, int startxref)
+static void writexref(fz_context *ctx, pdf_document *doc, pdf_write_state *opts, int from, int to, int first, int64_t main_xref_offset, int64_t startxref)
 {
 	pdf_obj *trailer = NULL;
 	pdf_obj *obj;
@@ -1955,65 +1996,54 @@ static void writexref(fz_context *ctx, pdf_document *doc, pdf_write_state *opts,
 	fz_write_string(ctx, opts->out, "\n");
 
 	fz_var(trailer);
-	fz_var(nobj);
 
-	fz_try(ctx)
+	if (opts->do_incremental)
 	{
-		if (opts->do_incremental)
-		{
-			trailer = pdf_keep_obj(ctx, pdf_trailer(ctx, doc));
-			pdf_dict_put_drop(ctx, trailer, PDF_NAME_Size, pdf_new_int(ctx, doc, pdf_xref_len(ctx, doc)));
-			pdf_dict_put_drop(ctx, trailer, PDF_NAME_Prev, pdf_new_int(ctx, doc, doc->startxref));
-			doc->startxref = startxref;
-		}
-		else
-		{
-			trailer = pdf_new_dict(ctx, doc, 5);
-
-			nobj = pdf_new_int(ctx, doc, to);
-			pdf_dict_put(ctx, trailer, PDF_NAME_Size, nobj);
-			pdf_drop_obj(ctx, nobj);
-			nobj = NULL;
-
-			if (first)
-			{
-				obj = pdf_dict_get(ctx, pdf_trailer(ctx, doc), PDF_NAME_Info);
-				if (obj)
-					pdf_dict_put(ctx, trailer, PDF_NAME_Info, obj);
-
-				obj = pdf_dict_get(ctx, pdf_trailer(ctx, doc), PDF_NAME_Root);
-				if (obj)
-					pdf_dict_put(ctx, trailer, PDF_NAME_Root, obj);
-
-				obj = pdf_dict_get(ctx, pdf_trailer(ctx, doc), PDF_NAME_ID);
-				if (obj)
-					pdf_dict_put(ctx, trailer, PDF_NAME_ID, obj);
-			}
-			if (main_xref_offset != 0)
-			{
-				nobj = pdf_new_int(ctx, doc, main_xref_offset);
-				pdf_dict_put(ctx, trailer, PDF_NAME_Prev, nobj);
-				pdf_drop_obj(ctx, nobj);
-				nobj = NULL;
-			}
-		}
+		trailer = pdf_keep_obj(ctx, pdf_trailer(ctx, doc));
+		pdf_dict_put_int(ctx, trailer, PDF_NAME_Size, pdf_xref_len(ctx, doc));
+		pdf_dict_put_int(ctx, trailer, PDF_NAME_Prev, doc->startxref);
+		doc->startxref = startxref;
 	}
-	fz_always(ctx)
+	else
 	{
-		pdf_drop_obj(ctx, nobj);
-	}
-	fz_catch(ctx)
-	{
-		fz_rethrow(ctx);
+		trailer = pdf_new_dict(ctx, doc, 5);
+
+		nobj = pdf_new_int(ctx, doc, to);
+		pdf_dict_put_drop(ctx, trailer, PDF_NAME_Size, nobj);
+
+		if (first)
+		{
+			obj = pdf_dict_get(ctx, pdf_trailer(ctx, doc), PDF_NAME_Info);
+			if (obj)
+				pdf_dict_put(ctx, trailer, PDF_NAME_Info, obj);
+
+			obj = pdf_dict_get(ctx, pdf_trailer(ctx, doc), PDF_NAME_Root);
+			if (obj)
+				pdf_dict_put(ctx, trailer, PDF_NAME_Root, obj);
+
+			obj = pdf_dict_get(ctx, pdf_trailer(ctx, doc), PDF_NAME_ID);
+			if (obj)
+				pdf_dict_put(ctx, trailer, PDF_NAME_ID, obj);
+
+			obj = pdf_dict_get(ctx, pdf_trailer(ctx, doc), PDF_NAME_Encrypt);
+			if (obj)
+				pdf_dict_put(ctx, trailer, PDF_NAME_Encrypt, obj);
+		}
+		if (main_xref_offset != 0)
+		{
+			nobj = pdf_new_int(ctx, doc, main_xref_offset);
+			pdf_dict_put_drop(ctx, trailer, PDF_NAME_Prev, nobj);
+		}
 	}
 
 	fz_write_string(ctx, opts->out, "trailer\n");
+	/* Trailer is NOT encrypted */
 	pdf_print_obj(ctx, opts->out, trailer, opts->do_tight);
 	fz_write_string(ctx, opts->out, "\n");
 
 	pdf_drop_obj(ctx, trailer);
 
-	fz_write_printf(ctx, opts->out, "startxref\n%d\n%%%%EOF\n", startxref);
+	fz_write_printf(ctx, opts->out, "startxref\n%lu\n%%%%EOF\n", startxref);
 
 	doc->has_xref_streams = 0;
 }
@@ -2022,8 +2052,8 @@ static void writexrefstreamsubsect(fz_context *ctx, pdf_document *doc, pdf_write
 {
 	int num;
 
-	pdf_array_push_drop(ctx, index, pdf_new_int(ctx, doc, from));
-	pdf_array_push_drop(ctx, index, pdf_new_int(ctx, doc, to - from));
+	pdf_array_push_int(ctx, index, from);
+	pdf_array_push_int(ctx, index, to - from);
 	for (num = from; num < to; num++)
 	{
 		fz_append_byte(ctx, fzbuf, opts->use_list[num] ? 1 : 0);
@@ -2035,7 +2065,7 @@ static void writexrefstreamsubsect(fz_context *ctx, pdf_document *doc, pdf_write
 	}
 }
 
-static void writexrefstream(fz_context *ctx, pdf_document *doc, pdf_write_state *opts, int from, int to, int first, int main_xref_offset, int startxref)
+static void writexrefstream(fz_context *ctx, pdf_document *doc, pdf_write_state *opts, int from, int to, int first, int64_t main_xref_offset, int64_t startxref)
 {
 	int num;
 	pdf_obj *dict = NULL;
@@ -2079,26 +2109,26 @@ static void writexrefstream(fz_context *ctx, pdf_document *doc, pdf_write_state
 			}
 		}
 
-		pdf_dict_put_drop(ctx, dict, PDF_NAME_Size, pdf_new_int(ctx, doc, to));
+		pdf_dict_put_int(ctx, dict, PDF_NAME_Size, to);
 
 		if (opts->do_incremental)
 		{
-			pdf_dict_put_drop(ctx, dict, PDF_NAME_Prev, pdf_new_int(ctx, doc, doc->startxref));
+			pdf_dict_put_int(ctx, dict, PDF_NAME_Prev, doc->startxref);
 			doc->startxref = startxref;
 		}
 		else
 		{
 			if (main_xref_offset != 0)
-				pdf_dict_put_drop(ctx, dict, PDF_NAME_Prev, pdf_new_int(ctx, doc, main_xref_offset));
+				pdf_dict_put_int(ctx, dict, PDF_NAME_Prev, main_xref_offset);
 		}
 
-		pdf_dict_put_drop(ctx, dict, PDF_NAME_Type, PDF_NAME_XRef);
+		pdf_dict_put(ctx, dict, PDF_NAME_Type, PDF_NAME_XRef);
 
 		w = pdf_new_array(ctx, doc, 3);
 		pdf_dict_put(ctx, dict, PDF_NAME_W, w);
-		pdf_array_push_drop(ctx, w, pdf_new_int(ctx, doc, 1));
-		pdf_array_push_drop(ctx, w, pdf_new_int(ctx, doc, 4));
-		pdf_array_push_drop(ctx, w, pdf_new_int(ctx, doc, 1));
+		pdf_array_push_int(ctx, w, 1);
+		pdf_array_push_int(ctx, w, 4);
+		pdf_array_push_int(ctx, w, 1);
 
 		index = pdf_new_array(ctx, doc, 2);
 		pdf_dict_put_drop(ctx, dict, PDF_NAME_Index, index);
@@ -2136,8 +2166,8 @@ static void writexrefstream(fz_context *ctx, pdf_document *doc, pdf_write_state
 
 		pdf_update_stream(ctx, doc, dict, fzbuf, 0);
 
-		writeobject(ctx, doc, opts, num, 0, 0);
-		fz_write_printf(ctx, opts->out, "startxref\n%Zd\n%%%%EOF\n", startxref);
+		writeobject(ctx, doc, opts, num, 0, 0, 0);
+		fz_write_printf(ctx, opts->out, "startxref\n%lu\n%%%%EOF\n", startxref);
 	}
 	fz_always(ctx)
 	{
@@ -2152,9 +2182,9 @@ static void writexrefstream(fz_context *ctx, pdf_document *doc, pdf_write_state
 }
 
 static void
-padto(fz_context *ctx, fz_output *out, fz_off_t target)
+padto(fz_context *ctx, fz_output *out, int64_t target)
 {
-	fz_off_t pos = fz_tell_output(ctx, out);
+	int64_t pos = fz_tell_output(ctx, out);
 
 	assert(pos <= target);
 	while (pos < target)
@@ -2193,7 +2223,7 @@ dowriteobject(fz_context *ctx, pdf_document *doc, pdf_write_state *opts, int num
 		if (!opts->do_incremental || pdf_xref_is_incremental(ctx, doc, num))
 		{
 			opts->ofs_list[num] = fz_tell_output(ctx, opts->out);
-			writeobject(ctx, doc, opts, num, opts->gen_list[num], 1);
+			writeobject(ctx, doc, opts, num, opts->gen_list[num], 1, num == opts->crypt_object_number);
 		}
 	}
 	else
@@ -2209,7 +2239,7 @@ writeobjects(fz_context *ctx, pdf_document *doc, pdf_write_state *opts, int pass
 	if (!opts->do_incremental)
 	{
 		fz_write_printf(ctx, opts->out, "%%PDF-%d.%d\n", doc->version / 10, doc->version % 10);
-		fz_write_string(ctx, opts->out, "%%\316\274\341\277\246\n\n");
+		fz_write_string(ctx, opts->out, "%\xC2\xB5\xC2\xB6\n\n");
 	}
 
 	dowriteobject(ctx, doc, opts, opts->start, pass);
@@ -2228,7 +2258,7 @@ writeobjects(fz_context *ctx, pdf_document *doc, pdf_write_state *opts, int pass
 		dowriteobject(ctx, doc, opts, num, pass);
 	if (opts->do_linear && pass == 1)
 	{
-		fz_off_t offset = (opts->start == 1 ? opts->main_xref_offset : opts->ofs_list[1] + opts->hintstream_len);
+		int64_t offset = (opts->start == 1 ? opts->main_xref_offset : opts->ofs_list[1] + opts->hintstream_len);
 		padto(ctx, opts->out, offset);
 	}
 	for (num = 1; num < opts->start; num++)
@@ -2489,7 +2519,7 @@ make_page_offset_hints(fz_context *ctx, pdf_document *doc, pdf_write_state *opts
 	for (j = 0; j < pop[0]->len; j++)
 	{
 		int o = pop[0]->object[j];
-		fz_off_t min, max;
+		int64_t min, max;
 		min = opts->ofs_list[o];
 		if (o == opts->start-1)
 			max = opts->main_xref_offset;
@@ -2586,106 +2616,113 @@ static void presize_unsaved_signature_byteranges(fz_context *ctx, pdf_document *
 
 				for (i = 0; i < n+1; i++)
 				{
-					pdf_array_push_drop(ctx, byte_range, pdf_new_int(ctx, doc, INT_MAX));
-					pdf_array_push_drop(ctx, byte_range, pdf_new_int(ctx, doc, INT_MAX));
+					pdf_array_push_int(ctx, byte_range, INT_MAX);
+					pdf_array_push_int(ctx, byte_range, INT_MAX);
 				}
 			}
 		}
 	}
 }
 
-static void complete_signatures(fz_context *ctx, pdf_document *doc, pdf_write_state *opts, const char *filename)
+static void complete_signatures(fz_context *ctx, pdf_document *doc, pdf_write_state *opts)
 {
 	pdf_unsaved_sig *usig;
 	char buf[5120];
 	int s;
 	int i;
 	int last_end;
-	FILE *f;
+	fz_stream *stm = NULL;
+	fz_var(stm);
 
-	for (s = 0; s < doc->num_incremental_sections; s++)
+	fz_try(ctx)
 	{
-		pdf_xref *xref = &doc->xref_sections[doc->num_incremental_sections - s - 1];
-
-		if (xref->unsaved_sigs)
+		for (s = 0; s < doc->num_incremental_sections; s++)
 		{
-			pdf_obj *byte_range;
+			pdf_xref *xref = &doc->xref_sections[doc->num_incremental_sections - s - 1];
 
-			f = fz_fopen(filename, "rb+");
-			if (!f)
-				fz_throw(ctx, FZ_ERROR_GENERIC, "Failed to open %s to complete signatures", filename);
-
-			/* Locate the byte ranges and contents in the saved file */
-			for (usig = xref->unsaved_sigs; usig; usig = usig->next)
+			if (xref->unsaved_sigs)
 			{
-				char *bstr, *cstr, *fstr;
-				int pnum = pdf_obj_parent_num(ctx, pdf_dict_getl(ctx, usig->field, PDF_NAME_V, PDF_NAME_ByteRange, NULL));
-				fz_fseek(f, opts->ofs_list[pnum], SEEK_SET);
-				(void)fread(buf, 1, sizeof(buf), f);
-				buf[sizeof(buf)-1] = 0;
+				pdf_obj *byte_range;
 
-				bstr = strstr(buf, "/ByteRange");
-				cstr = strstr(buf, "/Contents");
-				fstr = strstr(buf, "/Filter");
-
-				if (bstr && cstr && fstr && bstr < cstr && cstr < fstr)
+				stm = fz_stream_from_output(ctx, opts->out);
+				/* Locate the byte ranges and contents in the saved file */
+				for (usig = xref->unsaved_sigs; usig; usig = usig->next)
 				{
-					usig->byte_range_start = bstr - buf + 10 + opts->ofs_list[pnum];
-					usig->byte_range_end = cstr - buf + opts->ofs_list[pnum];
-					usig->contents_start = cstr - buf + 9 + opts->ofs_list[pnum];
-					usig->contents_end = fstr - buf + opts->ofs_list[pnum];
+					char *bstr, *cstr, *fstr;
+					int pnum = pdf_obj_parent_num(ctx, pdf_dict_getl(ctx, usig->field, PDF_NAME_V, PDF_NAME_ByteRange, NULL));
+					fz_seek(ctx, stm, opts->ofs_list[pnum], SEEK_SET);
+					(void)fz_read(ctx, stm, (unsigned char *)buf, sizeof(buf));
+					buf[sizeof(buf)-1] = 0;
+
+					bstr = strstr(buf, "/ByteRange");
+					cstr = strstr(buf, "/Contents");
+					fstr = strstr(buf, "/Filter");
+
+					if (bstr && cstr && fstr && bstr < cstr && cstr < fstr)
+					{
+						usig->byte_range_start = bstr - buf + 10 + opts->ofs_list[pnum];
+						usig->byte_range_end = cstr - buf + opts->ofs_list[pnum];
+						usig->contents_start = cstr - buf + 9 + opts->ofs_list[pnum];
+						usig->contents_end = fstr - buf + opts->ofs_list[pnum];
+					}
 				}
-			}
 
-			/* Recreate ByteRange with correct values. Initially store the
-			* recreated object in the first of the unsaved signatures */
-			byte_range = pdf_new_array(ctx, doc, 4);
-			pdf_dict_putl_drop(ctx, xref->unsaved_sigs->field, byte_range, PDF_NAME_V, PDF_NAME_ByteRange, NULL);
+				fz_drop_stream(ctx, stm);
+				stm = NULL;
 
-			last_end = 0;
-			for (usig = xref->unsaved_sigs; usig; usig = usig->next)
-			{
-				pdf_array_push_drop(ctx, byte_range, pdf_new_int(ctx, doc, last_end));
-				pdf_array_push_drop(ctx, byte_range, pdf_new_int(ctx, doc, usig->contents_start - last_end));
-				last_end = usig->contents_end;
-			}
-			pdf_array_push_drop(ctx, byte_range, pdf_new_int(ctx, doc, last_end));
-			pdf_array_push_drop(ctx, byte_range, pdf_new_int(ctx, doc, xref->end_ofs - last_end));
+				/* Recreate ByteRange with correct values. Initially store the
+				* recreated object in the first of the unsaved signatures */
+				byte_range = pdf_new_array(ctx, doc, 4);
+				pdf_dict_putl_drop(ctx, xref->unsaved_sigs->field, byte_range, PDF_NAME_V, PDF_NAME_ByteRange, NULL);
 
-			/* Copy the new ByteRange to the other unsaved signatures */
-			for (usig = xref->unsaved_sigs->next; usig; usig = usig->next)
-				pdf_dict_putl_drop(ctx, usig->field, pdf_copy_array(ctx, byte_range), PDF_NAME_V, PDF_NAME_ByteRange, NULL);
+				last_end = 0;
+				for (usig = xref->unsaved_sigs; usig; usig = usig->next)
+				{
+					pdf_array_push_int(ctx, byte_range, last_end);
+					pdf_array_push_int(ctx, byte_range, usig->contents_start - last_end);
+					last_end = usig->contents_end;
+				}
+				pdf_array_push_int(ctx, byte_range, last_end);
+				pdf_array_push_int(ctx, byte_range, xref->end_ofs - last_end);
 
-			/* Write the byte range into buf, padding with spaces*/
-			i = pdf_sprint_obj(ctx, buf, sizeof(buf), byte_range, 1);
-			memset(buf+i, ' ', sizeof(buf)-i);
+				/* Copy the new ByteRange to the other unsaved signatures */
+				for (usig = xref->unsaved_sigs->next; usig; usig = usig->next)
+					pdf_dict_putl_drop(ctx, usig->field, pdf_copy_array(ctx, byte_range), PDF_NAME_V, PDF_NAME_ByteRange, NULL);
 
-			/* Write the byte range to the file */
-			for (usig = xref->unsaved_sigs; usig; usig = usig->next)
-			{
-				fz_fseek(f, usig->byte_range_start, SEEK_SET);
-				fwrite(buf, 1, usig->byte_range_end - usig->byte_range_start, f);
-			}
+				/* Write the byte range into buf, padding with spaces*/
+				i = pdf_sprint_obj(ctx, buf, sizeof(buf), byte_range, 1);
+				memset(buf+i, ' ', sizeof(buf)-i);
 
-			fclose(f);
+				/* Write the byte range to the file */
+				for (usig = xref->unsaved_sigs; usig; usig = usig->next)
+				{
+					fz_seek_output(ctx, opts->out, usig->byte_range_start, SEEK_SET);
+					fz_write_data(ctx, opts->out, buf, usig->byte_range_end - usig->byte_range_start);
+				}
 
-			/* Write the digests into the file */
-			for (usig = xref->unsaved_sigs; usig; usig = usig->next)
-				pdf_write_digest(ctx, doc, filename, byte_range, usig->contents_start, usig->contents_end - usig->contents_start, usig->signer);
+				/* Write the digests into the file */
+				for (usig = xref->unsaved_sigs; usig; usig = usig->next)
+					pdf_write_digest(ctx, opts->out, byte_range, usig->contents_start, usig->contents_end - usig->contents_start, usig->signer);
 
-			/* delete the unsaved_sigs records */
-			while ((usig = xref->unsaved_sigs) != NULL)
-			{
-				xref->unsaved_sigs = usig->next;
-				pdf_drop_obj(ctx, usig->field);
-				pdf_drop_signer(ctx, usig->signer);
-				fz_free(ctx, usig);
+				/* delete the unsaved_sigs records */
+				while ((usig = xref->unsaved_sigs) != NULL)
+				{
+					xref->unsaved_sigs = usig->next;
+					pdf_drop_obj(ctx, usig->field);
+					usig->signer->drop(usig->signer);
+					fz_free(ctx, usig);
+				}
 			}
 		}
 	}
+	fz_catch(ctx)
+	{
+		fz_drop_stream(ctx, stm);
+		fz_rethrow(ctx);
+	}
 }
 
-static void sanitize(fz_context *ctx, pdf_document *doc, int ascii)
+static void clean_content_streams(fz_context *ctx, pdf_document *doc, int sanitize, int ascii)
 {
 	int n = pdf_count_pages(ctx, doc);
 	int i;
@@ -2694,11 +2731,11 @@ static void sanitize(fz_context *ctx, pdf_document *doc, int ascii)
 	{
 		pdf_annot *annot;
 		pdf_page *page = pdf_load_page(ctx, doc, i);
-		pdf_clean_page_contents(ctx, doc, page, NULL, NULL, NULL, ascii);
+		pdf_clean_page_contents(ctx, doc, page, NULL, NULL, NULL, sanitize, ascii);
 
 		for (annot = pdf_first_annot(ctx, page); annot != NULL; annot = pdf_next_annot(ctx, annot))
 		{
-			pdf_clean_annot_contents(ctx, doc, annot, NULL, NULL, NULL, ascii);
+			pdf_clean_annot_contents(ctx, doc, annot, NULL, NULL, NULL, sanitize, ascii);
 		}
 
 		fz_drop_page(ctx, &page->super);
@@ -2709,7 +2746,6 @@ static void sanitize(fz_context *ctx, pdf_document *doc, int ascii)
  * pdf_write_options, passed into pdf_save_document */
 static void initialise_write_state(fz_context *ctx, pdf_document *doc, const pdf_write_options *in_opts, pdf_write_state *opts)
 {
-	int num;
 	int xref_len = pdf_xref_len(ctx, doc);
 
 	opts->do_incremental = in_opts->do_incremental;
@@ -2729,21 +2765,16 @@ static void initialise_write_state(fz_context *ctx, pdf_document *doc, const pdf
 	/* We deliberately make these arrays long enough to cope with
 	* 1 to n access rather than 0..n-1, and add space for 2 new
 	* extra entries that may be required for linearization. */
-	opts->use_list = fz_malloc_array(ctx, xref_len + 3, sizeof(int));
-	opts->ofs_list = fz_malloc_array(ctx, xref_len + 3, sizeof(fz_off_t));
-	opts->gen_list = fz_calloc(ctx, xref_len + 3, sizeof(int));
-	opts->renumber_map = fz_malloc_array(ctx, xref_len + 3, sizeof(int));
-	opts->rev_renumber_map = fz_malloc_array(ctx, xref_len + 3, sizeof(int));
+	opts->list_len = 0;
+	opts->use_list = NULL;
+	opts->ofs_list = NULL;
+	opts->gen_list = NULL;
+	opts->renumber_map = NULL;
+	opts->rev_renumber_map = NULL;
 	opts->continue_on_error = in_opts->continue_on_error;
 	opts->errors = in_opts->errors;
 
-	for (num = 0; num < xref_len; num++)
-	{
-		opts->use_list[num] = 0;
-		opts->ofs_list[num] = 0;
-		opts->renumber_map[num] = num;
-		opts->rev_renumber_map[num] = num;
-	}
+	expand_lists(ctx, opts, xref_len);
 }
 
 /* Free the resources held by the dynamic write options */
@@ -2775,8 +2806,11 @@ const char *fz_pdf_write_options_usage =
 	"\tascii: ASCII hex encode binary streams\n"
 	"\tpretty: pretty-print objects with indentation\n"
 	"\tlinearize: optimize for web browsers\n"
-	"\tsanitize: clean up graphics commands in content streams\n"
+	"\tclean: pretty-print graphics commands in content streams\n"
+	"\tsanitize: sanitize graphics commands in content streams\n"
 	"\tgarbage: garbage collect unused objects\n"
+	"\tincremental: write changes as incremental update\n"
+	"\tcontinue-on-error: continue saving the document even if there is an error\n"
 	"\tor garbage=compact: ... and compact cross reference table\n"
 	"\tor garbage=deduplicate: ... and remove duplicate objects\n"
 	"\n";
@@ -2802,8 +2836,10 @@ pdf_parse_write_options(fz_context *ctx, pdf_write_options *opts, const char *ar
 		opts->do_pretty = fz_option_eq(val, "yes");
 	if (fz_has_option(ctx, args, "linearize", &val))
 		opts->do_linear = fz_option_eq(val, "yes");
-	if (fz_has_option(ctx, args, "sanitize", &val))
+	if (fz_has_option(ctx, args, "clean", &val))
 		opts->do_clean = fz_option_eq(val, "yes");
+	if (fz_has_option(ctx, args, "sanitize", &val))
+		opts->do_sanitize = fz_option_eq(val, "yes");
 	if (fz_has_option(ctx, args, "incremental", &val))
 		opts->do_incremental = fz_option_eq(val, "yes");
 	if (fz_has_option(ctx, args, "continue-on-error", &val))
@@ -2817,7 +2853,7 @@ pdf_parse_write_options(fz_context *ctx, pdf_write_options *opts, const char *ar
 		else if (fz_option_eq(val, "deduplicate"))
 			opts->do_garbage = 3;
 		else
-			opts->do_garbage = atoi(val);
+			opts->do_garbage = fz_atoi(val);
 	}
 
 	return opts;
@@ -2837,14 +2873,31 @@ prepare_for_save(fz_context *ctx, pdf_document *doc, pdf_write_options *in_opts)
 {
 	doc->freeze_updates = 1;
 
-	/* Sanitize the operator streams */
-	if (in_opts->do_clean)
-		sanitize(ctx, doc, in_opts->do_ascii);
+	/* Rewrite (and possibly sanitize) the operator streams */
+	if (in_opts->do_clean || in_opts->do_sanitize)
+		clean_content_streams(ctx, doc, in_opts->do_sanitize, in_opts->do_ascii);
 
 	pdf_finish_edit(ctx, doc);
 	presize_unsaved_signature_byteranges(ctx, doc);
 }
 
+static void
+change_identity(fz_context *ctx, pdf_document *doc)
+{
+	pdf_obj *identity = pdf_dict_get(ctx, pdf_trailer(ctx, doc), PDF_NAME_ID);
+	pdf_obj *str;
+	unsigned char rnd[16];
+
+	if (pdf_array_len(ctx, identity) < 2)
+		return;
+
+	/* Maybe recalculate this in future. For now, just change the second one. */
+	fz_memrnd(ctx, rnd, 16);
+	str = pdf_new_string(ctx, doc, (char *)rnd, 16);
+	pdf_array_put_drop(ctx, identity, 1, str);
+
+}
+
 static void
 do_pdf_save_document(fz_context *ctx, pdf_document *doc, pdf_write_state *opts, pdf_write_options *in_opts)
 {
@@ -2875,14 +2928,21 @@ do_pdf_save_document(fz_context *ctx, pdf_document *doc, pdf_write_state *opts,
 		{
 			pdf_ensure_solid_xref(ctx, doc, xref_len);
 			preloadobjstms(ctx, doc);
+			change_identity(ctx, doc);
+			xref_len = pdf_xref_len(ctx, doc); /* May have changed due to repair */
+			expand_lists(ctx, opts, xref_len);
 		}
 
 		/* Sweep & mark objects from the trailer */
 		if (opts->do_garbage >= 1 || opts->do_linear)
 			(void)markobj(ctx, doc, opts, pdf_trailer(ctx, doc));
 		else
+		{
+			xref_len = pdf_xref_len(ctx, doc); /* May have changed due to repair */
+			expand_lists(ctx, opts, xref_len);
 			for (num = 0; num < xref_len; num++)
 				opts->use_list[num] = 1;
+		}
 
 		/* Coalesce and renumber duplicate objects */
 		if (opts->do_garbage >= 3)
@@ -2892,14 +2952,26 @@ do_pdf_save_document(fz_context *ctx, pdf_document *doc, pdf_write_state *opts,
 		if (opts->do_garbage >= 2 || opts->do_linear)
 			compactxref(ctx, doc, opts);
 
+		opts->crypt_object_number = 0;
+		if (doc->crypt)
+		{
+			pdf_obj *crypt = pdf_dict_get(ctx, pdf_trailer(ctx, doc), PDF_NAME_Encrypt);
+			int crypt_num = pdf_to_num(ctx, crypt);
+			opts->crypt_object_number = opts->renumber_map[crypt_num];
+		}
+
 		/* Make renumbering affect all indirect references and update xref */
 		if (opts->do_garbage >= 2 || opts->do_linear)
 			renumberobjs(ctx, doc, opts);
 
 		/* Truncate the xref after compacting and renumbering */
 		if ((opts->do_garbage >= 2 || opts->do_linear) && !opts->do_incremental)
+		{
+			xref_len = pdf_xref_len(ctx, doc); /* May have changed due to repair */
+			expand_lists(ctx, opts, xref_len);
 			while (xref_len > 0 && !opts->use_list[xref_len-1])
 				xref_len--;
+		}
 
 		if (opts->do_linear)
 			linearize(ctx, doc, opts);
@@ -2962,7 +3034,7 @@ do_pdf_save_document(fz_context *ctx, pdf_document *doc, pdf_write_state *opts,
 				}
 			}
 
-			if (opts->do_linear)
+			if (opts->do_linear && opts->page_count > 0)
 			{
 				opts->main_xref_offset = fz_tell_output(ctx, opts->out);
 				writexref(ctx, doc, opts, 0, opts->start, 0, 0, opts->first_xref_offset);
@@ -2992,6 +3064,8 @@ do_pdf_save_document(fz_context *ctx, pdf_document *doc, pdf_write_state *opts,
 			doc->xref_sections[0].end_ofs = fz_tell_output(ctx, opts->out);
 		}
 
+		complete_signatures(ctx, doc, opts);
+
 		doc->dirty = 0;
 	}
 	fz_always(ctx)
@@ -3083,15 +3157,10 @@ void pdf_save_document(fz_context *ctx, pdf_document *doc, const char *filename,
 	{
 		opts.out = fz_new_output_with_path(ctx, filename, 0);
 	}
-
-	if (!opts.out)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "cannot open output file '%s'", filename);
-
 	fz_try(ctx)
 	{
 		do_pdf_save_document(ctx, doc, &opts, in_opts);
-
-		complete_signatures(ctx, doc, &opts, filename);
+		fz_close_output(ctx, opts.out);
 	}
 	fz_always(ctx)
 	{
@@ -3131,8 +3200,8 @@ make_page_tree_node(fz_context *ctx, pdf_document *doc, int l, int r, pdf_obj *p
 	fz_try(ctx)
 	{
 		me = pdf_new_dict(ctx, doc, 2);
-		pdf_dict_put_drop(ctx, me, PDF_NAME_Type, PDF_NAME_Pages);
-		pdf_dict_put_drop(ctx, me, PDF_NAME_Count, pdf_new_int(ctx, doc, r-l));
+		pdf_dict_put(ctx, me, PDF_NAME_Type, PDF_NAME_Pages);
+		pdf_dict_put_int(ctx, me, PDF_NAME_Count, r-l);
 		if (!root)
 			pdf_dict_put(ctx, me, PDF_NAME_Parent, parent_ref);
 		a = pdf_new_array(ctx, doc, KIDS_PER_LEVEL);
