@@ -3,9 +3,31 @@
  */
 
 #include "mupdf/fitz.h"
-#include "mupdf/pdf.h" /* for pdf output */
 
+#if FZ_ENABLE_PDF
+#include "mupdf/pdf.h" /* for pdf output */
+#endif
+
+#ifndef DISABLE_MUTHREADS
 #include "mupdf/helpers/mu-threads.h"
+#endif
+
+#include <string.h>
+#include <stdlib.h>
+#include <stdio.h>
+#ifdef _MSC_VER
+struct timeval;
+struct timezone;
+int gettimeofday(struct timeval *tv, struct timezone *tz);
+#else
+#include <sys/time.h>
+#endif
+
+/* Allow for windows stdout being made binary */
+#ifdef _WIN32
+#include <io.h>
+#include <fcntl.h>
+#endif
 
 /* Enable for helpful threading debug */
 /* #define DEBUG_THREADS(A) do { printf A; fflush(stdout); } while (0) */
@@ -14,8 +36,8 @@
 enum {
 	OUT_NONE,
 	OUT_PNG, OUT_TGA, OUT_PNM, OUT_PGM, OUT_PPM, OUT_PAM,
-	OUT_PBM, OUT_PKM, OUT_PWG, OUT_PCL, OUT_PS,
-	OUT_TEXT, OUT_HTML, OUT_STEXT,
+	OUT_PBM, OUT_PKM, OUT_PWG, OUT_PCL, OUT_PS, OUT_PSD,
+	OUT_TEXT, OUT_HTML, OUT_XHTML, OUT_STEXT, OUT_PCLM,
 	OUT_TRACE, OUT_SVG,
 #if FZ_ENABLE_PDF
 	OUT_PDF,
@@ -23,39 +45,45 @@ enum {
 	OUT_GPROOF
 };
 
-enum { CS_INVALID, CS_UNSET, CS_MONO, CS_GRAY, CS_GRAY_ALPHA, CS_RGB, CS_RGB_ALPHA, CS_CMYK, CS_CMYK_ALPHA };
+enum { CS_INVALID, CS_UNSET, CS_MONO, CS_GRAY, CS_GRAY_ALPHA, CS_RGB, CS_RGB_ALPHA, CS_CMYK, CS_CMYK_ALPHA, CS_ICC };
+
+enum { SPOTS_NONE, SPOTS_OVERPRINT_SIM, SPOTS_FULL };
 
 typedef struct
 {
 	char *suffix;
 	int format;
+	int spots;
 } suffix_t;
 
 static const suffix_t suffix_table[] =
 {
-	{ ".png", OUT_PNG },
-	{ ".pgm", OUT_PGM },
-	{ ".ppm", OUT_PPM },
-	{ ".pnm", OUT_PNM },
-	{ ".pam", OUT_PAM },
-	{ ".pbm", OUT_PBM },
-	{ ".pkm", OUT_PKM },
-	{ ".svg", OUT_SVG },
-	{ ".pwg", OUT_PWG },
-	{ ".pcl", OUT_PCL },
-	{ ".ps", OUT_PS },
+	{ ".png", OUT_PNG, 0 },
+	{ ".pgm", OUT_PGM, 0 },
+	{ ".ppm", OUT_PPM, 0 },
+	{ ".pnm", OUT_PNM, 0 },
+	{ ".pam", OUT_PAM, 0 },
+	{ ".pbm", OUT_PBM, 0 },
+	{ ".pkm", OUT_PKM, 0 },
+	{ ".svg", OUT_SVG, 0 },
+	{ ".pwg", OUT_PWG, 0 },
+	{ ".pclm", OUT_PCLM, 0 },
+	{ ".pcl", OUT_PCL, 0 },
 #if FZ_ENABLE_PDF
-	{ ".pdf", OUT_PDF },
+	{ ".pdf", OUT_PDF, 0 },
 #endif
-	{ ".tga", OUT_TGA },
+	{ ".psd", OUT_PSD, 1 },
+	{ ".ps", OUT_PS, 0 },
+	{ ".tga", OUT_TGA, 0 },
 
-	{ ".txt", OUT_TEXT },
-	{ ".text", OUT_TEXT },
-	{ ".html", OUT_HTML },
-	{ ".stext", OUT_STEXT },
+	{ ".txt", OUT_TEXT, 0 },
+	{ ".text", OUT_TEXT, 0 },
+	{ ".html", OUT_HTML, 0 },
+	{ ".xhtml", OUT_XHTML, 0 },
+	{ ".stext", OUT_STEXT, 0 },
 
-	{ ".trace", OUT_TRACE },
-	{ ".gproof", OUT_GPROOF },
+	{ ".trace", OUT_TRACE, 0 },
+	{ ".gproof", OUT_GPROOF, 0 },
 };
 
 typedef struct
@@ -86,12 +114,12 @@ typedef struct
 {
 	int format;
 	int default_cs;
-	int permitted_cs[6];
+	int permitted_cs[7];
 } format_cs_table_t;
 
 static const format_cs_table_t format_cs_table[] =
 {
-	{ OUT_PNG, CS_RGB, { CS_GRAY, CS_GRAY_ALPHA, CS_RGB, CS_RGB_ALPHA } },
+	{ OUT_PNG, CS_RGB, { CS_GRAY, CS_GRAY_ALPHA, CS_RGB, CS_RGB_ALPHA, CS_ICC } },
 	{ OUT_PPM, CS_RGB, { CS_GRAY, CS_RGB } },
 	{ OUT_PNM, CS_GRAY, { CS_GRAY, CS_RGB } },
 	{ OUT_PAM, CS_RGB_ALPHA, { CS_GRAY, CS_GRAY_ALPHA, CS_RGB, CS_RGB_ALPHA, CS_CMYK, CS_CMYK_ALPHA } },
@@ -100,7 +128,9 @@ static const format_cs_table_t format_cs_table[] =
 	{ OUT_PKM, CS_CMYK, { CS_CMYK } },
 	{ OUT_PWG, CS_RGB, { CS_MONO, CS_GRAY, CS_RGB, CS_CMYK } },
 	{ OUT_PCL, CS_MONO, { CS_MONO, CS_RGB } },
+	{ OUT_PCLM, CS_RGB, { CS_RGB, CS_GRAY } },
 	{ OUT_PS, CS_RGB, { CS_GRAY, CS_RGB, CS_CMYK } },
+	{ OUT_PSD, CS_CMYK, { CS_GRAY, CS_GRAY_ALPHA, CS_RGB, CS_RGB_ALPHA, CS_CMYK, CS_CMYK_ALPHA, CS_ICC } },
 	{ OUT_TGA, CS_RGB, { CS_GRAY, CS_GRAY_ALPHA, CS_RGB, CS_RGB_ALPHA } },
 
 	{ OUT_TRACE, CS_RGB, { CS_RGB } },
@@ -112,6 +142,7 @@ static const format_cs_table_t format_cs_table[] =
 
 	{ OUT_TEXT, CS_RGB, { CS_RGB } },
 	{ OUT_HTML, CS_RGB, { CS_RGB } },
+	{ OUT_XHTML, CS_RGB, { CS_RGB } },
 	{ OUT_STEXT, CS_RGB, { CS_RGB } },
 };
 
@@ -176,15 +207,16 @@ typedef struct worker_t {
 	fz_pixmap *pix;
 	fz_bitmap *bit;
 	fz_cookie cookie;
+#ifndef DISABLE_MUTHREADS
 	mu_semaphore start;
 	mu_semaphore stop;
 	mu_thread thread;
+#endif
 } worker_t;
 
 static char *output = NULL;
-fz_output *out = NULL;
+static fz_output *out = NULL;
 static int output_pagenum = 0;
-static int output_append = 0;
 static int output_file_per_page = 0;
 
 static char *format = NULL;
@@ -206,9 +238,6 @@ static float min_line_width = 0.0f;
 
 static int showfeatures = 0;
 static int showtime = 0;
-static size_t memtrace_current = 0;
-static size_t memtrace_peak = 0;
-static size_t memtrace_total = 0;
 static int showmemory = 0;
 static int showmd5 = 0;
 
@@ -222,19 +251,34 @@ static int alphabits_text = 8;
 static int alphabits_graphics = 8;
 
 static int out_cs = CS_UNSET;
+static const char *proof_filename = NULL;
+fz_colorspace *proof_cs = NULL;
+static const char *icc_filename = NULL;
 static float gamma_value = 1;
 static int invert = 0;
 static int band_height = 0;
 static int lowmemory = 0;
 
 static int errored = 0;
-static fz_stext_sheet *sheet = NULL;
 static fz_colorspace *colorspace;
+static fz_colorspace *oi = NULL;
+#ifdef FZ_ENABLE_SPOT_RENDERING
+static int spots = SPOTS_OVERPRINT_SIM;
+#else
+static int spots = SPOTS_NONE;
+#endif
 static int alpha;
 static char *filename;
 static int files = 0;
 static int num_workers = 0;
 static worker_t *workers;
+static fz_band_writer *bander = NULL;
+
+#ifdef NO_ICC
+static fz_cmm_engine *icc_engine = NULL;
+#else
+static fz_cmm_engine *icc_engine = &fz_cmm_engine_lcms;
+#endif
 
 static const char *layer_config = NULL;
 
@@ -242,14 +286,17 @@ static struct {
 	int active;
 	int started;
 	fz_context *ctx;
+#ifndef DISABLE_MUTHREADS
 	mu_thread thread;
 	mu_semaphore start;
 	mu_semaphore stop;
+#endif
 	int pagenum;
 	char *filename;
 	fz_display_list *list;
 	fz_page *page;
 	int interptime;
+	fz_separations *seps;
 } bgprint;
 
 static struct {
@@ -285,9 +332,11 @@ static void usage(void)
 		"\t-w -\twidth (in pixels) (maximum width if -r is specified)\n"
 		"\t-h -\theight (in pixels) (maximum height if -r is specified)\n"
 		"\t-f -\tfit width and/or height exactly; ignore original aspect ratio\n"
-		"\t-B -\tmaximum band_height (pgm, ppm, pam, png output only)\n"
-#ifdef MUDRAW_THREADS
+		"\t-B -\tmaximum band_height (pXm, pcl, pclm, ps, psd and png output only)\n"
+#ifndef DISABLE_MUTHREADS
 		"\t-T -\tnumber of threads to use for rendering (banded mode only)\n"
+#else
+		"\t-T -\tnumber of threads to use for rendering (disabled in this non-threading build)\n"
 #endif
 		"\n"
 		"\t-W -\tpage width for EPUB layout\n"
@@ -296,7 +345,8 @@ static void usage(void)
 		"\t-U -\tfile name of user stylesheet for EPUB layout\n"
 		"\t-X\tdisable document styles for EPUB layout\n"
 		"\n"
-		"\t-c -\tcolorspace (mono, gray, grayalpha, rgb, rgba, cmyk, cmykalpha)\n"
+		"\t-c -\tcolorspace (mono, gray, grayalpha, rgb, rgba, cmyk, cmykalpha, filename of ICC profile)\n"
+		"\t-e -\tproof icc profile (filename of ICC profile)\n"
 		"\t-G -\tapply gamma correction\n"
 		"\t-I\tinvert colors\n"
 		"\n"
@@ -306,7 +356,22 @@ static void usage(void)
 		"\t-D\tdisable use of display list\n"
 		"\t-i\tignore errors\n"
 		"\t-L\tlow memory mode (avoid caching, clear objects after each page)\n"
+#ifndef DISABLE_MUTHREADS
 		"\t-P\tparallel interpretation/rendering\n"
+#else
+		"\t-P\tparallel interpretation/rendering (disabled in this non-threading build)\n"
+#endif
+		"\t-N\tdisable ICC workflow (\"N\"o color management)\n"
+		"\t-O -\tControl spot/overprint rendering\n"
+#ifdef FZ_ENABLE_SPOT_RENDERING
+		"\t\t 0 = No spot rendering\n"
+		"\t\t 1 = Overprint simulation (default)\n"
+		"\t\t 2 = Full spot rendering\n"
+#else
+		"\t\t 0 = No spot rendering (default)\n"
+		"\t\t 1 = Overprint simulation (Disabled in this build)\n"
+		"\t\t 2 = Full spot rendering (Disabled in this build)\n"
+#endif
 		"\n"
 		"\t-y l\tList the layer configs to stderr\n"
 		"\t-y -\tSelect layer config (by number)\n"
@@ -354,22 +419,10 @@ file_level_headers(fz_context *ctx)
 	if (output_format == OUT_STEXT || output_format == OUT_TRACE)
 		fz_write_printf(ctx, out, "<?xml version=\"1.0\"?>\n");
 
-	if (output_format == OUT_TEXT || output_format == OUT_HTML || output_format == OUT_STEXT)
-		sheet = fz_new_stext_sheet(ctx);
-
 	if (output_format == OUT_HTML)
-	{
-		fz_write_printf(ctx, out, "<style>\n");
-		fz_write_printf(ctx, out, "body{background-color:gray;margin:12pt;}\n");
-		fz_write_printf(ctx, out, "div.page{background-color:white;margin:6pt;padding:6pt;}\n");
-		fz_write_printf(ctx, out, "div.block{border:1px solid gray;margin:6pt;padding:6pt;}\n");
-		fz_write_printf(ctx, out, "div.metaline{display:table;width:100%%}\n");
-		fz_write_printf(ctx, out, "div.line{display:table-row;padding:6pt}\n");
-		fz_write_printf(ctx, out, "div.cell{display:table-cell;padding-left:6pt;padding-right:6pt}\n");
-		fz_write_printf(ctx, out, "p{margin:0pt;padding:0pt;}\n");
-		fz_write_printf(ctx, out, "</style>\n");
-		fz_write_printf(ctx, out, "<body>\n");
-	}
+		fz_print_stext_header_as_html(ctx, out);
+	if (output_format == OUT_XHTML)
+		fz_print_stext_header_as_xhtml(ctx, out);
 
 	if (output_format == OUT_STEXT || output_format == OUT_TRACE)
 		fz_write_printf(ctx, out, "<document name=\"%s\">\n", filename);
@@ -379,6 +432,14 @@ file_level_headers(fz_context *ctx)
 
 	if (output_format == OUT_PWG)
 		fz_write_pwg_file_header(ctx, out);
+
+	if (output_format == OUT_PCLM)
+	{
+		fz_pclm_options opts = { 0 };
+		fz_parse_pclm_options(ctx, &opts, "compression=flate");
+		bander = fz_new_pclm_band_writer(ctx, out, &opts);
+	}
+
 }
 
 static void
@@ -388,23 +449,24 @@ file_level_trailers(fz_context *ctx)
 		fz_write_printf(ctx, out, "</document>\n");
 
 	if (output_format == OUT_HTML)
-	{
-		fz_write_printf(ctx, out, "</body>\n");
-		fz_write_printf(ctx, out, "<style>\n");
-		fz_print_stext_sheet(ctx, out, sheet);
-		fz_write_printf(ctx, out, "</style>\n");
-	}
+		fz_print_stext_trailer_as_html(ctx, out);
+	if (output_format == OUT_XHTML)
+		fz_print_stext_trailer_as_xhtml(ctx, out);
 
 	if (output_format == OUT_PS)
 		fz_write_ps_file_trailer(ctx, out, output_pagenum);
 
-	fz_drop_stext_sheet(ctx, sheet);
+	if (output_format == OUT_PCLM)
+		fz_drop_band_writer(ctx, bander);
+
 }
 
 static void drawband(fz_context *ctx, fz_page *page, fz_display_list *list, const fz_matrix *ctm, const fz_rect *tbounds, fz_cookie *cookie, int band_start, fz_pixmap *pix, fz_bitmap **bit)
 {
 	fz_device *dev = NULL;
 
+	fz_var(dev);
+
 	*bit = NULL;
 
 	fz_try(ctx)
@@ -414,7 +476,7 @@ static void drawband(fz_context *ctx, fz_page *page, fz_display_list *list, cons
 		else
 			fz_clear_pixmap_with_value(ctx, pix, 255);
 
-		dev = fz_new_draw_device(ctx, NULL, pix);
+		dev = fz_new_draw_device_with_proof(ctx, NULL, pix, proof_cs);
 		if (lowmemory)
 			fz_enable_device_hints(ctx, dev, FZ_NO_CACHE);
 		if (alphabits_graphics == 0)
@@ -432,9 +494,6 @@ static void drawband(fz_context *ctx, fz_page *page, fz_display_list *list, cons
 		if (gamma_value != 1)
 			fz_gamma_pixmap(ctx, pix, gamma_value);
 
-		if (pix->alpha)
-			fz_unmultiply_pixmap(ctx, pix);
-
 		if (((output_format == OUT_PCL || output_format == OUT_PWG) && out_cs == CS_MONO) || (output_format == OUT_PBM) || (output_format == OUT_PKM))
 			*bit = fz_new_bitmap_from_pixmap_band(ctx, pix, NULL, band_start);
 	}
@@ -445,12 +504,29 @@ static void drawband(fz_context *ctx, fz_page *page, fz_display_list *list, cons
 	}
 }
 
-static void dodrawpage(fz_context *ctx, fz_page *page, fz_display_list *list, int pagenum, fz_cookie *cookie, int start, int interptime, char *filename, int bg)
+static void dodrawpage(fz_context *ctx, fz_page *page, fz_display_list *list, int pagenum, fz_cookie *cookie, int start, int interptime, char *filename, int bg, fz_separations *seps)
 {
 	fz_rect mediabox;
 	fz_device *dev = NULL;
 
-	fz_bound_page(ctx, page, &mediabox);
+	fz_var(dev);
+
+	if (output_file_per_page)
+		file_level_headers(ctx);
+
+	fz_try(ctx)
+	{
+		if (list)
+			fz_bound_display_list(ctx, list, &mediabox);
+		else
+			fz_bound_page(ctx, page, &mediabox);
+	}
+	fz_catch(ctx)
+	{
+		fz_drop_separations(ctx, seps);
+		fz_drop_page(ctx, page);
+		fz_rethrow(ctx);
+	}
 
 	if (output_format == OUT_TRACE)
 	{
@@ -471,66 +547,70 @@ static void dodrawpage(fz_context *ctx, fz_page *page, fz_display_list *list, in
 		fz_always(ctx)
 		{
 			fz_drop_device(ctx, dev);
-			dev = NULL;
 		}
 		fz_catch(ctx)
 		{
 			fz_drop_display_list(ctx, list);
+			fz_drop_separations(ctx, seps);
 			fz_drop_page(ctx, page);
 			fz_rethrow(ctx);
 		}
 	}
 
-	else if (output_format == OUT_TEXT || output_format == OUT_HTML || output_format == OUT_STEXT)
+	else if (output_format == OUT_TEXT || output_format == OUT_HTML || output_format == OUT_XHTML || output_format == OUT_STEXT)
 	{
 		fz_stext_page *text = NULL;
+		float zoom;
+		fz_matrix ctm;
+
+		zoom = resolution / 72;
+		fz_pre_scale(fz_rotate(&ctm, rotation), zoom, zoom);
 
 		fz_var(text);
 
 		fz_try(ctx)
 		{
-			fz_rect mediabox;
-			if (list)
-				fz_bound_display_list(ctx, list, &mediabox);
-			else
-				fz_bound_page(ctx, page, &mediabox);
+			fz_stext_options stext_options;
+
+			stext_options.flags = (output_format == OUT_HTML || output_format == OUT_XHTML) ? FZ_STEXT_PRESERVE_IMAGES : 0;
 			text = fz_new_stext_page(ctx, &mediabox);
-			dev = fz_new_stext_device(ctx, sheet, text, 0);
+			dev = fz_new_stext_device(ctx,  text, &stext_options);
 			if (lowmemory)
 				fz_enable_device_hints(ctx, dev, FZ_NO_CACHE);
-			if (output_format == OUT_HTML)
-				fz_disable_device_hints(ctx, dev, FZ_IGNORE_IMAGE);
 			if (list)
-				fz_run_display_list(ctx, list, dev, &fz_identity, &fz_infinite_rect, cookie);
+				fz_run_display_list(ctx, list, dev, &ctm, &fz_infinite_rect, cookie);
 			else
-				fz_run_page(ctx, page, dev, &fz_identity, cookie);
+				fz_run_page(ctx, page, dev, &ctm, cookie);
 			fz_close_device(ctx, dev);
 			fz_drop_device(ctx, dev);
 			dev = NULL;
 			if (output_format == OUT_STEXT)
 			{
-				fz_print_stext_page_xml(ctx, out, text);
+				fz_print_stext_page_as_xml(ctx, out, text);
 			}
 			else if (output_format == OUT_HTML)
 			{
-				fz_analyze_text(ctx, sheet, text);
-				fz_print_stext_page_html(ctx, out, text);
+				fz_print_stext_page_as_html(ctx, out, text);
+			}
+			else if (output_format == OUT_XHTML)
+			{
+				fz_print_stext_page_as_xhtml(ctx, out, text);
 			}
 			else if (output_format == OUT_TEXT)
 			{
-				fz_print_stext_page(ctx, out, text);
+				fz_print_stext_page_as_text(ctx, out, text);
 				fz_write_printf(ctx, out, "\f\n");
 			}
 		}
 		fz_always(ctx)
 		{
 			fz_drop_device(ctx, dev);
-			dev = NULL;
 			fz_drop_stext_page(ctx, text);
 		}
 		fz_catch(ctx)
 		{
 			fz_drop_display_list(ctx, list);
+			fz_drop_separations(ctx, seps);
 			fz_drop_page(ctx, page);
 			fz_rethrow(ctx);
 		}
@@ -539,19 +619,24 @@ static void dodrawpage(fz_context *ctx, fz_page *page, fz_display_list *list, in
 #if FZ_ENABLE_PDF
 	else if (output_format == OUT_PDF)
 	{
-		fz_buffer *contents;
-		pdf_obj *resources;
+		fz_buffer *contents = NULL;
+		pdf_obj *resources = NULL;
+
+		fz_var(contents);
+		fz_var(resources);
 
-		dev = pdf_page_write(ctx, pdfout, &mediabox, &resources, &contents);
 		fz_try(ctx)
 		{
 			pdf_obj *page_obj;
 
+			dev = pdf_page_write(ctx, pdfout, &mediabox, &resources, &contents);
 			if (list)
 				fz_run_display_list(ctx, list, dev, &fz_identity, NULL, cookie);
 			else
 				fz_run_page(ctx, page, dev, &fz_identity, cookie);
 			fz_close_device(ctx, dev);
+			fz_drop_device(ctx, dev);
+			dev = NULL;
 
 			page_obj = pdf_add_page(ctx, pdfout, &mediabox, rotation, resources, contents);
 			pdf_insert_page(ctx, pdfout, -1, page_obj);
@@ -562,11 +647,11 @@ static void dodrawpage(fz_context *ctx, fz_page *page, fz_display_list *list, in
 			pdf_drop_obj(ctx, resources);
 			fz_drop_buffer(ctx, contents);
 			fz_drop_device(ctx, dev);
-			dev = NULL;
 		}
 		fz_catch(ctx)
 		{
 			fz_drop_display_list(ctx, list);
+			fz_drop_separations(ctx, seps);
 			fz_drop_page(ctx, page);
 			fz_rethrow(ctx);
 		}
@@ -577,26 +662,27 @@ static void dodrawpage(fz_context *ctx, fz_page *page, fz_display_list *list, in
 	{
 		float zoom;
 		fz_matrix ctm;
-		fz_rect bounds, tbounds;
+		fz_rect tbounds;
 		char buf[512];
-		fz_output *out;
+		fz_output *out = NULL;
 
-		if (!output || !strcmp(output, "-"))
-			out = fz_stdout(ctx);
-		else
-		{
-			sprintf(buf, output, pagenum);
-			out = fz_new_output_with_path(ctx, buf, 0);
-		}
+		fz_var(out);
 
-		fz_bound_page(ctx, page, &bounds);
 		zoom = resolution / 72;
 		fz_pre_rotate(fz_scale(&ctm, zoom, zoom), rotation);
-		tbounds = bounds;
+		tbounds = mediabox;
 		fz_transform_rect(&tbounds, &ctm);
 
 		fz_try(ctx)
 		{
+			if (!output || !strcmp(output, "-"))
+				out = fz_stdout(ctx);
+			else
+			{
+				fz_snprintf(buf, sizeof(buf), output, pagenum);
+				out = fz_new_output_with_path(ctx, buf, 0);
+			}
+
 			dev = fz_new_svg_device(ctx, out, tbounds.x1-tbounds.x0, tbounds.y1-tbounds.y0, FZ_SVG_TEXT_AS_PATH, 1);
 			if (lowmemory)
 				fz_enable_device_hints(ctx, dev, FZ_NO_CACHE);
@@ -605,18 +691,17 @@ static void dodrawpage(fz_context *ctx, fz_page *page, fz_display_list *list, in
 			else
 				fz_run_page(ctx, page, dev, &ctm, cookie);
 			fz_close_device(ctx, dev);
-			fz_drop_device(ctx, dev);
-			dev = NULL;
+			fz_close_output(ctx, out);
 		}
 		fz_always(ctx)
 		{
 			fz_drop_device(ctx, dev);
-			dev = NULL;
 			fz_drop_output(ctx, out);
 		}
 		fz_catch(ctx)
 		{
 			fz_drop_display_list(ctx, list);
+			fz_drop_separations(ctx, seps);
 			fz_drop_page(ctx, page);
 			fz_rethrow(ctx);
 		}
@@ -625,18 +710,16 @@ static void dodrawpage(fz_context *ctx, fz_page *page, fz_display_list *list, in
 	{
 		float zoom;
 		fz_matrix ctm;
-		fz_rect bounds, tbounds;
+		fz_rect tbounds;
 		fz_irect ibounds;
 		fz_pixmap *pix = NULL;
 		int w, h;
-		fz_band_writer *bander = NULL;
 		fz_bitmap *bit = NULL;
 
 		fz_var(pix);
 		fz_var(bander);
 		fz_var(bit);
 
-		fz_bound_page(ctx, page, &bounds);
 		zoom = resolution / 72;
 		fz_pre_scale(fz_rotate(&ctm, rotation), zoom, zoom);
 
@@ -645,7 +728,7 @@ static void dodrawpage(fz_context *ctx, fz_page *page, fz_display_list *list, in
 			fz_pre_scale(fz_pre_translate(&ctm, 0, -height), 1, -1);
 		}
 
-		tbounds = bounds;
+		tbounds = mediabox;
 		fz_round_rect(&ibounds, fz_transform_rect(&tbounds, &ctm));
 
 		/* Make local copies of our width/height */
@@ -695,7 +778,7 @@ static void dodrawpage(fz_context *ctx, fz_page *page, fz_display_list *list, in
 			}
 			fz_scale(&scale_mat, scalex, scaley);
 			fz_concat(&ctm, &ctm, &scale_mat);
-			tbounds = bounds;
+			tbounds = mediabox;
 			fz_transform_rect(&tbounds, &ctm);
 		}
 		fz_round_rect(&ibounds, &tbounds);
@@ -729,17 +812,19 @@ static void dodrawpage(fz_context *ctx, fz_page *page, fz_display_list *list, in
 					workers[band].tbounds = tbounds;
 					memset(&workers[band].cookie, 0, sizeof(fz_cookie));
 					workers[band].list = list;
-					workers[band].pix = fz_new_pixmap_with_bbox(ctx, colorspace, &band_ibounds, alpha);
+					workers[band].pix = fz_new_pixmap_with_bbox(ctx, colorspace, &band_ibounds, seps, alpha);
 					fz_set_pixmap_resolution(ctx, workers[band].pix, resolution, resolution);
+#ifndef DISABLE_MUTHREADS
 					DEBUG_THREADS(("Worker %d, Pre-triggering band %d\n", band, band));
 					mu_trigger_semaphore(&workers[band].start);
+#endif
 					ctm.f -= drawheight;
 				}
 				pix = workers[0].pix;
 			}
 			else
 			{
-				pix = fz_new_pixmap_with_bbox(ctx, colorspace, &band_ibounds, alpha);
+				pix = fz_new_pixmap_with_bbox(ctx, colorspace, &band_ibounds, seps, alpha);
 				fz_set_pixmap_resolution(ctx, pix, resolution, resolution);
 			}
 
@@ -758,8 +843,10 @@ static void dodrawpage(fz_context *ctx, fz_page *page, fz_display_list *list, in
 					bander = fz_new_pkm_band_writer(ctx, out);
 				else if (output_format == OUT_PS)
 					bander = fz_new_ps_band_writer(ctx, out);
+				else if (output_format == OUT_PSD)
+					bander = fz_new_psd_band_writer(ctx, out);
 				else if (output_format == OUT_TGA)
-					bander = fz_new_tga_band_writer(ctx, out, colorspace == fz_device_bgr(ctx));
+					bander = fz_new_tga_band_writer(ctx, out, fz_colorspace_is_bgr(ctx, colorspace));
 				else if (output_format == OUT_PWG)
 				{
 					if (out_cs == CS_MONO)
@@ -775,7 +862,9 @@ static void dodrawpage(fz_context *ctx, fz_page *page, fz_display_list *list, in
 						bander = fz_new_color_pcl_band_writer(ctx, out, NULL);
 				}
 				if (bander)
-					fz_write_header(ctx, bander, pix->w, totalheight, pix->n, pix->alpha, pix->xres, pix->yres, output_pagenum++);
+				{
+					fz_write_header(ctx, bander, pix->w, totalheight, pix->n, pix->alpha, pix->xres, pix->yres, output_pagenum++, pix->colorspace, pix->seps);
+				}
 			}
 
 			for (band = 0; band < bands; band++)
@@ -783,8 +872,10 @@ static void dodrawpage(fz_context *ctx, fz_page *page, fz_display_list *list, in
 				if (num_workers > 0)
 				{
 					worker_t *w = &workers[band % num_workers];
+#ifndef DISABLE_MUTHREADS
 					DEBUG_THREADS(("Waiting for worker %d to complete band %d\n", w->num, band));
 					mu_wait_semaphore(&w->stop);
+#endif
 					pix = w->pix;
 					bit = w->bit;
 					w->bit = NULL;
@@ -808,8 +899,10 @@ static void dodrawpage(fz_context *ctx, fz_page *page, fz_display_list *list, in
 					w->ctm = ctm;
 					w->tbounds = tbounds;
 					memset(&w->cookie, 0, sizeof(fz_cookie));
+#ifndef DISABLE_MUTHREADS
 					DEBUG_THREADS(("Triggering worker %d for band %d\n", w->num, w->band));
 					mu_trigger_semaphore(&w->start);
+#endif
 				}
 				ctm.f -= drawheight;
 			}
@@ -828,7 +921,8 @@ static void dodrawpage(fz_context *ctx, fz_page *page, fz_display_list *list, in
 		}
 		fz_always(ctx)
 		{
-			fz_drop_band_writer(ctx, bander);
+			if (output_format != OUT_PCLM)
+				fz_drop_band_writer(ctx, bander);
 			fz_drop_bitmap(ctx, bit);
 			bit = NULL;
 			if (num_workers > 0)
@@ -843,6 +937,7 @@ static void dodrawpage(fz_context *ctx, fz_page *page, fz_display_list *list, in
 		fz_catch(ctx)
 		{
 			fz_drop_display_list(ctx, list);
+			fz_drop_separations(ctx, seps);
 			fz_drop_page(ctx, page);
 			fz_rethrow(ctx);
 		}
@@ -850,9 +945,11 @@ static void dodrawpage(fz_context *ctx, fz_page *page, fz_display_list *list, in
 
 	fz_drop_display_list(ctx, list);
 
-	if (!output_append)
+	if (output_file_per_page)
 		file_level_trailers(ctx);
 
+	fz_drop_separations(ctx, seps);
+
 	fz_drop_page(ctx, page);
 
 	if (showtime)
@@ -904,14 +1001,10 @@ static void dodrawpage(fz_context *ctx, fz_page *page, fz_display_list *list, in
 	fprintf(stderr, "\n");
 
 	if (lowmemory)
-	{
 		fz_empty_store(ctx);
-	}
 
 	if (showmemory)
-	{
 		fz_dump_glyph_cache_stats(ctx);
-	}
 
 	fz_flush_warnings(ctx);
 
@@ -924,7 +1017,9 @@ static void bgprint_flush(void)
 	if (!bgprint.active || !bgprint.started)
 		return;
 
+#ifndef DISABLE_MUTHREADS
 	mu_wait_semaphore(&bgprint.stop);
+#endif
 	bgprint.started = 0;
 }
 
@@ -935,19 +1030,53 @@ static void drawpage(fz_context *ctx, fz_document *doc, int pagenum)
 	fz_device *dev = NULL;
 	int start;
 	fz_cookie cookie = { 0 };
-	int first_page = !output_append;
 	fz_rect bounds;
+	fz_separations *seps = NULL;
 
 	fz_var(list);
 	fz_var(dev);
+	fz_var(seps);
 
 	start = (showtime ? gettime() : 0);
 
 	page = fz_load_page(ctx, doc, pagenum - 1);
 
-	/* Output any file level (as opposed to page level) headers. */
-	if (first_page)
-		file_level_headers(ctx);
+	if (spots != SPOTS_NONE)
+	{
+		fz_try(ctx)
+		{
+			seps = fz_page_separations(ctx, page);
+			if (seps)
+			{
+				int i, n = fz_count_separations(ctx, seps);
+				if (spots == SPOTS_FULL)
+					for (i = 0; i < n; i++)
+						fz_set_separation_behavior(ctx, seps, i, FZ_SEPARATION_SPOT);
+				else
+					for (i = 0; i < n; i++)
+						fz_set_separation_behavior(ctx, seps, i, FZ_SEPARATION_COMPOSITE);
+			}
+			else if (fz_page_uses_overprint(ctx, page))
+			{
+				/* This page uses overprint, so we need an empty
+				 * sep object to force the overprint simulation on. */
+				seps = fz_new_separations(ctx, 0);
+			}
+			else if (oi && fz_colorspace_n(ctx, oi) != fz_colorspace_n(ctx, colorspace))
+			{
+				/* We have an output intent, and it's incompatible
+				 * with the colorspace our device needs. Force the
+				 * overprint simulation on, because this ensures that
+				 * we 'simulate' the output intent too. */
+				seps = fz_new_separations(ctx, 0);
+			}
+		}
+		fz_catch(ctx)
+		{
+			fz_drop_page(ctx, page);
+			fz_rethrow(ctx);
+		}
+	}
 
 	if (uselist)
 	{
@@ -968,6 +1097,7 @@ static void drawpage(fz_context *ctx, fz_document *doc, int pagenum)
 		fz_catch(ctx)
 		{
 			fz_drop_display_list(ctx, list);
+			fz_drop_separations(ctx, seps);
 			fz_drop_page(ctx, page);
 			fz_rethrow(ctx);
 		}
@@ -1000,6 +1130,9 @@ static void drawpage(fz_context *ctx, fz_document *doc, int pagenum)
 		}
 		fz_catch(ctx)
 		{
+			fz_drop_display_list(ctx, list);
+			fz_drop_separations(ctx, seps);
+			fz_drop_page(ctx, page);
 			fz_rethrow(ctx);
 		}
 		fprintf(stderr, " %s", iscolor ? "color" : "grayscale");
@@ -1010,10 +1143,13 @@ static void drawpage(fz_context *ctx, fz_document *doc, int pagenum)
 		char text_buffer[512];
 
 		bgprint_flush();
-		fz_drop_output(ctx, out);
+		if (out)
+		{
+			fz_close_output(ctx, out);
+			fz_drop_output(ctx, out);
+		}
 		fz_snprintf(text_buffer, sizeof(text_buffer), output, pagenum);
-		out = fz_new_output_with_path(ctx, text_buffer, output_append);
-		output_append = 1;
+		out = fz_new_output_with_path(ctx, text_buffer, 0);
 	}
 
 	if (bgprint.active)
@@ -1027,15 +1163,21 @@ static void drawpage(fz_context *ctx, fz_document *doc, int pagenum)
 		bgprint.started = 1;
 		bgprint.page = page;
 		bgprint.list = list;
+		bgprint.seps = seps;
 		bgprint.filename = filename;
 		bgprint.pagenum = pagenum;
 		bgprint.interptime = start;
+#ifndef DISABLE_MUTHREADS
 		mu_trigger_semaphore(&bgprint.start);
+#else
+		fz_drop_display_list(ctx, list);
+		fz_drop_page(ctx, page);
+#endif
 	}
 	else
 	{
 		fprintf(stderr, "page %s %d", filename, pagenum);
-		dodrawpage(ctx, page, list, pagenum, &cookie, start, 0, filename, 0);
+		dodrawpage(ctx, page, list, pagenum, &cookie, start, 0, filename, 0, seps);
 	}
 }
 
@@ -1086,8 +1228,10 @@ parse_colorspace(const char *name)
 		if (!strcmp(name, cs_name_table[i].name))
 			return cs_name_table[i].colorspace;
 	}
-	fprintf(stderr, "Unknown colorspace \"%s\"\n", name);
-	exit(1);
+
+	/* Assume ICC. We will error out later if not the case. */
+	icc_filename = name;
+	return CS_ICC;
 }
 
 typedef struct
@@ -1098,9 +1242,17 @@ typedef struct
 #endif
 } trace_header;
 
+typedef struct
+{
+	size_t current;
+	size_t peak;
+	size_t total;
+} trace_info;
+
 static void *
 trace_malloc(void *arg, size_t size)
 {
+	trace_info *info = (trace_info *) arg;
 	trace_header *p;
 	if (size == 0)
 		return NULL;
@@ -1108,27 +1260,29 @@ trace_malloc(void *arg, size_t size)
 	if (p == NULL)
 		return NULL;
 	p[0].size = size;
-	memtrace_current += size;
-	memtrace_total += size;
-	if (memtrace_current > memtrace_peak)
-		memtrace_peak = memtrace_current;
+	info->current += size;
+	info->total += size;
+	if (info->current > info->peak)
+		info->peak = info->current;
 	return (void *)&p[1];
 }
 
 static void
 trace_free(void *arg, void *p_)
 {
+	trace_info *info = (trace_info *) arg;
 	trace_header *p = (trace_header *)p_;
 
 	if (p == NULL)
 		return;
-	memtrace_current -= p[-1].size;
+	info->current -= p[-1].size;
 	free(&p[-1]);
 }
 
 static void *
 trace_realloc(void *arg, void *p_, size_t size)
 {
+	trace_info *info = (trace_info *) arg;
 	trace_header *p = (trace_header *)p_;
 	size_t oldsize;
 
@@ -1143,11 +1297,11 @@ trace_realloc(void *arg, void *p_, size_t size)
 	p = realloc(&p[-1], size + sizeof(trace_header));
 	if (p == NULL)
 		return NULL;
-	memtrace_current += size - oldsize;
+	info->current += size - oldsize;
 	if (size > oldsize)
-		memtrace_total += size - oldsize;
-	if (memtrace_current > memtrace_peak)
-		memtrace_peak = memtrace_current;
+		info->total += size - oldsize;
+	if (info->current > info->peak)
+		info->peak = info->current;
 	p[0].size = size;
 	return &p[1];
 }
@@ -1187,7 +1341,7 @@ static void bgprint_worker(void *arg)
 		{
 			int start = gettime();
 			memset(&cookie, 0, sizeof(cookie));
-			dodrawpage(bgprint.ctx, bgprint.page, bgprint.list, pagenum, &cookie, start, bgprint.interptime, bgprint.filename, 1);
+			dodrawpage(bgprint.ctx, bgprint.page, bgprint.list, pagenum, &cookie, start, bgprint.interptime, bgprint.filename, 1, bgprint.seps);
 		}
 		DEBUG_THREADS(("BGPrint completed page %d\n", pagenum));
 		mu_trigger_semaphore(&bgprint.stop);
@@ -1205,6 +1359,7 @@ static inline int iswhite(int ch)
 
 static void apply_layer_config(fz_context *ctx, fz_document *doc, const char *lc)
 {
+#if FZ_ENABLE_PDF
 	pdf_document *pdoc = pdf_specifics(ctx, doc);
 	int config = -1;
 	int n, j;
@@ -1299,6 +1454,7 @@ static void apply_layer_config(fz_context *ctx, fz_document *doc, const char *lc
 			fprintf(stderr, " <locked>");
 		fprintf(stderr, "\n");
 	}
+#endif
 }
 
 #ifdef MUDRAW_STANDALONE
@@ -1311,12 +1467,13 @@ int mudraw_main(int argc, char **argv)
 	fz_document *doc = NULL;
 	int c;
 	fz_context *ctx;
-	fz_alloc_context alloc_ctx = { NULL, trace_malloc, trace_realloc, trace_free };
+	trace_info info = { 0, 0, 0 };
+	fz_alloc_context alloc_ctx = { &info, trace_malloc, trace_realloc, trace_free };
 	fz_locks_context *locks = NULL;
 
 	fz_var(doc);
 
-	while ((c = fz_getopt(argc, argv, "p:o:F:R:r:w:h:fB:c:G:Is:A:DiW:H:S:T:U:XLvPl:y:")) != -1)
+	while ((c = fz_getopt(argc, argv, "p:o:F:R:r:w:h:fB:c:e:G:Is:A:DiW:H:S:T:U:XLvPl:y:NO:")) != -1)
 	{
 		switch (c)
 		{
@@ -1335,6 +1492,7 @@ int mudraw_main(int argc, char **argv)
 		case 'B': band_height = atoi(fz_optarg); break;
 
 		case 'c': out_cs = parse_colorspace(fz_optarg); break;
+		case 'e': proof_filename = fz_optarg; break;
 		case 'G': gamma_value = fz_atof(fz_optarg); break;
 		case 'I': invert++; break;
 
@@ -1344,6 +1502,13 @@ int mudraw_main(int argc, char **argv)
 		case 'U': layout_css = fz_optarg; break;
 		case 'X': layout_use_doc_css = 0; break;
 
+		case 'O': spots = fz_atof(fz_optarg);
+#ifndef FZ_ENABLE_SPOT_RENDERING
+			fprintf(stderr, "Spot rendering/Overprint/Overprint simulation not enabled in this build\n");
+			spots = SPOTS_NONE;
+#endif
+			break;
+
 		case 's':
 			if (strchr(fz_optarg, 't')) ++showtime;
 			if (strchr(fz_optarg, 'm')) ++showmemory;
@@ -1365,6 +1530,7 @@ int mudraw_main(int argc, char **argv)
 		case 'D': uselist = 0; break;
 		case 'l': min_line_width = fz_atof(fz_optarg); break;
 		case 'i': ignore_errors = 1; break;
+		case 'N': icc_engine = NULL; break;
 
 		case 'T':
 #ifndef DISABLE_MUTHREADS
@@ -1374,8 +1540,13 @@ int mudraw_main(int argc, char **argv)
 			break;
 #endif
 		case 'L': lowmemory = 1; break;
-		case 'P': bgprint.active = 1; break;
-
+		case 'P':
+#ifndef DISABLE_MUTHREADS
+			bgprint.active = 1; break;
+#else
+			fprintf(stderr, "Threads not enabled in this build\n");
+			break;
+#endif
 		case 'y': layer_config = fz_optarg; break;
 
 		case 'v': fprintf(stderr, "mudraw version %s\n", FZ_VERSION); return 1;
@@ -1424,9 +1595,13 @@ int mudraw_main(int argc, char **argv)
 		exit(1);
 	}
 
+	if (proof_filename)
+		proof_cs = fz_new_icc_colorspace_from_file(ctx, NULL, proof_filename);
+
 	fz_set_text_aa_level(ctx, alphabits_text);
 	fz_set_graphics_aa_level(ctx, alphabits_graphics);
 	fz_set_graphics_min_line_width(ctx, min_line_width);
+	fz_set_cmm_engine(ctx, icc_engine);
 
 #ifndef DISABLE_MUTHREADS
 	if (bgprint.active)
@@ -1490,6 +1665,11 @@ int mudraw_main(int argc, char **argv)
 			if (!strcmp(format, suffix_table[i].suffix+1))
 			{
 				output_format = suffix_table[i].format;
+				if (spots == SPOTS_FULL && suffix_table[i].spots == 0)
+				{
+					fprintf(stderr, "Output format '%s' does not support spot rendering.\nDoing overprint simulation instead.\n", suffix_table[i].suffix+1);
+					spots = SPOTS_OVERPRINT_SIM;
+				}
 				break;
 			}
 		}
@@ -1512,6 +1692,11 @@ int mudraw_main(int argc, char **argv)
 			{
 				suffix = s+1;
 				output_format = suffix_table[i].format;
+				if (spots == SPOTS_FULL && suffix_table[i].spots == 0)
+				{
+					fprintf(stderr, "Output format '%s' does not support spot rendering.\nDoing overprint simulation instead.\n", suffix_table[i].suffix+1);
+					spots = SPOTS_OVERPRINT_SIM;
+				}
 				i = 0;
 			}
 		}
@@ -1519,9 +1704,9 @@ int mudraw_main(int argc, char **argv)
 
 	if (band_height)
 	{
-		if (output_format != OUT_PAM && output_format != OUT_PGM && output_format != OUT_PPM && output_format != OUT_PNM && output_format != OUT_PNG && output_format != OUT_PBM && output_format != OUT_PKM && output_format != OUT_PCL && output_format != OUT_PS)
+		if (output_format != OUT_PAM && output_format != OUT_PGM && output_format != OUT_PPM && output_format != OUT_PNM && output_format != OUT_PNG && output_format != OUT_PBM && output_format != OUT_PKM && output_format != OUT_PCL && output_format != OUT_PCLM && output_format != OUT_PS && output_format != OUT_PSD && output_format != OUT_TGA)
 		{
-			fprintf(stderr, "Banded operation only possible with PAM, PBM, PGM, PKM, PPM, PNM, PCL, PS and PNG outputs\n");
+			fprintf(stderr, "Banded operation only possible with PxM, PCL, PCLM, PS, PSD, PNG and TGA outputs\n");
 			exit(1);
 		}
 		if (showmd5)
@@ -1573,12 +1758,71 @@ int mudraw_main(int argc, char **argv)
 		colorspace = fz_device_cmyk(ctx);
 		alpha = (out_cs == CS_CMYK_ALPHA);
 		break;
+	case CS_ICC:
+		fz_try(ctx)
+			colorspace = fz_new_icc_colorspace_from_file(ctx, NULL, icc_filename);
+		fz_catch(ctx)
+		{
+			fprintf(stderr, "Invalid ICC destination color space\n");
+			exit(1);
+		}
+		if (colorspace == NULL)
+		{
+			fprintf(stderr, "Invalid ICC destination color space\n");
+			exit(1);
+		}
+		alpha = 0;
+		break;
 	default:
 		fprintf(stderr, "Unknown colorspace!\n");
 		exit(1);
 		break;
 	}
 
+	if (out_cs != CS_ICC)
+		colorspace = fz_keep_colorspace(ctx, colorspace);
+	else
+	{
+		int i, j, okay;
+
+		/* Check to make sure this icc profile is ok with the output format */
+		okay = 0;
+		for (i = 0; i < nelem(format_cs_table); i++)
+		{
+			if (format_cs_table[i].format == output_format)
+			{
+				for (j = 0; j < nelem(format_cs_table[i].permitted_cs); j++)
+				{
+					switch (format_cs_table[i].permitted_cs[j])
+					{
+					case CS_MONO:
+					case CS_GRAY:
+					case CS_GRAY_ALPHA:
+						if (fz_colorspace_is_gray(ctx, colorspace))
+							okay = 1;
+						break;
+					case CS_RGB:
+					case CS_RGB_ALPHA:
+						if (fz_colorspace_is_rgb(ctx, colorspace))
+							okay = 1;
+						break;
+					case CS_CMYK:
+					case CS_CMYK_ALPHA:
+						if (fz_colorspace_is_cmyk(ctx, colorspace))
+							okay = 1;
+						break;
+					}
+				}
+			}
+		}
+
+		if (!okay)
+		{
+			fprintf(stderr, "ICC profile uses a colorspace that cannot be used for this format\n");
+			exit(1);
+		}
+	}
+
 #if FZ_ENABLE_PDF
 	if (output_format == OUT_PDF)
 	{
@@ -1599,7 +1843,17 @@ int mudraw_main(int argc, char **argv)
 			out = fz_new_output_with_path(ctx, output, 0);
 	}
 	else
+	{
+#ifdef _WIN32
+		/* Windows specific code to make stdout binary. */
+		if (output_format != OUT_TEXT && output_format != OUT_STEXT && output_format != OUT_HTML && output_format != OUT_XHTML && output_format != OUT_TRACE)
+			setmode(fileno(stdout), O_BINARY);
+#endif
 		out = fz_stdout(ctx);
+	}
+
+	if (!output_file_per_page)
+		file_level_headers(ctx);
 
 	timing.count = 0;
 	timing.total = 0;
@@ -1627,6 +1881,23 @@ int mudraw_main(int argc, char **argv)
 
 				doc = fz_open_document(ctx, filename);
 
+				/* Once document is open check for output intent colorspace */
+				oi = fz_document_output_intent(ctx, doc);
+				if (oi)
+				{
+					/* See if we had explicitly set a profile to render */
+					if (out_cs != CS_ICC)
+					{
+						/* In this case, we want to render to the output intent
+						 * color space if the number of channels is the same */
+						if (fz_colorspace_n(ctx, oi) == fz_colorspace_n(ctx, colorspace))
+						{
+							fz_drop_colorspace(ctx, colorspace);
+							colorspace = fz_keep_colorspace(ctx, oi);
+						}
+					}
+				}
+
 				if (fz_needs_password(ctx, doc))
 				{
 					if (!fz_authenticate_password(ctx, doc, password))
@@ -1656,12 +1927,13 @@ int mudraw_main(int argc, char **argv)
 			}
 			fz_catch(ctx)
 			{
+				fz_drop_document(ctx, doc);
+				doc = NULL;
+
 				if (!ignore_errors)
 					fz_rethrow(ctx);
 
 				bgprint_flush();
-				fz_drop_document(ctx, doc);
-				doc = NULL;
 				fz_warn(ctx, "ignoring error in '%s'", filename);
 			}
 		}
@@ -1674,7 +1946,7 @@ int mudraw_main(int argc, char **argv)
 		errored = 1;
 	}
 
-	if (output_append)
+	if (!output_file_per_page)
 		file_level_trailers(ctx);
 
 #if FZ_ENABLE_PDF
@@ -1693,6 +1965,7 @@ int mudraw_main(int argc, char **argv)
 	}
 	else
 	{
+		fz_close_output(ctx, out);
 		fz_drop_output(ctx, out);
 		out = NULL;
 	}
@@ -1757,6 +2030,8 @@ int mudraw_main(int argc, char **argv)
 	}
 #endif /* DISABLE_MUTHREADS */
 
+	fz_drop_colorspace(ctx, colorspace);
+	fz_drop_colorspace(ctx, proof_cs);
 	fz_drop_context(ctx);
 
 #ifndef DISABLE_MUTHREADS
@@ -1765,9 +2040,9 @@ int mudraw_main(int argc, char **argv)
 
 	if (showmemory)
 	{
-		fprintf(stderr, "Total memory use = " FMT_zu " bytes\n", memtrace_total);
-		fprintf(stderr, "Peak memory use = " FMT_zu " bytes\n", memtrace_peak);
-		fprintf(stderr, "Current memory use = " FMT_zu " bytes\n", memtrace_current);
+		char buf[100];
+		fz_snprintf(buf, sizeof buf, "Memory use total=%zu peak=%zu current=%zu", info.total, info.peak, info.current);
+		fprintf(stderr, "%s\n", buf);
 	}
 
 	return (errored != 0);
