@@ -2,8 +2,12 @@
  * pdfextract -- the ultimate way to extract images and fonts from pdfs
  */
 
+#include "mupdf/fitz.h"
 #include "mupdf/pdf.h"
 
+#include <stdlib.h>
+#include <stdio.h>
+
 static pdf_document *doc = NULL;
 static fz_context *ctx = NULL;
 static int dorgb = 0;
@@ -38,19 +42,19 @@ static void writepixmap(fz_context *ctx, fz_pixmap *pix, char *file, int dorgb)
 
 	if (dorgb && pix->colorspace && pix->colorspace != fz_device_rgb(ctx))
 	{
-		rgb = fz_convert_pixmap(ctx, pix, fz_device_rgb(ctx), 1);
+		rgb = fz_convert_pixmap(ctx, pix, fz_device_rgb(ctx), NULL, NULL, NULL /* FIXME */, 1);
 		pix = rgb;
 	}
 
 	if (pix->n - pix->alpha <= 3)
 	{
-		snprintf(buf, sizeof(buf), "%s.png", file);
+		fz_snprintf(buf, sizeof(buf), "%s.png", file);
 		printf("extracting image %s\n", buf);
 		fz_save_pixmap_as_png(ctx, pix, buf);
 	}
 	else
 	{
-		snprintf(buf, sizeof(buf), "%s.pam", file);
+		fz_snprintf(buf, sizeof(buf), "%s.pam", file);
 		printf("extracting image %s\n", buf);
 		fz_save_pixmap_as_pam(ctx, pix, buf);
 	}
@@ -62,21 +66,19 @@ static void
 writejpeg(fz_context *ctx, const unsigned char *data, size_t len, const char *file)
 {
 	char buf[1024];
-	fz_output *out = NULL;
+	fz_output *out;
 
-	fz_var(out);
-
-	snprintf(buf, sizeof(buf), "%s.jpg", file);
+	fz_snprintf(buf, sizeof(buf), "%s.jpg", file);
 
+	out = fz_new_output_with_path(ctx, buf, 0);
 	fz_try(ctx)
 	{
-		out = fz_new_output_with_path(ctx, buf, 0);
+		printf("extracting image %s\n", buf);
 		fz_write_data(ctx, out, data, len);
+		fz_close_output(ctx, out);
 	}
 	fz_always(ctx)
-	{
 		fz_drop_output(ctx, out);
-	}
 	fz_catch(ctx)
 		fz_rethrow(ctx);
 }
@@ -99,25 +101,31 @@ static void saveimage(int num)
 	{
 		image = pdf_load_image(ctx, doc, ref);
 		cbuf = fz_compressed_image_buffer(ctx, image);
-		snprintf(buf, sizeof(buf), "img-%04d", num);
+		fz_snprintf(buf, sizeof(buf), "img-%04d", num);
 		type = cbuf == NULL ? FZ_IMAGE_UNKNOWN : cbuf->params.type;
+
 		if (image->use_colorkey)
 			type = FZ_IMAGE_UNKNOWN;
 		if (image->use_decode)
 			type = FZ_IMAGE_UNKNOWN;
 		if (image->mask)
 			type = FZ_IMAGE_UNKNOWN;
-		switch (type)
+		if (dorgb)
 		{
-		case FZ_IMAGE_JPEG:
+			enum fz_colorspace_type ctype = fz_colorspace_type(ctx, image->colorspace);
+			if (ctype != FZ_COLORSPACE_RGB && ctype != FZ_COLORSPACE_GRAY)
+				type = FZ_IMAGE_UNKNOWN;
+		}
+
+		if (type == FZ_IMAGE_JPEG)
 		{
 			unsigned char *data;
 			size_t len = fz_buffer_storage(ctx, cbuf->buffer, &data);
-			snprintf(buf, sizeof(buf), "img-%04d", num);
 			writejpeg(ctx, data, len, buf);
 			break;
 		}
-		default:
+		else
+		{
 			pix = fz_get_pixmap_from_image(ctx, image, NULL, NULL, 0, 0);
 			writepixmap(ctx, pix, buf, dorgb);
 		}
@@ -140,7 +148,7 @@ static void savefont(pdf_obj *dict, int num)
 	pdf_obj *obj;
 	char *ext = "";
 	fz_output *out;
-	char *fontname = "font";
+	const char *fontname = "font";
 	size_t len;
 	unsigned char *data;
 
@@ -191,11 +199,14 @@ static void savefont(pdf_obj *dict, int num)
 	len = fz_buffer_storage(ctx, buf, &data);
 	fz_try(ctx)
 	{
-		snprintf(namebuf, sizeof(namebuf), "%s-%04d.%s", fontname, num, ext);
+		fz_snprintf(namebuf, sizeof(namebuf), "%s-%04d.%s", fontname, num, ext);
 		printf("extracting font %s\n", namebuf);
 		out = fz_new_output_with_path(ctx, namebuf, 0);
 		fz_try(ctx)
+		{
 			fz_write_data(ctx, out, data, len);
+			fz_close_output(ctx, out);
+		}
 		fz_always(ctx)
 			fz_drop_output(ctx, out);
 		fz_catch(ctx)
