@@ -2,8 +2,13 @@
  * PDF posteriser; split pages within a PDF file into smaller lumps.
  */
 
+#include "mupdf/fitz.h"
 #include "mupdf/pdf.h"
 
+#include <string.h>
+#include <stdlib.h>
+#include <stdio.h>
+
 static int x_factor = 0;
 static int y_factor = 0;
 
@@ -42,10 +47,10 @@ intersect_box(fz_context *ctx, pdf_document *doc, pdf_obj *page, pdf_obj *box_na
 		old_rect.y1 = mb->y1;
 
 	newbox = pdf_new_array(ctx, doc, 4);
-	pdf_array_push(ctx, newbox, pdf_new_real(ctx, doc, old_rect.x0));
-	pdf_array_push(ctx, newbox, pdf_new_real(ctx, doc, old_rect.y0));
-	pdf_array_push(ctx, newbox, pdf_new_real(ctx, doc, old_rect.x1));
-	pdf_array_push(ctx, newbox, pdf_new_real(ctx, doc, old_rect.y1));
+	pdf_array_push_real(ctx, newbox, old_rect.x0);
+	pdf_array_push_real(ctx, newbox, old_rect.y0);
+	pdf_array_push_real(ctx, newbox, old_rect.x1);
+	pdf_array_push_real(ctx, newbox, old_rect.y1);
 	pdf_dict_put_drop(ctx, page, box_name, newbox);
 }
 
@@ -84,6 +89,7 @@ static void decimatepages(fz_context *ctx, pdf_document *doc)
 		int x, y;
 
 		pdf_page_transform(ctx, page_details, &mediabox, &page_ctm);
+		fz_drop_page(ctx, (fz_page *) page_details);
 
 		w = mediabox.x1 - mediabox.x0;
 		h = mediabox.y1 - mediabox.y0;
@@ -125,10 +131,10 @@ static void decimatepages(fz_context *ctx, pdf_document *doc)
 				else
 					mb.y1 = mediabox.y0 + (h/yf)*(y+1);
 
-				pdf_array_push(ctx, newmediabox, pdf_new_real(ctx, doc, mb.x0));
-				pdf_array_push(ctx, newmediabox, pdf_new_real(ctx, doc, mb.y0));
-				pdf_array_push(ctx, newmediabox, pdf_new_real(ctx, doc, mb.x1));
-				pdf_array_push(ctx, newmediabox, pdf_new_real(ctx, doc, mb.y1));
+				pdf_array_push_real(ctx, newmediabox, mb.x0);
+				pdf_array_push_real(ctx, newmediabox, mb.y0);
+				pdf_array_push_real(ctx, newmediabox, mb.x1);
+				pdf_array_push_real(ctx, newmediabox, mb.y1);
 
 				pdf_dict_put(ctx, newpageobj, PDF_NAME_Parent, pages);
 				pdf_dict_put_drop(ctx, newpageobj, PDF_NAME_MediaBox, newmediabox);
@@ -139,7 +145,8 @@ static void decimatepages(fz_context *ctx, pdf_document *doc)
 				intersect_box(ctx, doc, newpageobj, PDF_NAME_ArtBox, &mb);
 
 				/* Store page object in new kids array */
-				pdf_array_push(ctx, kids, newpageref);
+				pdf_drop_obj(ctx, newpageobj);
+				pdf_array_push_drop(ctx, kids, newpageref);
 
 				kidcount++;
 			}
@@ -147,7 +154,7 @@ static void decimatepages(fz_context *ctx, pdf_document *doc)
 	}
 
 	/* Update page count and kids array */
-	pdf_dict_put_drop(ctx, pages, PDF_NAME_Count, pdf_new_int(ctx, doc, kidcount));
+	pdf_dict_put_int(ctx, pages, PDF_NAME_Count, kidcount);
 	pdf_dict_put_drop(ctx, pages, PDF_NAME_Kids, kids);
 }
 
