@@ -1,4 +1,4 @@
-/* Copyright (C) 2001-2012 Artifex Software, Inc.
+/* Copyright (C) 2001-2018 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -9,8 +9,8 @@
    of the license contained in the file LICENSE in this distribution.
 
    Refer to licensing information at http://www.artifex.com or contact
-   Artifex Software, Inc.,  7 Mt. Lassen Drive - Suite A-134, San Rafael,
-   CA  94903, U.S.A., +1(415)492-9861, for further information.
+   Artifex Software, Inc.,  1305 Grant Avenue - Suite 200, Novato,
+   CA 94945, U.S.A., +1(415)492-9861, for further information.
 */
 
 /*
@@ -378,26 +378,21 @@ typedef struct {
     size_t size;
 } Jbig2WordStreamBuf;
 
-static int
+static void
 jbig2_word_stream_buf_get_next_word(Jbig2WordStream *self, size_t offset, uint32_t *word)
 {
     Jbig2WordStreamBuf *z = (Jbig2WordStreamBuf *) self;
     const byte *data = z->data;
-    uint32_t result;
 
+    *word = 0;
     if (offset + 4 < z->size)
-        result = (data[offset] << 24) | (data[offset + 1] << 16) | (data[offset + 2] << 8) | data[offset + 3];
-    else if (offset > z->size)
-        return -1;
-    else {
+        *word = (data[offset] << 24) | (data[offset + 1] << 16) | (data[offset + 2] << 8) | data[offset + 3];
+    else if (offset <= z->size) {
         size_t i;
 
-        result = 0;
         for (i = 0; i < z->size - offset; i++)
-            result |= data[offset + i] << ((3 - i) << 3);
+            *word |= data[offset + i] << ((3 - i) << 3);
     }
-    *word = result;
-    return 0;
 }
 
 Jbig2WordStream *
