@@ -1,4 +1,4 @@
-/* Copyright (C) 2001-2012 Artifex Software, Inc.
+/* Copyright (C) 2001-2018 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -9,8 +9,8 @@
    of the license contained in the file LICENSE in this distribution.
 
    Refer to licensing information at http://www.artifex.com or contact
-   Artifex Software, Inc.,  7 Mt. Lassen Drive - Suite A-134, San Rafael,
-   CA  94903, U.S.A., +1(415)492-9861, for further information.
+   Artifex Software, Inc.,  1305 Grant Avenue - Suite 200, Novato,
+   CA 94945, U.S.A., +1(415)492-9861, for further information.
 */
 
 /*
@@ -61,6 +61,7 @@ jbig2_generic_stats_size(Jbig2Ctx *ctx, int template)
     byte *gbreg_line = (byte *) image->data;
 
     /* todo: currently we only handle the nominal gbat location */
+    /* when resolved make sure to call jbig2_check_adaptive_pixel_in_field() */
 
 #ifdef OUTPUT_PBM
     printf("P4\n%d %d\n", GBW, GBH);
@@ -96,8 +97,6 @@ jbig2_generic_stats_size(Jbig2Ctx *ctx, int template)
                 bool bit;
 
                 bit = jbig2_arith_decode(as, &GB_stats[CONTEXT]);
-                if (bit < 0)
-                            return -1;
                 result |= bit << (7 - x_minor);
                 CONTEXT = ((CONTEXT & 0x7bf7) << 1) | bit | ((line_m1 >> (7 - x_minor)) & 0x10) | ((line_m2 >> (7 - x_minor)) & 0x800);
             }
@@ -112,6 +111,9 @@ jbig2_generic_stats_size(Jbig2Ctx *ctx, int template)
     return 0;
 }
 
+#define pixel_outside_field(x, y) \
+    ((y) < -128 || (y) > 0 || (x) < -128 || ((y) < 0 && (x) > 127) || ((y) == 0 && (x) >= 0))
+
 static int
 jbig2_decode_generic_template0_unopt(Jbig2Ctx *ctx,
                                      Jbig2Segment *segment,
@@ -123,6 +125,13 @@ jbig2_decode_generic_template0_unopt(Jbig2Ctx *ctx,
     int x, y;
     bool bit;
 
+    if (pixel_outside_field(params->gbat[0], params->gbat[1]) ||
+        pixel_outside_field(params->gbat[2], params->gbat[3]) ||
+        pixel_outside_field(params->gbat[4], params->gbat[5]) ||
+        pixel_outside_field(params->gbat[6], params->gbat[7]))
+        return jbig2_error(ctx, JBIG2_SEVERITY_FATAL, segment->number,
+                           "adaptive template pixel is out of field");
+
     /* this version is generic and easy to understand, but very slow */
 
     for (y = 0; y < GBH; y++) {
@@ -145,8 +154,6 @@ jbig2_decode_generic_template0_unopt(Jbig2Ctx *ctx,
             CONTEXT |= jbig2_image_get_pixel(image, x - 1, y - 2) << 14;
             CONTEXT |= jbig2_image_get_pixel(image, x + params->gbat[6], y + params->gbat[7]) << 15;
             bit = jbig2_arith_decode(as, &GB_stats[CONTEXT]);
-            if (bit < 0)
-                        return -1;
             jbig2_image_set_pixel(image, x, y, bit);
         }
     }
@@ -165,6 +172,7 @@ jbig2_decode_generic_template1(Jbig2Ctx *ctx,
     byte *gbreg_line = (byte *) image->data;
 
     /* todo: currently we only handle the nominal gbat location */
+    /* when resolved make sure to call jbig2_check_adaptive_pixel_in_field() */
 
 #ifdef OUTPUT_PBM
     printf("P4\n%d %d\n", GBW, GBH);
@@ -200,8 +208,6 @@ jbig2_decode_generic_template1(Jbig2Ctx *ctx,
                 bool bit;
 
                 bit = jbig2_arith_decode(as, &GB_stats[CONTEXT]);
-                if (bit < 0)
-                            return -1;
                 result |= bit << (7 - x_minor);
                 CONTEXT = ((CONTEXT & 0xefb) << 1) | bit | ((line_m1 >> (8 - x_minor)) & 0x8) | ((line_m2 >> (8 - x_minor)) & 0x200);
             }
@@ -228,6 +234,7 @@ jbig2_decode_generic_template2(Jbig2Ctx *ctx,
     byte *gbreg_line = (byte *) image->data;
 
     /* todo: currently we only handle the nominal gbat location */
+    /* when resolved make sure to call jbig2_check_adaptive_pixel_in_field() */
 
 #ifdef OUTPUT_PBM
     printf("P4\n%d %d\n", GBW, GBH);
@@ -263,8 +270,6 @@ jbig2_decode_generic_template2(Jbig2Ctx *ctx,
                 bool bit;
 
                 bit = jbig2_arith_decode(as, &GB_stats[CONTEXT]);
-                if (bit < 0)
-                            return -1;
                 result |= bit << (7 - x_minor);
                 CONTEXT = ((CONTEXT & 0x1bd) << 1) | bit | ((line_m1 >> (10 - x_minor)) & 0x4) | ((line_m2 >> (10 - x_minor)) & 0x80);
             }
@@ -326,8 +331,6 @@ jbig2_decode_generic_template2a(Jbig2Ctx *ctx,
                 bool bit;
 
                 bit = jbig2_arith_decode(as, &GB_stats[CONTEXT]);
-                if (bit < 0)
-                            return -1;
                 result |= bit << (7 - x_minor);
                 CONTEXT = ((CONTEXT & 0x1b9) << 1) | bit |
                            ((line_m1 >> (10 - x_minor)) & 0x8) | ((line_m1 >> (9 - x_minor)) & 0x4) | ((line_m2 >> (10 - x_minor)) & 0x80);
@@ -386,8 +389,6 @@ jbig2_decode_generic_template3(Jbig2Ctx *ctx,
                 bool bit;
 
                 bit = jbig2_arith_decode(as, &GB_stats[CONTEXT]);
-                if (bit < 0)
-                            return -1;
                 result |= bit << (7 - x_minor);
                 CONTEXT = ((CONTEXT & 0x1f7) << 1) | bit | ((line_m1 >> (10 - x_minor)) & 0x010);
             }
@@ -414,6 +415,10 @@ jbig2_decode_generic_template3_unopt(Jbig2Ctx *ctx,
     int x, y;
     bool bit;
 
+    if (pixel_outside_field(params->gbat[0], params->gbat[1]))
+        return jbig2_error(ctx, JBIG2_SEVERITY_FATAL, segment->number,
+                           "adaptive template pixel is out of field");
+
     /* this version is generic and easy to understand, but very slow */
 
     for (y = 0; y < GBH; y++) {
@@ -430,8 +435,6 @@ jbig2_decode_generic_template3_unopt(Jbig2Ctx *ctx,
             CONTEXT |= jbig2_image_get_pixel(image, x - 2, y - 1) << 8;
             CONTEXT |= jbig2_image_get_pixel(image, x - 3, y - 1) << 9;
             bit = jbig2_arith_decode(as, &GB_stats[CONTEXT]);
-            if (bit < 0)
-                        return -1;
             jbig2_image_set_pixel(image, x, y, bit);
         }
     }
@@ -464,11 +467,15 @@ jbig2_decode_generic_template0_TPGDON(Jbig2Ctx *ctx,
     bool bit;
     int LTP = 0;
 
+    if (pixel_outside_field(params->gbat[0], params->gbat[1]) ||
+        pixel_outside_field(params->gbat[2], params->gbat[3]) ||
+        pixel_outside_field(params->gbat[4], params->gbat[5]) ||
+        pixel_outside_field(params->gbat[6], params->gbat[7]))
+        return jbig2_error(ctx, JBIG2_SEVERITY_FATAL, segment->number,
+                           "adaptive template pixel is out of field");
+
     for (y = 0; y < GBH; y++) {
-        bit = jbig2_arith_decode(as, &GB_stats[0x9B25]);
-        if (bit < 0)
-                    return -1;
-        LTP ^= bit;
+        LTP ^= jbig2_arith_decode(as, &GB_stats[0x9B25]);
         if (!LTP) {
             for (x = 0; x < GBW; x++) {
                 CONTEXT = jbig2_image_get_pixel(image, x - 1, y);
@@ -488,8 +495,6 @@ jbig2_decode_generic_template0_TPGDON(Jbig2Ctx *ctx,
                 CONTEXT |= jbig2_image_get_pixel(image, x - 1, y - 2) << 14;
                 CONTEXT |= jbig2_image_get_pixel(image, x + params->gbat[6], y + params->gbat[7]) << 15;
                 bit = jbig2_arith_decode(as, &GB_stats[CONTEXT]);
-                if (bit < 0)
-                            return -1;
                 jbig2_image_set_pixel(image, x, y, bit);
             }
         } else {
@@ -512,11 +517,12 @@ jbig2_decode_generic_template1_TPGDON(Jbig2Ctx *ctx,
     bool bit;
     int LTP = 0;
 
+    if (pixel_outside_field(params->gbat[0], params->gbat[1]))
+        return jbig2_error(ctx, JBIG2_SEVERITY_FATAL, segment->number,
+                           "adaptive template pixel is out of field");
+
     for (y = 0; y < GBH; y++) {
-        bit = jbig2_arith_decode(as, &GB_stats[0x0795]);
-        if (bit < 0)
-                    return -1;
-        LTP ^= bit;
+        LTP ^= jbig2_arith_decode(as, &GB_stats[0x0795]);
         if (!LTP) {
             for (x = 0; x < GBW; x++) {
                 CONTEXT = jbig2_image_get_pixel(image, x - 1, y);
@@ -533,8 +539,6 @@ jbig2_decode_generic_template1_TPGDON(Jbig2Ctx *ctx,
                 CONTEXT |= jbig2_image_get_pixel(image, x, y - 2) << 11;
                 CONTEXT |= jbig2_image_get_pixel(image, x - 1, y - 2) << 12;
                 bit = jbig2_arith_decode(as, &GB_stats[CONTEXT]);
-                if (bit < 0)
-                            return -1;
                 jbig2_image_set_pixel(image, x, y, bit);
             }
         } else {
@@ -557,11 +561,12 @@ jbig2_decode_generic_template2_TPGDON(Jbig2Ctx *ctx,
     bool bit;
     int LTP = 0;
 
+    if (pixel_outside_field(params->gbat[0], params->gbat[1]))
+        return jbig2_error(ctx, JBIG2_SEVERITY_FATAL, segment->number,
+                           "adaptive template pixel is out of field");
+
     for (y = 0; y < GBH; y++) {
-        bit = jbig2_arith_decode(as, &GB_stats[0xE5]);
-        if (bit < 0)
-                    return -1;
-        LTP ^= bit;
+        LTP ^= jbig2_arith_decode(as, &GB_stats[0xE5]);
         if (!LTP) {
             for (x = 0; x < GBW; x++) {
                 CONTEXT = jbig2_image_get_pixel(image, x - 1, y);
@@ -575,8 +580,6 @@ jbig2_decode_generic_template2_TPGDON(Jbig2Ctx *ctx,
                 CONTEXT |= jbig2_image_get_pixel(image, x, y - 2) << 8;
                 CONTEXT |= jbig2_image_get_pixel(image, x - 1, y - 2) << 9;
                 bit = jbig2_arith_decode(as, &GB_stats[CONTEXT]);
-                if (bit < 0)
-                            return -1;
                 jbig2_image_set_pixel(image, x, y, bit);
             }
         } else {
@@ -599,11 +602,12 @@ jbig2_decode_generic_template3_TPGDON(Jbig2Ctx *ctx,
     bool bit;
     int LTP = 0;
 
+    if (pixel_outside_field(params->gbat[0], params->gbat[1]))
+        return jbig2_error(ctx, JBIG2_SEVERITY_FATAL, segment->number,
+                           "adaptive template pixel is out of field");
+
     for (y = 0; y < GBH; y++) {
-        bit = jbig2_arith_decode(as, &GB_stats[0x0195]);
-        if (bit < 0)
-                    return -1;
-        LTP ^= bit;
+        LTP ^= jbig2_arith_decode(as, &GB_stats[0x0195]);
         if (!LTP) {
             for (x = 0; x < GBW; x++) {
                 CONTEXT = jbig2_image_get_pixel(image, x - 1, y);
@@ -617,8 +621,6 @@ jbig2_decode_generic_template3_TPGDON(Jbig2Ctx *ctx,
                 CONTEXT |= jbig2_image_get_pixel(image, x - 2, y - 1) << 8;
                 CONTEXT |= jbig2_image_get_pixel(image, x - 3, y - 1) << 9;
                 bit = jbig2_arith_decode(as, &GB_stats[CONTEXT]);
-                if (bit < 0)
-                            return -1;
                 jbig2_image_set_pixel(image, x, y, bit);
             }
         } else {
@@ -671,7 +673,7 @@ jbig2_decode_generic_region(Jbig2Ctx *ctx,
 {
     const int8_t *gbat = params->gbat;
 
-    if (image->stride * image->height > (1 << 24) && segment->data_length < image->stride * image->height / 256) {
+    if (image->stride * image->height > (1 << 26) && segment->data_length < image->stride * image->height / (1 << 16)) {
         return jbig2_error(ctx, JBIG2_SEVERITY_FATAL, segment->number,
                            "region is far larger than data provided (%d << %d), aborting to prevent DOS", segment->data_length, image->stride * image->height);
     }
