@@ -1,4 +1,4 @@
-/* Copyright (C) 2001-2012 Artifex Software, Inc.
+/* Copyright (C) 2001-2018 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -9,8 +9,8 @@
    of the license contained in the file LICENSE in this distribution.
 
    Refer to licensing information at http://www.artifex.com or contact
-   Artifex Software, Inc.,  7 Mt. Lassen Drive - Suite A-134, San Rafael,
-   CA  94903, U.S.A., +1(415)492-9861, for further information.
+   Artifex Software, Inc.,  1305 Grant Avenue - Suite 200, Novato,
+   CA 94945, U.S.A., +1(415)492-9861, for further information.
 */
 
 /*
@@ -61,8 +61,7 @@ huff_get_next_word(Jbig2HuffmanState *hs, uint32_t offset)
     uint32_t word = 0;
     Jbig2WordStream *ws = hs->ws;
 
-    if ((ws->get_next_word(ws, offset, &word)) && ((hs->offset_limit == 0) || (offset < hs->offset_limit)))
-        hs->offset_limit = offset;
+    ws->get_next_word(ws, offset, &word);
     return word;
 }
 
@@ -352,7 +351,7 @@ jbig2_build_huffman_table(Jbig2Ctx *ctx, const Jbig2HuffmanParams *params)
     const Jbig2HuffmanLine *lines = params->lines;
     int n_lines = params->n_lines;
     int i, j;
-    int max_j;
+    uint32_t max_j;
     int log_table_size = 0;
     Jbig2HuffmanTable *result;
     Jbig2HuffmanEntry *entries;
@@ -422,8 +421,9 @@ jbig2_build_huffman_table(Jbig2Ctx *ctx, const Jbig2HuffmanParams *params)
 
             if (PREFLEN == CURLEN) {
                 int RANGELEN = lines[CURTEMP].RANGELEN;
-                int start_j = CURCODE << shift;
-                int end_j = (CURCODE + 1) << shift;
+                uint32_t start_j = CURCODE << shift;
+                uint32_t end_j = (CURCODE + 1) << shift;
+                uint32_t cur_j;
                 byte eflags = 0;
 
                 if (end_j > max_j) {
@@ -439,23 +439,23 @@ jbig2_build_huffman_table(Jbig2Ctx *ctx, const Jbig2HuffmanParams *params)
                 if (CURTEMP == n_lines - (params->HTOOB ? 3 : 2))
                     eflags |= JBIG2_HUFFMAN_FLAGS_ISLOW;
                 if (PREFLEN + RANGELEN > LOG_TABLE_SIZE_MAX) {
-                    for (j = start_j; j < end_j; j++) {
-                        entries[j].u.RANGELOW = lines[CURTEMP].RANGELOW;
-                        entries[j].PREFLEN = PREFLEN;
-                        entries[j].RANGELEN = RANGELEN;
-                        entries[j].flags = eflags;
+                    for (cur_j = start_j; cur_j < end_j; cur_j++) {
+                        entries[cur_j].u.RANGELOW = lines[CURTEMP].RANGELOW;
+                        entries[cur_j].PREFLEN = PREFLEN;
+                        entries[cur_j].RANGELEN = RANGELEN;
+                        entries[cur_j].flags = eflags;
                     }
                 } else {
-                    for (j = start_j; j < end_j; j++) {
-                        int32_t HTOFFSET = (j >> (shift - RANGELEN)) & ((1 << RANGELEN) - 1);
+                    for (cur_j = start_j; cur_j < end_j; cur_j++) {
+                        int32_t HTOFFSET = (cur_j >> (shift - RANGELEN)) & ((1 << RANGELEN) - 1);
 
                         if (eflags & JBIG2_HUFFMAN_FLAGS_ISLOW)
-                            entries[j].u.RANGELOW = lines[CURTEMP].RANGELOW - HTOFFSET;
+                            entries[cur_j].u.RANGELOW = lines[CURTEMP].RANGELOW - HTOFFSET;
                         else
-                            entries[j].u.RANGELOW = lines[CURTEMP].RANGELOW + HTOFFSET;
-                        entries[j].PREFLEN = PREFLEN + RANGELEN;
-                        entries[j].RANGELEN = 0;
-                        entries[j].flags = eflags;
+                            entries[cur_j].u.RANGELOW = lines[CURTEMP].RANGELOW + HTOFFSET;
+                        entries[cur_j].PREFLEN = PREFLEN + RANGELEN;
+                        entries[cur_j].RANGELEN = 0;
+                        entries[cur_j].flags = eflags;
                     }
                 }
                 CURCODE++;
