@@ -4,7 +4,7 @@
 
 #include <time.h>
 
-#if defined(__unix__)
+#if defined(__unix__) || defined(__APPLE__)
 #include <sys/time.h>
 #elif defined(_WIN32)
 #include <sys/timeb.h>
@@ -14,7 +14,7 @@
 
 static double Now(void)
 {
-#if defined(__unix__)
+#if defined(__unix__) || defined(__APPLE__)
 	struct timeval tv;
 	gettimeofday(&tv, NULL);
 	return floor(tv.tv_sec * 1000.0 + tv.tv_usec / 1000.0);
@@ -43,7 +43,7 @@ static double LocalTZA(void)
 
 static double DaylightSavingTA(double t)
 {
-	return 0; // TODO
+	return 0; /* TODO */
 }
 
 /* Helpers from the ECMA 262 specification */
@@ -207,12 +207,17 @@ static double MakeDay(double y, double m, double date)
 	};
 
 	double yd, md;
+	int im;
 
 	y += floor(m / 12);
 	m = pmod(m, 12);
 
+	im = (int)m;
+	if (im < 0 || im >= 12)
+		return NAN;
+
 	yd = floor(TimeFromYear(y) / msPerDay);
-	md = firstDayOfMonth[InLeapYear(y)][(int)m];
+	md = firstDayOfMonth[InLeapYear(y)][im];
 
 	return yd + md + date - 1;
 }
@@ -298,12 +303,15 @@ static double parseDateTime(const char *s)
 
 	if (*s) return NAN;
 
-	if (m > 11) return NAN;
+	if (m < 1 || m > 12) return NAN;
 	if (d < 1 || d > 31) return NAN;
-	if (H > 24 || M > 59 || S > 59) return NAN;
+	if (H < 0 || H > 24) return NAN;
+	if (M < 0 || M > 59) return NAN;
+	if (S < 0 || S > 59) return NAN;
+	if (ms < 0 || ms > 999) return NAN;
 	if (H == 24 && (M != 0 || S != 0 || ms != 0)) return NAN;
 
-	// TODO: DaylightSavingTA on local times
+	/* TODO: DaylightSavingTA on local times */
 	t = MakeDate(MakeDay(y, m-1, d), MakeTime(H, M, S, ms));
 	return t - tza;
 }
