@@ -9,22 +9,24 @@
 
 static const char *astname[] = {
 #include "astnames.h"
+NULL
 };
 
 static const char *opname[] = {
 #include "opnames.h"
+NULL
 };
 
 const char *jsP_aststring(enum js_AstType type)
 {
-	if (type < nelem(astname))
+	if (type < nelem(astname)-1)
 		return astname[type];
 	return "<unknown>";
 }
 
 const char *jsC_opcodestring(enum js_OpCode opcode)
 {
-	if (opcode < nelem(opname))
+	if (opcode < nelem(opname)-1)
 		return opname[opcode];
 	return "<unknown>";
 }
@@ -274,8 +276,12 @@ static void puna(int d, int p, js_Ast *exp, const char *pre, const char *suf)
 
 static void pexpi(int d, int p, js_Ast *exp)
 {
-	int tp = prec(exp->type);
-	int paren = 0;
+	int tp, paren;
+
+	if (!exp) return;
+
+	tp = prec(exp->type);
+	paren = 0;
 	if (tp < p) {
 		pc('(');
 		paren = 1;
@@ -387,7 +393,7 @@ static void pexpi(int d, int p, js_Ast *exp)
 	case EXP_FUN:
 		if (p == 0) pc('(');
 		ps("function ");
-		if (exp->a) pexpi(d, 0, exp->a);
+		pexpi(d, 0, exp->a);
 		pc('(');
 		pargs(d, exp->b);
 		ps(") {\n");
@@ -810,26 +816,26 @@ void js_dumpvalue(js_State *J, js_Value v)
 			break;
 		}
 		switch (v.u.object->type) {
-		case JS_COBJECT: printf("[Object %p]", v.u.object); break;
-		case JS_CARRAY: printf("[Array %p]", v.u.object); break;
+		case JS_COBJECT: printf("[Object %p]", (void*)v.u.object); break;
+		case JS_CARRAY: printf("[Array %p]", (void*)v.u.object); break;
 		case JS_CFUNCTION:
 			printf("[Function %p, %s, %s:%d]",
-				v.u.object,
+				(void*)v.u.object,
 				v.u.object->u.f.function->name,
 				v.u.object->u.f.function->filename,
 				v.u.object->u.f.function->line);
 			break;
 		case JS_CSCRIPT: printf("[Script %s]", v.u.object->u.f.function->filename); break;
-		case JS_CCFUNCTION: printf("[CFunction %p]", v.u.object->u.c.function); break;
+		case JS_CCFUNCTION: printf("[CFunction %s]", v.u.object->u.c.name); break;
 		case JS_CBOOLEAN: printf("[Boolean %d]", v.u.object->u.boolean); break;
 		case JS_CNUMBER: printf("[Number %g]", v.u.object->u.number); break;
 		case JS_CSTRING: printf("[String'%s']", v.u.object->u.s.string); break;
-		case JS_CERROR: printf("[Error %s]", v.u.object->u.s.string); break;
-		case JS_CITERATOR: printf("[Iterator %p]", v.u.object); break;
+		case JS_CERROR: printf("[Error]"); break;
+		case JS_CITERATOR: printf("[Iterator %p]", (void*)v.u.object); break;
 		case JS_CUSERDATA:
 			printf("[Userdata %s %p]", v.u.object->u.user.tag, v.u.object->u.user.data);
 			break;
-		default: printf("[Object %p]", v.u.object); break;
+		default: printf("[Object %p]", (void*)v.u.object); break;
 		}
 		break;
 	}
