@@ -1,6 +1,7 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <unistd.h>
 
 #include "mujs.h"
 
@@ -119,31 +120,68 @@ static const char *require_js =
 	"require.cache = Object.create(null);\n"
 ;
 
-int eval_print(js_State *J, const char *source)
+static const char *stacktrace_js =
+	"Error.prototype.toString = function() {\n"
+	"if (this.stackTrace) return this.name + ': ' + this.message + this.stackTrace;\n"
+	"return this.name + ': ' + this.message;\n"
+	"};\n"
+;
+
+static int eval_print(js_State *J, const char *source)
 {
 	if (js_ploadstring(J, "[string]", source)) {
-		fprintf(stderr, "%s\n", js_tostring(J, -1));
+		fprintf(stderr, "%s\n", js_trystring(J, -1, "Error"));
 		js_pop(J, 1);
 		return 1;
 	}
-	js_pushglobal(J);
+	js_pushundefined(J);
 	if (js_pcall(J, 0)) {
-		fprintf(stderr, "%s\n", js_tostring(J, -1));
+		fprintf(stderr, "%s\n", js_trystring(J, -1, "Error"));
 		js_pop(J, 1);
 		return 1;
 	}
 	if (js_isdefined(J, -1))
-		printf("%s\n", js_tostring(J, -1));
+		printf("%s\n", js_trystring(J, -1, "can't convert to string"));
 	js_pop(J, 1);
 	return 0;
 }
 
+static char *read_stdin(void)
+{
+	int n = 0;
+	int t = 512;
+	char *s = NULL;
+
+	for (;;) {
+		char *ss = realloc(s, t);
+		if (!ss) {
+			free(s);
+			fprintf(stderr, "cannot allocate storage for stdin contents\n");
+			return NULL;
+		}
+		s = ss;
+		n += fread(s + n, 1, t - n - 1, stdin);
+		if (n < t - 1)
+			break;
+		t *= 2;
+	}
+
+	if (ferror(stdin)) {
+		free(s);
+		fprintf(stderr, "error reading stdin\n");
+		return NULL;
+	}
+
+	s[n] = 0;
+	return s;
+}
+
 int
 main(int argc, char **argv)
 {
 	char line[256];
 	js_State *J;
-	int i;
+	int i, status = 0;
 
 	J = js_newstate(NULL, NULL, JS_STRICT);
 
@@ -169,24 +207,30 @@ main(int argc, char **argv)
 	js_setglobal(J, "quit");
 
 	js_dostring(J, require_js);
+	js_dostring(J, stacktrace_js);
 
 	if (argc > 1) {
-		for (i = 1; i < argc; ++i) {
+		for (i = 1; i < argc; ++i)
 			if (js_dofile(J, argv[i]))
-				return 1;
-			js_gc(J, 0);
-		}
+				status = 1;
 	} else {
-		fputs(PS1, stdout);
-		while (fgets(line, sizeof line, stdin)) {
-			eval_print(J, line);
+		if (isatty(0)) {
 			fputs(PS1, stdout);
+			while (fgets(line, sizeof line, stdin)) {
+				eval_print(J, line);
+				fputs(PS1, stdout);
+			}
+			putchar('\n');
+		} else {
+			char *input = read_stdin();
+			if (!input || !js_dostring(J, input))
+				status = 1;
+			free(input);
 		}
-		putchar('\n');
-		js_gc(J, 1);
 	}
 
+	js_gc(J, 0);
 	js_freestate(J);
 
-	return 0;
+	return status;
 }
