@@ -14,8 +14,8 @@
 #define nelem(a) (int)(sizeof (a) / sizeof (a)[0])
 
 #define REPINF 255
-#define MAXTHREAD 1000
 #define MAXSUB REG_MAXSUB
+#define MAXPROG (32 << 10)
 
 typedef struct Reclass Reclass;
 typedef struct Renode Renode;
@@ -250,7 +250,7 @@ static int lexclass(struct cstate *g)
 {
 	int type = L_CCLASS;
 	int quoted, havesave, havedash;
-	Rune save;
+	Rune save = 0;
 
 	newcclass(g);
 
@@ -595,23 +595,25 @@ struct Reinst {
 	Reinst *y;
 };
 
-static int count(Renode *node)
+static int count(struct cstate *g, Renode *node)
 {
-	int min, max;
+	int min, max, n;
 	if (!node) return 0;
 	switch (node->type) {
 	default: return 1;
-	case P_CAT: return count(node->x) + count(node->y);
-	case P_ALT: return count(node->x) + count(node->y) + 2;
+	case P_CAT: return count(g, node->x) + count(g, node->y);
+	case P_ALT: return count(g, node->x) + count(g, node->y) + 2;
 	case P_REP:
 		min = node->m;
 		max = node->n;
-		if (min == max) return count(node->x) * min;
-		if (max < REPINF) return count(node->x) * max + (max - min);
-		return count(node->x) * (min + 1) + 2;
-	case P_PAR: return count(node->x) + 2;
-	case P_PLA: return count(node->x) + 2;
-	case P_NLA: return count(node->x) + 2;
+		if (min == max) n = count(g, node->x) * min;
+		else if (max < REPINF) n = count(g, node->x) * max + (max - min);
+		else n = count(g, node->x) * (min + 1) + 2;
+		if (n < 0 || n > MAXPROG) die(g, "program too large");
+		return n;
+	case P_PAR: return count(g, node->x) + 2;
+	case P_PLA: return count(g, node->x) + 2;
+	case P_NLA: return count(g, node->x) + 2;
 	}
 }
 
@@ -651,6 +653,7 @@ static void compile(Reprog *prog, Renode *node)
 		break;
 
 	case P_REP:
+		inst = NULL; /* silence compiler warning. assert(node->m > 0). */
 		for (i = 0; i < node->m; ++i) {
 			inst = prog->end;
 			compile(prog, node->x);
@@ -807,23 +810,31 @@ static void dumpprog(Reprog *prog)
 }
 #endif
 
-Reprog *regcomp(const char *pattern, int cflags, const char **errorp)
+Reprog *regcompx(void *(*alloc)(void *ctx, void *p, int n), void *ctx,
+	const char *pattern, int cflags, const char **errorp)
 {
 	struct cstate g;
 	Renode *node;
 	Reinst *split, *jump;
-	int i;
+	int i, n;
 
-	g.prog = malloc(sizeof (Reprog));
-	g.pstart = g.pend = malloc(sizeof (Renode) * strlen(pattern) * 2);
+	g.pstart = NULL;
+	g.prog = NULL;
 
 	if (setjmp(g.kaboom)) {
 		if (errorp) *errorp = g.error;
-		free(g.pstart);
-		free(g.prog);
+		alloc(ctx, g.pstart, 0);
+		alloc(ctx, g.prog, 0);
 		return NULL;
 	}
 
+	g.prog = alloc(ctx, NULL, sizeof (Reprog));
+	if (!g.prog)
+		die(&g, "cannot allocate regular expression");
+	g.pstart = g.pend = alloc(ctx, NULL, sizeof (Renode) * strlen(pattern) * 2);
+	if (!g.pstart)
+		die(&g, "cannot allocate regular expression parse list");
+
 	g.source = pattern;
 	g.ncclass = 0;
 	g.nsub = 1;
@@ -839,8 +850,19 @@ Reprog *regcomp(const char *pattern, int cflags, const char **errorp)
 	if (g.lookahead != 0)
 		die(&g, "syntax error");
 
+#ifdef TEST
+	dumpnode(node);
+	putchar('\n');
+#endif
+
+	n = 6 + count(&g, node);
+	if (n < 0 || n > MAXPROG)
+		die(&g, "program too large");
+
 	g.prog->nsub = g.nsub;
-	g.prog->start = g.prog->end = malloc((count(node) + 6) * sizeof (Reinst));
+	g.prog->start = g.prog->end = alloc(ctx, NULL, n * sizeof (Reinst));
+	if (!g.prog->start)
+		die(&g, "cannot allocate regular expression instruction list");
 
 	split = emit(g.prog, I_SPLIT);
 	split->x = split + 3;
@@ -854,25 +876,38 @@ Reprog *regcomp(const char *pattern, int cflags, const char **errorp)
 	emit(g.prog, I_END);
 
 #ifdef TEST
-	dumpnode(node);
-	putchar('\n');
 	dumpprog(g.prog);
 #endif
 
-	free(g.pstart);
+	alloc(ctx, g.pstart, 0);
 
 	if (errorp) *errorp = NULL;
 	return g.prog;
 }
 
-void regfree(Reprog *prog)
+void regfreex(void *(*alloc)(void *ctx, void *p, int n), void *ctx, Reprog *prog)
 {
 	if (prog) {
-		free(prog->start);
-		free(prog);
+		alloc(ctx, prog->start, 0);
+		alloc(ctx, prog, 0);
 	}
 }
 
+static void *default_alloc(void *ctx, void *p, int n)
+{
+	return realloc(p, (size_t)n);
+}
+
+Reprog *regcomp(const char *pattern, int cflags, const char **errorp)
+{
+	return regcompx(default_alloc, NULL, pattern, cflags, errorp);
+}
+
+void regfree(Reprog *prog)
+{
+	regfreex(default_alloc, NULL, prog);
+}
+
 /* Match */
 
 static int isnewline(int c)
@@ -923,169 +958,155 @@ static int strncmpcanon(const char *a, const char *b, int n)
 	return 0;
 }
 
-struct Rethread {
-	Reinst *pc;
-	const char *sp;
-	Resub sub;
-};
-
-static void spawn(Rethread *t, Reinst *pc, const char *sp, Resub *sub)
-{
-	t->pc = pc;
-	t->sp = sp;
-	memcpy(&t->sub, sub, sizeof t->sub);
-}
-
 static int match(Reinst *pc, const char *sp, const char *bol, int flags, Resub *out)
 {
-	Rethread ready[MAXTHREAD];
 	Resub scratch;
-	Resub sub;
-	Rune c;
-	int nready;
 	int i;
+	Rune c;
 
-	/* queue initial thread */
-	spawn(ready + 0, pc, sp, out);
-	nready = 1;
-
-	/* run threads in stack order */
-	while (nready > 0) {
-		--nready;
-		pc = ready[nready].pc;
-		sp = ready[nready].sp;
-		memcpy(&sub, &ready[nready].sub, sizeof sub);
-		for (;;) {
-			switch (pc->opcode) {
-			case I_END:
-				for (i = 0; i < MAXSUB; ++i) {
-					out->sub[i].sp = sub.sub[i].sp;
-					out->sub[i].ep = sub.sub[i].ep;
-				}
+	for (;;) {
+		switch (pc->opcode) {
+		case I_END:
+			return 1;
+		case I_JUMP:
+			pc = pc->x;
+			break;
+		case I_SPLIT:
+			scratch = *out;
+			if (match(pc->x, sp, bol, flags, &scratch)) {
+				*out = scratch;
 				return 1;
-			case I_JUMP:
-				pc = pc->x;
-				continue;
-			case I_SPLIT:
-				if (nready >= MAXTHREAD) {
-					fprintf(stderr, "regexec: backtrack overflow!\n");
+			}
+			pc = pc->y;
+			break;
+
+		case I_PLA:
+			if (!match(pc->x, sp, bol, flags, out))
+				return 0;
+			pc = pc->y;
+			break;
+		case I_NLA:
+			scratch = *out;
+			if (match(pc->x, sp, bol, flags, &scratch))
+				return 0;
+			pc = pc->y;
+			break;
+
+		case I_ANYNL:
+			sp += chartorune(&c, sp);
+			if (c == 0)
+				return 0;
+			pc = pc + 1;
+			break;
+		case I_ANY:
+			sp += chartorune(&c, sp);
+			if (c == 0)
+				return 0;
+			if (isnewline(c))
+				return 0;
+			pc = pc + 1;
+			break;
+		case I_CHAR:
+			sp += chartorune(&c, sp);
+			if (c == 0)
+				return 0;
+			if (flags & REG_ICASE)
+				c = canon(c);
+			if (c != pc->c)
+				return 0;
+			pc = pc + 1;
+			break;
+		case I_CCLASS:
+			sp += chartorune(&c, sp);
+			if (c == 0)
+				return 0;
+			if (flags & REG_ICASE) {
+				if (!incclasscanon(pc->cc, canon(c)))
+					return 0;
+			} else {
+				if (!incclass(pc->cc, c))
 					return 0;
-				}
-				spawn(&ready[nready++], pc->y, sp, &sub);
-				pc = pc->x;
-				continue;
-
-			case I_PLA:
-				if (!match(pc->x, sp, bol, flags, &sub))
-					goto dead;
-				pc = pc->y;
-				continue;
-			case I_NLA:
-				memcpy(&scratch, &sub, sizeof scratch);
-				if (match(pc->x, sp, bol, flags, &scratch))
-					goto dead;
-				pc = pc->y;
-				continue;
-
-			case I_ANYNL:
-				sp += chartorune(&c, sp);
-				if (c == 0)
-					goto dead;
-				break;
-			case I_ANY:
-				sp += chartorune(&c, sp);
-				if (c == 0)
-					goto dead;
-				if (isnewline(c))
-					goto dead;
-				break;
-			case I_CHAR:
-				sp += chartorune(&c, sp);
-				if (c == 0)
-					goto dead;
-				if (flags & REG_ICASE)
-					c = canon(c);
-				if (c != pc->c)
-					goto dead;
-				break;
-			case I_CCLASS:
-				sp += chartorune(&c, sp);
-				if (c == 0)
-					goto dead;
-				if (flags & REG_ICASE) {
-					if (!incclasscanon(pc->cc, canon(c)))
-						goto dead;
-				} else {
-					if (!incclass(pc->cc, c))
-						goto dead;
-				}
-				break;
-			case I_NCCLASS:
-				sp += chartorune(&c, sp);
-				if (c == 0)
-					goto dead;
-				if (flags & REG_ICASE) {
-					if (incclasscanon(pc->cc, canon(c)))
-						goto dead;
-				} else {
-					if (incclass(pc->cc, c))
-						goto dead;
-				}
-				break;
-			case I_REF:
-				i = sub.sub[pc->n].ep - sub.sub[pc->n].sp;
-				if (flags & REG_ICASE) {
-					if (strncmpcanon(sp, sub.sub[pc->n].sp, i))
-						goto dead;
-				} else {
-					if (strncmp(sp, sub.sub[pc->n].sp, i))
-						goto dead;
-				}
-				if (i > 0)
-					sp += i;
-				break;
-
-			case I_BOL:
-				if (sp == bol && !(flags & REG_NOTBOL))
-					break;
-				if (flags & REG_NEWLINE)
-					if (sp > bol && isnewline(sp[-1]))
-						break;
-				goto dead;
-			case I_EOL:
-				if (*sp == 0)
-					break;
-				if (flags & REG_NEWLINE)
-					if (isnewline(*sp))
-						break;
-				goto dead;
-			case I_WORD:
-				i = sp > bol && iswordchar(sp[-1]);
-				i ^= iswordchar(sp[0]);
-				if (i)
-					break;
-				goto dead;
-			case I_NWORD:
-				i = sp > bol && iswordchar(sp[-1]);
-				i ^= iswordchar(sp[0]);
-				if (!i)
-					break;
-				goto dead;
-
-			case I_LPAR:
-				sub.sub[pc->n].sp = sp;
-				break;
-			case I_RPAR:
-				sub.sub[pc->n].ep = sp;
-				break;
-			default:
-				goto dead;
 			}
 			pc = pc + 1;
+			break;
+		case I_NCCLASS:
+			sp += chartorune(&c, sp);
+			if (c == 0)
+				return 0;
+			if (flags & REG_ICASE) {
+				if (incclasscanon(pc->cc, canon(c)))
+					return 0;
+			} else {
+				if (incclass(pc->cc, c))
+					return 0;
+			}
+			pc = pc + 1;
+			break;
+		case I_REF:
+			i = out->sub[pc->n].ep - out->sub[pc->n].sp;
+			if (flags & REG_ICASE) {
+				if (strncmpcanon(sp, out->sub[pc->n].sp, i))
+					return 0;
+			} else {
+				if (strncmp(sp, out->sub[pc->n].sp, i))
+					return 0;
+			}
+			if (i > 0)
+				sp += i;
+			pc = pc + 1;
+			break;
+
+		case I_BOL:
+			if (sp == bol && !(flags & REG_NOTBOL)) {
+				pc = pc + 1;
+				break;
+			}
+			if (flags & REG_NEWLINE) {
+				if (sp > bol && isnewline(sp[-1])) {
+					pc = pc + 1;
+					break;
+				}
+			}
+			return 0;
+		case I_EOL:
+			if (*sp == 0) {
+				pc = pc + 1;
+				break;
+			}
+			if (flags & REG_NEWLINE) {
+				if (isnewline(*sp)) {
+					pc = pc + 1;
+					break;
+				}
+			}
+			return 0;
+		case I_WORD:
+			i = sp > bol && iswordchar(sp[-1]);
+			i ^= iswordchar(sp[0]);
+			if (!i)
+				return 0;
+			pc = pc + 1;
+			break;
+		case I_NWORD:
+			i = sp > bol && iswordchar(sp[-1]);
+			i ^= iswordchar(sp[0]);
+			if (i)
+				return 0;
+			pc = pc + 1;
+			break;
+
+		case I_LPAR:
+			out->sub[pc->n].sp = sp;
+			pc = pc + 1;
+			break;
+		case I_RPAR:
+			out->sub[pc->n].ep = sp;
+			pc = pc + 1;
+			break;
+		default:
+			return 0;
 		}
-dead: ;
 	}
-	return 0;
 }
 
 int regexec(Reprog *prog, const char *sp, Resub *sub, int eflags)
